/*
 * linux/arch/arm/mach-pvc01s/core.c
 *
 * Copyright (c) 2012 Panasonic, Inc
 *
 * based on : arm versatile arch
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/init.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <linux/sysdev.h>
#include <linux/interrupt.h>
#include <linux/amba/bus.h>
#include <linux/amba/mmci.h>
#include <linux/clocksource.h>
#include <linux/clockchips.h>
#include <linux/cnt32_to_63.h>
#include <linux/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/plat-ram.h>

#include <asm/clkdev.h>
#include <asm/system.h>
#include <mach/hardware.h>
#include <mach/pvc01s.h>
#include <asm/irq.h>
#include <asm/hardware/arm_timer.h>
#include <asm/mach-types.h>

#include <asm/mach/arch.h>
#include <asm/mach/flash.h>
#include <asm/mach/irq.h>
#include <asm/mach/time.h>
#include <asm/mach/map.h>

#include <mach/memory.h>

#include "core.h"
#include "clock.h"

#define MTD_DRAM_ROOTFS

//	io
static struct map_desc pvc01s_io_desc[] __initdata = 
{
	//	interrupt group
	{
		.virtual	=  IO_ADDRESS(PVC01S_IRC_GROUP_VIRT_ADDR),
		.pfn		= __phys_to_pfn(PVC01S_IRCGROUP_PHY_ADDR),
		.length		= SZ_4K,
		.type		= MT_DEVICE
	},

	//	interrupt controller
	{
		.virtual	=  IO_ADDRESS(PVC01S_IRC_TOP_MAP_VIRT_ADDR),
		.pfn		= __phys_to_pfn(PVC01S_IRC_MAPPED_TOP_PHY_ADDR),
		.length		= SZ_4K,		//	PVC01S_IRC_PHY_SIZE=256byte
		.type		= MT_DEVICE
	}, 

	//	timer
	{
		.virtual	=  IO_ADDRESS(PVC01S_TIMER_VIRT_ADDR),
		.pfn		= __phys_to_pfn(PVC01S_TIMER1_PHY_ADDR),
		.length		= SZ_4K,
		.type		= MT_DEVICE
	},
};

void __init pvc01s_map_io(void)
{
	iotable_init(pvc01s_io_desc, ARRAY_SIZE(pvc01s_io_desc));
}

#ifdef MTD_DRAM_ROOTFS
static struct mtd_partition pvc01s_ram_partitions[] = 
{
	{
		.name	= "testrootfs",
		.offset = 0,
		.size	= PVC01S_MTD_DRAM_ROOTFS_SIZE,
	},
};

struct platdata_mtd_ram pvc01s_dram_pdata = {
	.mapname	= "DRAM",
	.bankwidth	= 4,
	.partitions		= pvc01s_ram_partitions,
	.nr_partitions	= ARRAY_SIZE(pvc01s_ram_partitions),
};

static struct resource pvc01s_dram_resource[] = {
	[0] = {
		.start = PVC01S_MTD_DRAM_ROOTFS_PHYS_ADDR,
		.end   = PVC01S_MTD_DRAM_ROOTFS_PHYS_ADDR + PVC01S_MTD_DRAM_ROOTFS_SIZE - 1,
		.flags = IORESOURCE_MEM,
	},
};

static struct platform_device pvc01s_dram = {
	.name		= "mtd-ram",
	.id		= 0,
	.resource	= pvc01s_dram_resource,
	.num_resources	= ARRAY_SIZE(pvc01s_dram_resource),
	.dev	= {
		.platform_data = &pvc01s_dram_pdata,
	},
};
#else
//
//	sample for onenand
//
static struct mtd_partition pvc01s_onenand_partitions[] = 
{
	{
		.name	= "kernel",
		.offset = 0,
		.size	= SZ_1M * 2,
	}, {
		.name	= "rootfs",
		.offset	= MTDPART_OFS_APPEND,
		.size	= SZ_1M * 8,
	}, 
};

static struct flash_platform_data pvc01s_onenand_data = {
	.parts		= pvc01s_onenand_partitions,
	.nr_parts	= ARRAY_SIZE(pvc01s_onenand_partitions),
};

static struct resource pvc01s_onenand_resource = {
	.start			= PVC01S_CS0_PHY_ADDR,
	.end			= (PVC01S_CS_SIZE*2) - 1,
	.flags			= IORESOURCE_MEM,
};

static struct platform_device pvc01s_nand_device = {
	.name			= "onenand",
	.id			= 0,
	.dev			= {
		.platform_data	= &pvc01s_onenand_data,
	},
	.num_resources		= 1,
	.resource		= &pvc01s_onenand_resource,
};
#endif

//	primecells
#define UART1_IRQ	{ IRQ_UARTINT1, NO_IRQ }
AMBA_DEVICE(uart1, "dev:uart",  UART1,    NULL);

static struct amba_device *amba_devs[] __initdata = {
	&uart1_device,
};

//	These are fixed clocks.
static struct clk ref135_clk = {
	.rate	= 13500000,
};

static struct clk_lookup lookups[] = {
	//	uart1
	{	
		.dev_id		= "dev:uart",
		.clk		= &ref135_clk,
	}, 
	//	timer
	{	
		.dev_id		= "dev:timer",
		.clk		= &ref135_clk,
	}
};


/*****************************************************************************
 * SD/SDIO/MMC
 ****************************************************************************/
static struct resource pvcsdio_resources[] = {
	[0] = {
		.start	= PVCSD_SDX_BASE,
		.end	= PVCSD_SDX_BASE + SZ_4K - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= IRQ_SDIO,
		.end	= IRQ_SDIO,
		.flags	= IORESOURCE_IRQ,
	},
};

static u64 pvcsdio_dmamask = 0xffffffffUL;

static struct platform_device pvc01s_sdio = {
	.name		= "pvsdio",
	.id		= -1,
	.dev		= {
		.dma_mask = &pvcsdio_dmamask,
		.coherent_dma_mask = 0xffffffff,
	},
	.num_resources	= ARRAY_SIZE(pvcsdio_resources),
	.resource	= pvcsdio_resources,
};

void __init pvc01s_sdio_init(struct pvcsdio_platform_data *pvcsdio_data)
{
	pvcsdio_data->clock = 43875000;
	pvc01s_sdio.dev.platform_data = pvcsdio_data;
	platform_device_register(&pvc01s_sdio);
}


static struct pvcsdio_platform_data pvcsdio_data;

void __init pvc01s_init(void)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(lookups); i++)
		clkdev_add(&lookups[i]);

#ifdef MTD_DRAM_ROOTFS
	platform_device_register(&pvc01s_dram);
#else
	platform_device_register(&pvc01s_nand_device);
#endif

	for (i = 0; i < ARRAY_SIZE(amba_devs); i++) {
		struct amba_device *d = amba_devs[i];
		amba_device_register(d, &iomem_resource);
	}
	pvc01s_sdio_init(&pvcsdio_data);
}

//	timer
extern struct sys_timer pvc01s_timer;
//	interrupt
extern void __init pvc01s_init_interrupts(void);

MACHINE_START(PVC01S, "PVC01S board")
	.phys_io	= 0xfff21000,
	.io_pg_offst	= ((0xfff21000) >> 18) & 0xfffc,
#if 0 /* not use this boot_params area, set params from CONFIG_CMDLINE */
	.boot_params	= PHYS_OFFSET+0x100,
#endif
	.map_io		= pvc01s_map_io,
	.init_irq	= pvc01s_init_interrupts,
	.timer		= &pvc01s_timer,
	.init_machine	= pvc01s_init,
MACHINE_END
