/*
 * linux/arch/arm/mach-uniphier/irq.c
 *
 * Copyright (C) 2011 Panasonic Corporation
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */
#include <linux/module.h>
#include <linux/irq.h>
#include <linux/interrupt.h>

#include <asm/io.h>
#include <asm/hardware/gic.h>
#include <mach/hardware.h>
#include <mach/smp.h>

#include <mach/intctl-regs.h>
#ifdef CONFIG_LINUX_CLI_MASK
#define in_fastintr()	\
	(__mn_intc_get_mask_priority() < CONFIG_LINUX_CLI_LEVEL)

#define FASTINTR_IPI_IRQ	8
#define FASTINTR_IPI_LEVEL	0

#ifdef CONFIG_SMP
# include <asm/spinlock.h>
# define fastintr_spin_lock(lock)	__raw_spin_lock(&(lock)->raw_lock)
# define fastintr_spin_unlock(lock)	__raw_spin_unlock(&(lock)->raw_lock)
#else /* CONFIG_SMP */
# define fastintr_spin_lock(lock)	do {} while (0)
# define fastintr_spin_unlock(lock)	do {} while (0)
#endif /* CONFIG_SMP */
#endif /* CONFIG_LINUX_CLI_MASK */

static void mn_cpupic_noop(unsigned int irq)
{
}

static struct irq_chip mn_cpu_pic_empty = {
	.name		= "cpu_empty",
	.disable	= mn_cpupic_noop,
	.enable		= mn_cpupic_noop,
	.ack		= mn_cpupic_noop,
	.mask		= mn_cpupic_noop,
	.mask_ack	= mn_cpupic_noop,
	.unmask		= mn_cpupic_noop,
};


extern spinlock_t irq_controller_lock;

/*
 * change the level at which an IRQ executes
 * - must not be called whilst interrupts are being processed!
 */
void mn_intc_set_level(unsigned int irq, unsigned int pri)
{
	unsigned long flags;
	int shift;
	u32 mask, new_pri, tmp32, reg;

#ifdef CONFIG_LINUX_CLI_MASK
	if (in_fastintr()) {
		FI_BUG("mn_intc_set_level : call in fast intrrupt");
	}
#endif /* CONFIG_LINUX_CLI_MASK */
	if (in_interrupt())
		BUG();

	shift = (irq % 4) * 8;
	mask = 0xff << shift;
	new_pri = ((pri << UNIPHIER_GIC_PRI_SHIFT) << shift) & mask;
	reg = UNIPHIER_GIC_DIST_START + GIC_DIST_PRI + (irq / 4) * 4;

	spin_lock_irqsave(&irq_controller_lock, flags);
	tmp32 = inl(reg);
	tmp32 = (tmp32 & ~mask) | new_pri;
	outl(tmp32, reg);
	spin_unlock_irqrestore(&irq_controller_lock, flags);
}
EXPORT_SYMBOL(mn_intc_set_level);


#ifdef CONFIG_LINUX_CLI_MASK
void mn_fastintr_clear(unsigned int irq)
{
	outl(irq, UNIPHIER_GIC_CPU_START + GIC_CPU_EOI);
}
EXPORT_SYMBOL(mn_fastintr_clear);
#endif /* CONFIG_LINUX_CLI_MASK */


void mn_intc_clear(unsigned int irq)
{
	unsigned long flags;
	u32 bit = 1 << (irq % 32);
	u32 status;

#ifdef CONFIG_LINUX_CLI_MASK
	if (in_fastintr()) {
		outl(bit,
		     UNIPHIER_GIC_DIST_START + GIC_DIST_PENDING_CLEAR + (irq / 32) * 4);
		return;
	}
	local_fastintr_save(flags);
	spin_lock(&irq_controller_lock);
#else /* CONFIG_LINUX_CLI_MASK */
	spin_lock_irqsave(&irq_controller_lock, flags);
#endif /* CONFIG_LINUX_CLI_MASK */
	status = inl(UNIPHIER_GIC_DIST_START + GIC_DIST_ACTIVE_BIT + (irq / 32) * 4);
	if (status & bit) {
		outl(irq, UNIPHIER_GIC_CPU_START + GIC_CPU_EOI);
	}
	else {
		outl(bit, UNIPHIER_GIC_DIST_START + GIC_DIST_PENDING_CLEAR + (irq / 32) * 4);
	}
#ifdef CONFIG_LINUX_CLI_MASK
	spin_unlock(&irq_controller_lock);
	local_fastintr_restore(flags);
#else /* CONFIG_LINUX_CLI_MASK */
	spin_unlock_irqrestore(&irq_controller_lock, flags);
#endif /* CONFIG_LINUX_CLI_MASK */
}
EXPORT_SYMBOL(mn_intc_clear);


void mn_intc_set(unsigned int irq)
{
	u32 bit = 1 << (irq % 32);
	u32 reg = UNIPHIER_GIC_DIST_START + GIC_DIST_PENDING_SET + (irq / 32) * 4;

	outl(bit, reg);
}
EXPORT_SYMBOL(mn_intc_set);


void mn_intc_enable(unsigned int irq)
{
	u32 bit = 1 << (irq % 32);
	u32 reg = UNIPHIER_GIC_DIST_START + GIC_DIST_ENABLE_SET + (irq / 32) * 4;

	outl(bit, reg);
}
EXPORT_SYMBOL(mn_intc_enable);


void mn_intc_disable(unsigned int irq)
{
	u32 bit = 1 << (irq % 32);
	u32 reg = UNIPHIER_GIC_DIST_START + GIC_DIST_ENABLE_CLEAR + (irq / 32) * 4;

	outl(bit, reg);
}
EXPORT_SYMBOL(mn_intc_disable);


extern void gic_set_irq_chip(unsigned int gic_nr, unsigned int irq);

void mn_set_lateack_irq_type(int irq)
{
	gic_set_irq_chip(0, irq);
}
EXPORT_SYMBOL(mn_set_lateack_irq_type);


void mn_set_empty_irq_type(int irq)
{
	set_irq_chip(irq, &mn_cpu_pic_empty);
}
EXPORT_SYMBOL(mn_set_empty_irq_type);


unsigned int mn_intc_get_level(unsigned int irq)
{
	int shift;
	u32 mask, tmp32, reg;

	shift = (irq % 4) * 8;
	mask = 0xff << shift;
	reg = UNIPHIER_GIC_DIST_START + GIC_DIST_PRI + (irq / 4) * 4;

	tmp32 = inl(reg);

	return (unsigned int)(((tmp32 & mask) >> shift) >> UNIPHIER_GIC_PRI_SHIFT);
}


unsigned int mn_intc_get_running_priority(void)
{
	unsigned long flags;
	unsigned int pri;
	u32 reg = UNIPHIER_GIC_CPU_START + GIC_CPU_RUNNINGPRI;

	local_irq_save(flags);
	pri = inl(reg);
	local_irq_restore(flags);
	return pri;
}

#define UNIPHIER_INTC_MAX_PREEMPT_COUNT	16

static unsigned int cpu0_mask_priority[UNIPHIER_INTC_MAX_PREEMPT_COUNT];
static unsigned int cpu1_mask_priority[UNIPHIER_INTC_MAX_PREEMPT_COUNT];

unsigned int *uniphier_intc_mask_stack[] = {
	cpu0_mask_priority,
	cpu1_mask_priority
};

unsigned int __mn_intc_get_mask_priority(void)
{
	u32 reg = UNIPHIER_GIC_CPU_START + GIC_CPU_PRIMASK;
	return inl(reg) >> UNIPHIER_GIC_PRI_SHIFT;
}
EXPORT_SYMBOL(__mn_intc_get_mask_priority);

#ifdef CONFIG_LINUX_CLI_MASK
extern void __iomem *gic_cpu_base_addr;
#endif /* CONFIG_LINUX_CLI_MASK */

unsigned int mn_intc_get_mask_priority(void)
{
	unsigned int pri;
	unsigned long flags;

#ifdef CONFIG_LINUX_CLI_MASK
	if (!gic_cpu_base_addr) {
		return 0;
	}
	local_fastintr_save(flags);
#else /* CONFIG_LINUX_CLI_MASK */
	local_irq_save(flags);
#endif /* CONFIG_LINUX_CLI_MASK */
	pri = __mn_intc_get_mask_priority();
#ifdef CONFIG_LINUX_CLI_MASK
	local_fastintr_restore(flags);
#else /* CONFIG_LINUX_CLI_MASK */
	local_irq_restore(flags);
#endif /* CONFIG_LINUX_CLI_MASK */
	return pri;
}
EXPORT_SYMBOL(mn_intc_get_mask_priority);


void __mn_intc_set_mask_priority(unsigned int pri)
{
	u32 reg = UNIPHIER_GIC_CPU_START + GIC_CPU_PRIMASK;
	outl(pri << UNIPHIER_GIC_PRI_SHIFT, reg);
}


void mn_intc_set_mask_priority(unsigned int pri)
{
	unsigned long flags;
#ifdef CONFIG_LINUX_CLI_MASK
	if (!gic_cpu_base_addr) {
		return;
	}
	local_fastintr_save(flags);
#else /* CONFIG_LINUX_CLI_MASK */
	local_irq_save(flags);
#endif /* CONFIG_LINUX_CLI_MASK */
	__mn_intc_set_mask_priority(pri);
#ifdef CONFIG_LINUX_CLI_MASK
	local_fastintr_restore(flags);
#else /* CONFIG_LINUX_CLI_MASK */
	local_irq_restore(flags);
#endif /* CONFIG_LINUX_CLI_MASK */
}
EXPORT_SYMBOL(mn_intc_set_mask_priority);


unsigned int mn_intc_set_and_get_mask_priority(unsigned int pri)
{
	unsigned int ret;
	unsigned long flags;
#ifdef CONFIG_LINUX_CLI_MASK
	if (!gic_cpu_base_addr) {
		return 0;
	}
	local_fastintr_save(flags);
#else /* CONFIG_LINUX_CLI_MASK */
	local_irq_save(flags);
#endif /* CONFIG_LINUX_CLI_MASK */
	ret = __mn_intc_get_mask_priority();
	__mn_intc_set_mask_priority(pri);
#ifdef CONFIG_LINUX_CLI_MASK
	local_fastintr_restore(flags);
#else /* CONFIG_LINUX_CLI_MASK */
	local_irq_restore(flags);
#endif /* CONFIG_LINUX_CLI_MASK */
	return ret;
}
EXPORT_SYMBOL(mn_intc_set_and_get_mask_priority);

#ifdef CONFIG_LINUX_CLI_MASK
#include <linux/hardirq.h>

unsigned int get_current_irq_priority(void)
{
	int cpu = hard_smp_processor_id();
	BUG_ON(!in_irq());
	return *(uniphier_intc_mask_stack[cpu] - 1) >> UNIPHIER_GIC_PRI_SHIFT;
}

void __mn_intc_pop_mask_priority(void)
{
	uniphier_intc_mask_stack[hard_smp_processor_id()]--;
}
#else /* CONFIG_LINUX_CLI_MASK */
void __mn_intc_pop_mask_priority(void)
{
	unsigned int cpu, *p;
	cpu = hard_smp_processor_id();

	p = uniphier_intc_mask_stack[cpu];
	__mn_intc_set_mask_priority(*(--p) >> UNIPHIER_GIC_PRI_SHIFT);
	uniphier_intc_mask_stack[cpu] = p;
}
#endif /* CONFIG_LINUX_CLI_MASK */

#ifdef CONFIG_LINUX_CLI_MASK
void *__fastintr_handler_table[CONFIG_LINUX_CLI_LEVEL];

#ifdef CONFIG_SMP
static DEFINE_SPINLOCK(fastintr_handler_lock);
#endif /* CONFIG_SMP */

static void default_fastintr_handler(unsigned int irq, struct pt_regs *pt_regs)
{
	mn_fastintr_clear(irq);
}

/* Must be called from the arch with interrupts disabled. */
static int __set_fastintr_handler(unsigned int level,
				  void (*handler)(unsigned int, struct pt_regs *))
{
	FI_CHECK_STATE(1, -1, -1);
	if (handler) {
		if (__fastintr_handler_table[level] != default_fastintr_handler) {
			return -EBUSY;
		}
	} else {
		handler = default_fastintr_handler;
	}
	__fastintr_handler_table[level] = handler;
	return 0;
}

int set_fastintr_handler(unsigned int level,
			 void (*handler)(unsigned int, struct pt_regs *))
{
	unsigned long flags;
	int ret;
	if ((level == FASTINTR_IPI_LEVEL) ||
	    (level >= CONFIG_LINUX_CLI_LEVEL)) {
		return -EINVAL;
	}
	local_fastintr_save(flags);
	fastintr_spin_lock(&fastintr_handler_lock);

	ret = __set_fastintr_handler(level, handler);

	fastintr_spin_unlock(&fastintr_handler_lock);
	local_fastintr_restore(flags);
	return ret;
}
EXPORT_SYMBOL(set_fastintr_handler);

#ifdef CONFIG_SMP

static void fastintr_smp_call_function_single_interrupt(void);

/******** based on arch/arm/mach-uniphier/include/mach/smp.h ********/

static inline void fastintr_smp_cross_call(const struct cpumask *mask)
{
	gic_raise_softirq(mask, FASTINTR_IPI_IRQ);
}

/******** based on arch/arm/kernel/smp.c ********/

struct fastintr_ipi_data {
	spinlock_t lock;
	unsigned long ipi_count;
	unsigned long bits;
};

static struct fastintr_ipi_data fastintr_ipi_data[NR_CPUS] = {
	[0 ... NR_CPUS - 1] = {
		.lock	= SPIN_LOCK_UNLOCKED,
	}
};

enum fastintr_ipi_msg_type {
	FASTINTR_IPI_CALL_FUNC_SINGLE,
};

static void fastintr_send_ipi_message(const struct cpumask *mask,
				      enum fastintr_ipi_msg_type msg)
{
	unsigned long flags;
	unsigned int cpu;

	local_fastintr_save(flags);

	/* for_each_cpu(cpu, mask) */
	for (cpu = -1;
	     cpu = find_next_bit(cpumask_bits(mask), NR_CPUS, cpu + 1),
	     cpu < NR_CPUS;) {
		struct fastintr_ipi_data *ipi = &fastintr_ipi_data[cpu];

		fastintr_spin_lock(&ipi->lock);
		ipi->bits |= 1 << msg;
		fastintr_spin_unlock(&ipi->lock);
	}

	/*
	 * Call the platform specific cross-CPU call function.
	 */
	fastintr_smp_cross_call(mask);

	local_fastintr_restore(flags);
}

static void fastintr_send_call_function_single_ipi(int cpu)
{
	fastintr_send_ipi_message(cpumask_of(cpu), FASTINTR_IPI_CALL_FUNC_SINGLE);
}

/* Must be called from GIC_CPU_PRIMASK is FASTINTR_IPI_LEVEL. */
static void fastintr_do_IPI(unsigned int irq, struct pt_regs *regs)
{
	unsigned int cpu = hard_smp_processor_id();
	struct fastintr_ipi_data *ipi = &fastintr_ipi_data[cpu];
	FI_CHECK_STATE(1, FASTINTR_IPI_LEVEL, FASTINTR_IPI_LEVEL);

	/* irq(first argument) is including cpu bits. */
	outl(irq, UNIPHIER_GIC_CPU_START + GIC_CPU_EOI);
	local_fastintr_enable();

	ipi->ipi_count++;

	for (;;) {
		unsigned long msgs;

		fastintr_spin_lock(&ipi->lock);
		msgs = ipi->bits;
		ipi->bits = 0;
		fastintr_spin_unlock(&ipi->lock);
		if (!msgs)
			break;

		do {
			unsigned nextmsg;

			nextmsg = msgs & -msgs;
			msgs &= ~nextmsg;
			nextmsg = ffz(~nextmsg);

			switch (nextmsg) {
			case FASTINTR_IPI_CALL_FUNC_SINGLE:
				fastintr_smp_call_function_single_interrupt();
				break;

			default:
				FI_BUG("Unknown fast IPI message");
				break;
			}
		} while (msgs);
	}
}

/******** based on include/linux/smp.h ********/

struct fastintr_call_single_data {
	struct list_head list;
	void (*func) (void *info);
	void *info;
	u16 flags;
	u16 priv;
};

/******** based on kernel/smp.c ********/

enum {
	FCSD_FLAG_LOCK		= 0x01,
};

struct fastintr_call_single_queue {
	struct list_head	list;
	spinlock_t		lock;
};

static struct fastintr_call_single_queue fastintr_call_single_queue[NR_CPUS];

/* Must not be called from the arch with interrupts disabled. */
static void fcsd_lock_wait(struct fastintr_call_single_data *data)
{
	FI_CHECK_STATE(0, CONFIG_LINUX_TASK_LEVEL, FASTINTR_IPI_LEVEL + 1);
	while (data->flags & FCSD_FLAG_LOCK)
		cpu_relax();
}

static void fcsd_lock(struct fastintr_call_single_data *data)
{
	fcsd_lock_wait(data);
	data->flags = FCSD_FLAG_LOCK;

	/*
	 * prevent CPU from reordering the above assignment
	 * to ->flags with any subsequent assignments to other
	 * fields of the specified call_single_data structure:
	 */
	smp_mb();
}

static void fcsd_unlock(struct fastintr_call_single_data *data)
{
	FI_ASSERT((data->flags & FCSD_FLAG_LOCK) == FCSD_FLAG_LOCK);
	/*
	 * ensure we're all done before releasing data:
	 */
	smp_mb();

	data->flags &= ~FCSD_FLAG_LOCK;
}

static void __fastintr_exec_single(int cpu,
				   struct fastintr_call_single_data *data,
				   int wait)
{
	struct fastintr_call_single_queue *dst =
		&fastintr_call_single_queue[cpu];
	unsigned long flags;
	int ipi;
	FI_CHECK_STATE(0, CONFIG_LINUX_TASK_LEVEL, FASTINTR_IPI_LEVEL + 1);

	FI_ASSERT((cpu >= 0) || (cpu < NR_CPUS));
	FI_ASSERT((data->flags & FCSD_FLAG_LOCK) == FCSD_FLAG_LOCK);

	local_fastintr_save(flags);
	fastintr_spin_lock(&dst->lock);
	ipi = list_empty(&dst->list);
	list_add_tail(&data->list, &dst->list);
	fastintr_spin_unlock(&dst->lock);

	if (ipi) {
		fastintr_send_call_function_single_ipi(cpu);
	}
	local_fastintr_restore(flags);

	if (wait) {
		fcsd_lock_wait(data);
	}
}

/* Must be called from GIC_CPU_PRIMASK is FASTINTR_IPI_LEVEL. */
static void fastintr_smp_call_function_single_interrupt(void)
{
	struct fastintr_call_single_queue *q =
		&fastintr_call_single_queue[hard_smp_processor_id()];
	LIST_HEAD(list);
	FI_CHECK_STATE(0, FASTINTR_IPI_LEVEL, FASTINTR_IPI_LEVEL);

	fastintr_spin_lock(&q->lock);
	list_replace_init(&q->list, &list);
	fastintr_spin_unlock(&q->lock);

	while (!list_empty(&list)) {
		struct fastintr_call_single_data *data;

		data = list_entry(list.next,
				  struct fastintr_call_single_data, list);
		list_del(&data->list);

		data->func(data->info);

		fcsd_unlock(data);
	}
}

static struct
fastintr_call_single_data fcsd_data[NR_CPUS][UNIPHIER_INTC_MAX_PREEMPT_COUNT];

int fastintr_smp_call_function_single(int cpu, void (*func) (void *info),
				      void *info, int wait)
{
	unsigned long flags, level;
	int this_cpu;
	int err = 0;
	struct fastintr_call_single_data *data;

	raw_local_save_flags(flags);
	level = get_flags_intc_pri(flags);
	if ((flags & PSR_I_BIT) || (level <= FASTINTR_IPI_LEVEL)) {
		return -EBUSY;
	}

	if (level == CONFIG_LINUX_TASK_LEVEL) {
		preempt_disable();
	}
	this_cpu = hard_smp_processor_id();

	if ((cpu == this_cpu) || (cpu >= NR_CPUS)) {
		err = -ENXIO;
		goto out;
	}

	data = &fcsd_data[this_cpu][level];
	fcsd_lock(data);

	data->func = func;
	data->info = info;
	__fastintr_exec_single(cpu, data, wait);
out:
	if (level == CONFIG_LINUX_TASK_LEVEL) {
		preempt_enable();
	}
	return err;
}
EXPORT_SYMBOL(fastintr_smp_call_function_single);
#endif /* CONFIG_SMP */

#ifdef CONFIG_SMP
void __cpuinit init_fastintr_call_function_ipi(void)
{
	FI_CHECK_STATE(1, -1, -1);
	mn_intc_set_level(FASTINTR_IPI_IRQ, FASTINTR_IPI_LEVEL);
}

void __init init_fastintr_call_single_data(void)
{
	int i;
	for (i = 0; i < NR_CPUS; i++) {
		struct fastintr_call_single_queue *q = &fastintr_call_single_queue[i];
		spin_lock_init(&q->lock);
		INIT_LIST_HEAD(&q->list);
	}
}
#endif /* CONFIG_SMP */

/* Must be called from the arch with interrupts disabled. */
void __init init_fastintr(void)
{
	int level;
	FI_CHECK_STATE(1, -1, -1);
	for (level = 0; level < CONFIG_LINUX_CLI_LEVEL; level++) {
		__fastintr_handler_table[level] = default_fastintr_handler;
	}
#ifdef CONFIG_SMP
	__set_fastintr_handler(FASTINTR_IPI_LEVEL, fastintr_do_IPI);

	init_fastintr_call_single_data();
#endif /* CONFIG_SMP */
}
#endif /* CONFIG_LINUX_CLI_MASK */
