/*
 * UniPhier shared locking registers.
 *
 * Copyright (C) 2012 Panasonic Corporation
 *
 */
#ifndef _UNIT_SSCLOCK_H
#define _UNIT_SSCLOCK_H

#ifdef __KERNEL__

#include <linux/irqflags.h>
#include <asm/io.h>

#define SSC_LOCK_REG_ADDR	UNIPHIER_VSSC_START0
#define SSC_LOCK_REG_SIZE	0x20	/* byte */

#define SSC_LOCK_REG_BIT(nr, b)		((nr) * 8 + (b))

#define SSC_LOCK_ABORT()				\
do {							\
	panic("BUG: ssc lock invalid argument.");	\
} while(0)

#define __ssc_memory_barrier()	__asm__ __volatile__("dmb": : : "memory")
#define __ssc_io(paddr)		__io(paddr)

static inline unsigned long __ssc_local_irqsave(void)
{
	unsigned long flags;
	local_irq_save(flags);
	return flags;
}

static inline void __ssc_local_irqrestore(unsigned long flags)
{
	local_irq_restore(flags);
        return;
}


/* Locking functions */
static inline void __ssc_lock(int bit)
{
	void *addr;
	unsigned int value;
	unsigned int tmp1, tmp2;

	if (bit < 0 || bit >= (SSC_LOCK_REG_SIZE * 8)) {
		SSC_LOCK_ABORT();
	}
	addr = __ssc_io(SSC_LOCK_REG_ADDR + bit / 8);
	value = 1 << (bit & 0x7);

	__asm__ __volatile__(
		"1:					\n"	/* retry:			*/
		"	ldrexb		%0, [%2]	\n"	/* tmp1 = *addr;		*/
		"	mov		%1, %0		\n"	/* tmp2 = tmp1;			*/
		"	ands		%1, %3		\n"	/* if (!(tmp2 &= value)) {	*/
		"	orreq		%0, %3		\n"	/*   tmp1 |= value;		*/
		"	strexbeq	%1, %0, [%2]	\n"	/*   *addr=tmp1; tmp2=state;	*/
		"	cmpeq		%1, #0		\n"	/*   if (tmp2) goto retry;	*/
		"	bne		1b		\n"	/* } else goto retry;		*/
		: "=&r" (tmp1), "=&r" (tmp2)
		: "r" (addr),"r" (value)
		: "memory", "cc");

	return;
}

static inline int __ssc_trylock(int bit)
{
	void *addr;
	unsigned int value;
	unsigned int tmp1, tmp2;

	if (bit < 0 || bit >= (SSC_LOCK_REG_SIZE * 8)) {
		SSC_LOCK_ABORT();
	}
	addr = __ssc_io(SSC_LOCK_REG_ADDR + bit / 8);
	value = 1 << (bit & 0x7);

	__asm__ __volatile__(
		"	ldrexb		%0, [%2]	\n"	/* tmp1 = *addr;		*/
		"	mov		%1, %0		\n"	/* tmp2 = tmp1;			*/
		"	ands		%1, %3		\n"	/* if (!(tmp2 &= value)) {	*/
		"	orreq		%0, %3		\n"	/*   tmp1 |= value;		*/
		"	strexbeq	%1, %0, [%2]	\n"	/*   *addr=tmp1; tmp2=state; }	*/
		: "=&r" (tmp1), "=&r" (tmp2)
		: "r" (addr),"r" (value)
		: "memory", "cc");
	if (tmp2 == 0) {
		return 1;
	} else {
		return 0;
	}
}

static inline void __ssc_unlock(int bit)
{
	void *addr;
	unsigned int value;
	unsigned int tmp1, tmp2;

	if (bit < 0 || bit >= (SSC_LOCK_REG_SIZE * 8)) {
		SSC_LOCK_ABORT();
	}
	addr = __ssc_io(SSC_LOCK_REG_ADDR + bit / 8);
	value = 1 << (bit & 0x7);

	__asm__ __volatile__(
		"1:					\n"	/* retry:			*/
		"	ldrexb		%0, [%2]	\n"	/* tmp1 = *addr;		*/
		"	and		%0, %3		\n"	/* tmp1 &= ~value;		*/
		"	strexb		%1, %0, [%2]	\n"	/* *addr=tmp1; tmp2=state;	*/
		"	cmp		%1, #0		\n"	/* if (tmp2) {			*/
		"	bne		1b		\n"	/*   goto retry; }		*/
		: "=&r" (tmp1), "=&r" (tmp2)
		: "r" (addr),"r" (~value)
		: "memory", "cc");
	return;
}

static inline unsigned long __ssc_lock_irqsave(int bit)
{
	unsigned long flags;

	flags = __ssc_local_irqsave();
	while (!__ssc_trylock(bit)) {
		__ssc_local_irqrestore(flags);
		flags = __ssc_local_irqsave();
	}
	return flags;
}

#define ssc_lock_irqsave(bit, flags)		\
do {						\
	(flags) = __ssc_lock_irqsave(bit);	\
} while (0)


static inline void ssc_unlock_irqrestore(int bit, unsigned long flags)
{
	__ssc_unlock(bit);
	__ssc_local_irqrestore(flags);
	return;
}

#endif /* __KERNEL__ */
#endif /* _UNIT_SSCLOCK_H */
