/**
 *	@file	kaoint.c
 *	@brief	kaoint driver
 */

//	----------------------------------------------------------------->>	INCLUDES

#include	<linux/module.h>
#include	<linux/fs.h>
#include	<linux/cdev.h>
#include	<linux/interrupt.h>
#include	<linux/sched.h>
#include	<linux/wait.h>
#include	<linux/poll.h>
#include	<linux/mm.h>

#include 	<asm/io.h>
#include	<asm/uaccess.h>

#include	"kaoint.h"			// kaoint header


//	----------------------------------------------------------------->>	MACROS

//	----------------------------------------------------------------->>	ENUMS

/**
 *	@brief		SET / RESET
 */
enum {
	KI_RESET	= 0	,
	KI_SET
};


//	----------------------------------------------------------------->>	STRUCTS

//	----------------------------------------------------------------->>	EXTERNAL FUNCS

//	----------------------------------------------------------------->>	EXTERNAL VARS

//	----------------------------------------------------------------->>	GLOBAL VARS

//	----------------------------------------------------------------->>	INLINE & MACRO FUNCS

static	wait_queue_head_t	kaoint_que;				// linux wait process que
static	int					kaoint_condition;		// kaoint condition
static	void __iomem		*kaoint_iovmap = NULL;	// kaoint virtual address


//	----------------------------------------------------------------->> FUNCS

/* open */
int		kaoint_open(	struct inode*	inode,
						struct file*	filp	);
/* read */
int		kaoint_read(	struct file*	filp,
						char*			buf,
						size_t			count,
						loff_t*			f_pos	);

/* write */
int		kaoint_write(	struct file*	filp,
						const char*		buf,
						size_t			count,
						loff_t*			f_pos	);

/* init */
int		kaoint_init( void );

/* exit */
void	kaoint_exit( void );


/* receive irq handler */
static	irqreturn_t	kaoint_recvIntHdr(	int		irq,
										void*	dev_id	);


/**
 *	@brief	kaoint open
 *
 *	@param	inode
 *	@param	file
 */
int		kaoint_open(	struct inode*	inode,
						struct file*	filp	)
{
	// module accses count check
	if( module_refcount(THIS_MODULE) >= DEV_ACCESS_MAX ) {
		// busy condition
		return	-EBUSY;
	}

	// try reference module( access nm + 1 )
	try_module_get(THIS_MODULE);

	printk( "[%s] open() : succeed\n", DEV_NAME );

	return	0;
}

/**
 *	@brief	receive irq handler
 *
 *	@param	irq
 *	@param	dev_id
 */
static	irqreturn_t	kaoint_recvIntHdr(	int		irq,
										void*	dev_id	)
{
	// non-open iomap?
	if( NULL == kaoint_iovmap ) {
		printk(	"[%s] handler() : fail(iovmap)\n", DEV_NAME );
		/* invalid irq */
		return IRQ_NONE;
	}

	/* clear interrupts */
	kaoint_irq_clear(	(unsigned long)kaoint_iovmap,
						INTNO_BE_TO_LINUX	);

	/* kaoint condition set */
	kaoint_condition	= KI_SET;

	/* wake up que tsk */
	wake_up_interruptible( &kaoint_que );

//	printk( "[%s] handler() : succeed\n", DEV_NAME);

	/* return irq handle */
	return	IRQ_HANDLED;
}

/**
 *	@brief		wait be -> net interrupt.
 *
 *	@param		file
 *	@param		buf
 *	@param		count
 *	@param		f_pos
 */
int		kaoint_read(	struct file*	filp,
						char*			buf,
						size_t			count,
						loff_t*			f_pos	)
{
	int	ret;

//	module_put(THIS_MODULE);

	// non-open iomap?
	if( NULL == kaoint_iovmap ) {
		printk( "[%s] read() : fail(iovmap)\n", DEV_NAME );
		/* invalid */
		return -EBUSY;
	}

	/* wait be -> net interrupt */
	ret	= wait_event_interruptible(	kaoint_que,
									kaoint_condition );
	/* error ? */
	if(	ret != 0) {
		/* return invalid */
		printk( "[%s] read() : fail(waitevent)\n", DEV_NAME );
		return	ret;
	}

	/* reset kaoint condition */
	kaoint_condition	= KI_RESET;

//	printk( "[%s] read() : succeed\n", DEV_NAME );

	return	0;
}

/**
 *	@brief		set net -> be interrupt.
 *
 *	@param		file
 *	@param		buf
 *	@param		count
 *	@param		f_pos
 */
int		kaoint_write(	struct file*	filp,
						const char*		buf,
						size_t			count,
						loff_t*			f_pos	)
{
	// non-open iomap?
	if( NULL == kaoint_iovmap ) {
		printk( "[%s] write() : fail\n", DEV_NAME );
		/* invalid */
		return -EBUSY;
	}

	/* set net -> be interrupt */
	kaoint_set_irq(	(unsigned long)kaoint_iovmap,
					INTNO_LINUX_TO_BE	);

//	printk( "[%s] write() : succeed\n", DEV_NAME);

	return	0;
}

/**
 *	@brief drivr file operations
 */
struct file_operations	kaoint_fops = {

	.open		= kaoint_open,
	.read		= kaoint_read,
	.write		= kaoint_write
};

/**
 *	@brief		kaoint driver init function
 *
 *	@param		nothing
 */
int kaoint_init( void )
{
	int	ret;

#if 0	// for debug

	/* tile message display */
	printk(" *** [%s] opened kaoint driver!! ***\n", __func__);
	printk("\n");

	printk("KAOINT_CPU_NO        : 0x%X\n", (unsigned int)KAOINT_CPU_NO );
	printk("KAOINT_BE_CPU_NO     : 0x%X\n", (unsigned int)KAOINT_BE_CPU_NO );
	printk("INTNO_BE_TO_LINUX    : 0x%X\n", (unsigned int)INTNO_BE_TO_LINUX );
	printk("INTNO_LINUX_TO_BE    : 0x%X\n", (unsigned int)INTNO_LINUX_TO_BE );
	printk("\n");

	printk("MAP_GIC_REG_BASE         : 0x%X\n", (unsigned int)MAP_GIC_REG_BASE );
	printk("MAP_GIC_REG_SZ           : 0x%X(page:0x%X)\n", (unsigned int)MAP_GIC_REG_SZ, (unsigned int)PAGE_SIZE );
	printk("MAP_GIC_OFFSET_PERCPU    : 0x%X\n", (unsigned int)MAP_GIC_OFFSET_PERCPU );
	printk("\n");

	printk("MAP_GIC_OFFSET_DIST_BASE : 0x%X\n", (unsigned int)MAP_GIC_OFFSET_DIST_BASE );
	printk("MAP_OFFSET_GIC_DIST_ISER : 0x%X\n", (unsigned int)MAP_OFFSET_GIC_DIST_ISER );
	printk("MAP_OFFSET_GIC_DIST_ICPR : 0x%X\n", (unsigned int)MAP_OFFSET_GIC_DIST_ICPR );
	printk("MAP_OFFSET_GIC_DIST_ABR  : 0x%X\n", (unsigned int)MAP_OFFSET_GIC_DIST_ABR );
	printk("MAP_OFFSET_GIC_DIST_SGIR : 0x%X\n", (unsigned int)MAP_OFFSET_GIC_DIST_SGIR );
	printk("\n");

	printk("MAP_GIC_OFFSET_CPU_IF_BASE : 0x%X\n", (unsigned int)MAP_GIC_OFFSET_CPU_IF_BASE );
	printk("MAP_GIC_OFFSET_CPU_IF_EOI : 0x%X\n", (unsigned int)MAP_GIC_OFFSET_CPU_IF_EOI );
	printk("\n");

#endif

	/* check state of other using. */
	ret = register_chrdev(	DEV_MAJOR_NUM,
							DEV_NAME,
							&kaoint_fops	);
	if( ret < 0 ) {
		/* return failed */
		printk( "[%s] register_chrdev() : fail\n", DEV_NAME );
		return	ret;
	}
//	printk( "[%s] register_chrdev() : succeed\n", DEV_NAME );

	/* get virtual address. */
	kaoint_iovmap		= ioremap_nocache(	MAP_GIC_REG_BASE,
											MAP_GIC_REG_SZ	);

	/* initialize condition */
	kaoint_condition	= KI_RESET;

	/* initiallize que */
	init_waitqueue_head( &kaoint_que );


	/* register interrupt handler entry */
	ret = request_irq(	INTNO_BE_TO_LINUX,		// irq no
						kaoint_recvIntHdr,		// irq handler
						IRQF_DISABLED,			// other irq disabled
						DEV_NAME,				// handler name
						NULL	);
	if( 0 != ret ){

		/* return failed */
		printk(	"[%s] request_irq() fail(%d)\n", DEV_NAME, ret);
		unregister_chrdev( DEV_MAJOR_NUM, DEV_NAME );

		/* failed */
		return	-EBUSY;
	}

	/* return succeed */
//	printk( "request_irq() : succeed : 0x%08X\n", (u32)kaoint_iovmap );
	printk( "[%s] init succeed\n", DEV_NAME );
	return	0;
}

/**
 *	@brief
 */
void kaoint_exit( void )
{
	/* virtual address acquisition? */
	if( NULL != kaoint_iovmap ) {
		/* free vmap memory */
		iounmap((void*)kaoint_iovmap);
		kaoint_iovmap = NULL;
	}

	/* release state of using. */
	unregister_chrdev(	DEV_MAJOR_NUM,
						DEV_NAME);

	printk(	"[%s] exit succeed\n", DEV_NAME );

	return;
}

// ================================================================
// 	kaoint driver module description
// ================================================================

/* module init */
module_init(kaoint_init);

/* module exit */
module_exit(kaoint_exit);

/* module description */
MODULE_DESCRIPTION("kaoint driver");

/* module lisense */
MODULE_LICENSE("Proprietary");

/* module author */
MODULE_AUTHOR("Panasonic");


// ================================================================


/* end */

