/**
 *	@file	kaoint.h
 *	@brief	kaoint driver header
 */

#ifndef	_KAO_INT_H_
#define	_KAO_INT_H_

//	----------------------------------------------------------------->>	INCLUDES

#include	<linux/spinlock.h>


//	----------------------------------------------------------------->>	MACROS

// ------------------------------------------------
// 				Driver Config
// ------------------------------------------------

// driver name
#define	DEV_NAME				"kaoint"

// device major number
#define	DEV_MAJOR_NUM			120

// max driver access number
#define	DEV_ACCESS_MAX			1

// kaoint on cpu no
#define KAOINT_CPU_NO			2

// BE cpu no
#define KAOINT_BE_CPU_NO		0


// ------------------------------------------------
// 		kaoint interrupt no.
// ------------------------------------------------

//	BE	-> NET	: 6
//	NET -> BE	: 7
//
//	[ARMINTC]	[Interrupt ID]	[unit]		[sub]
//	SGI			6				MIU(CPU)	GIC
//	SGI			7				MIU(CPU)	GIC
//
//	[document]
//	.../LSI外部仕様書\BLK\EXT\MIU\V9AHD_割込み要因.xls
//	.../BE_Soft/Product/Common/V9AHD/mk/SystemParam.h

// BE -> NET Interrupt No.
#define INTNO_BE_TO_LINUX		6

// NET -> BE Interrupt No.
#define INTNO_LINUX_TO_BE		7


// ------------------------------------------------
// 		GIC register map
// ------------------------------------------------

// register base address
#define MAP_GIC_REG_BASE			0x70000000

// register accsess total size
#define MAP_GIC_REG_SZ				0x00020000

// on CPU offset( cpu unit : 0x8000 )
#define MAP_GIC_OFFSET_PERCPU		( 0x8000 * KAOINT_CPU_NO )


// Distributor offset
#define MAP_GIC_OFFSET_DIST_BASE	( MAP_GIC_OFFSET_PERCPU + 0x1000 )

// Interrupt Set Enable Registers
#define MAP_OFFSET_GIC_DIST_ISER	( MAP_GIC_OFFSET_DIST_BASE + 0x100 )

// Interrupt Clear Pending Registers
#define MAP_OFFSET_GIC_DIST_ICPR	( MAP_GIC_OFFSET_DIST_BASE + 0x280 )

// Interrupt Set Active Registers
#define MAP_OFFSET_GIC_DIST_ABR		( MAP_GIC_OFFSET_DIST_BASE + 0x300 )

// Software Generated Interrupt Register
#define MAP_OFFSET_GIC_DIST_SGIR	( MAP_GIC_OFFSET_DIST_BASE + 0xF00 )

// Clear Pending Registers
#define MAP_OFFSET_GIC_DIST_CPENDSGIR	( MAP_GIC_OFFSET_DIST_BASE + 0xF10 )


// CPU interfaces offset
#define MAP_GIC_OFFSET_CPU_IF_BASE	( MAP_GIC_OFFSET_PERCPU + 0x2000 )

// End of Interrupt Register
#define MAP_GIC_OFFSET_CPU_IF_EOI	( MAP_GIC_OFFSET_CPU_IF_BASE + 0x10 )


//	----------------------------------------------------------------->>	ENUMS

//	----------------------------------------------------------------->>	STRUCTS

//	----------------------------------------------------------------->>	EXTERNAL FUNCS

//	----------------------------------------------------------------->>	EXTERNAL VARS

//	----------------------------------------------------------------->>	GLOBAL VARS

//	----------------------------------------------------------------->>	INLINE & MACRO FUNCS

/**
 *	@brief	kaoint memory barrier
 *
 *	@param	nothing
 */
#if defined(_PGR)  /* dummy for PGRelief */
#define memory_barrier()
#else /* _PGR */
#define memory_barrier()	__asm__ __volatile__("dmb": : : "memory")
#endif /* _PGR */

/**
 *	@brief	kaoint irq save(arm status core register save)
 *
 *	@param	nothing
 */
#if defined(_PGR)  /* dummy for PGRelief */
static inline unsigned long kaoint_irqsave(void) { return 0; }
#else /* _PGR */
static inline unsigned long kaoint_irqsave(void)
{
	unsigned long flags;
	__asm__ __volatile__(
		"	mrs	%0, cpsr\n"
		"	cpsid	i	\n"
		: "=r"(flags)
		:
		: "memory", "cc"
		);

	memory_barrier();
	return flags;
}
#endif /* _PGR */


/**
 *	@brief	kaoint irq save(arm status core register restore)
 *
 *	@param	flags
 */
#if defined(_PGR)  /* dummy for PGRelief */
static inline int kaoint_irqrestore(unsigned long flags) { return 0; }
#else /* _PGR */
static inline int kaoint_irqrestore(unsigned long flags)
{
	memory_barrier();

	__asm__ __volatile__(
		"	msr	cpsr_c, %0\n"
		:
		: "r" (flags)
		: "memory", "cc"
		);

	return 0;
}
#endif /* _PGR */

/**
 *	@brief	kaoint irq clear
 *
 *	@param	vadr	virtual base address
 *	@param	irq		interrupt no
 */
static inline void kaoint_irq_clear( unsigned long vadr, int irq )
{
	unsigned long gic_dist_cpendsgir = vadr + MAP_OFFSET_GIC_DIST_CPENDSGIR	+ irq;

	/* SGI Pending clear */
	*(volatile unsigned char*)gic_dist_cpendsgir = 0xFF;
}

/**
 *	@brief	kaoint set irqs(other cpu)
 *
 *	@param	vadr	virtual base address
 *	@param	irq		interrupt no
 */
static inline int kaoint_set_irq( unsigned long vadr, int irq )
{
	spinlock_t		lock;	// spinlock
	unsigned long	flgs;	// spinlock flgs

	unsigned long map = 1 << KAOINT_BE_CPU_NO;
	unsigned long gic_dist_sgir = vadr + MAP_OFFSET_GIC_DIST_SGIR;

	/* lock spinlock */
	spin_lock_irqsave( &lock, flgs );

	// set SGIR( software generated interrupt register )
	*(volatile unsigned long *)gic_dist_sgir = ( irq | map << 16 );

	/* unlock spinlock */
	spin_unlock_irqrestore( &lock, flgs );

	return 0;
}

#endif	// _KAO_INT_H_
