/**
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 * @file
 * @brief   CPU間通信D/D カーネルAPI
 */
#include <asm/io.h>         /* ioremap, iounmap */
#include <linux/errno.h>
#include <linux/fs.h>       /* inode, file, file_operations */
#include <linux/kernel.h>   /* printk() */
#include <linux/module.h>
#include <linux/ioctl.h>    /* ioctl */
#include <linux/poll.h>     /* POLLIN ... */
#include <linux/wait.h>     /* wait_queue_head_t */
#include <linux/sched.h>    /* wait_event_interruptible, wake_up_interruptible */

#include "ppcdev.h"
#include "ppcdev_main.h"
#include "ppcdev_util.h"
#include "ppcdev_ioctl.h"


// 他カーネルモジュールからの参照
EXPORT_SYMBOL( ppcdev_kopen );
EXPORT_SYMBOL( ppcdev_kclose );
EXPORT_SYMBOL( ppcdev_kmalloc );
EXPORT_SYMBOL( ppcdev_kfree );
EXPORT_SYMBOL( ppcdev_ksend );
EXPORT_SYMBOL( ppcdev_kreqirq );
EXPORT_SYMBOL( ppcdev_kioctl );


/**
 * @brief  デバイスオープン[Kernel API]
 * @param  [in]  id         受信ID
 * @return 受信キューインデックス
 */
int ppcdev_kopen( int id )
{
    switch ( id ) {
    case PPCDEV_SVETH_ID:
        return __ppcdev_open( _ETH, id );
    case PPCDEV_SDIO_ID:
        return __ppcdev_open( _DMA, id );
    default:
        return -EINVAL;
    }
}


/**
 * @brief  デバイスクローズ[Kernel API]
 * @param  [in]  fd         受信キューインデックス
 * @return エラーコード
 */
int ppcdev_kclose( int fd )
{
    return __ppcdev_release( fd );
}


/**
 * @brief  DRAM領域確保[Kernel API]
 * @param  [in]  size       メモリブロックサイズ
 * @return カーネル仮想アドレス
 */
void *ppcdev_kmalloc( size_t size )
{
    return __ppcdev_kmalloc( size );
}


/**
 * @brief  DRAM領域開放[Kernel API]
 * @param  [in]  buf        物理アドレス
 */
void ppcdev_kfree( const void *buf )
{
    __ppcdev_kfree( buf );
}


/**
 * @brief  送信バッファライト[Kernel API]
 * @param  [in]  fd         受信キューインデックス
 * @param  [in]  id         送信ID
 * @param  [in]  buf        DRAM送信バッファアドレス
 * @param  [in]  size       ライトバイトサイズ
 * @return ライトバイトサイズ
 */
int ppcdev_ksend( int fd, int id, const void *buf, size_t size )
{
    ppcdev_msgh_t txmsg;
    int copy_len;

    txmsg.id = ( unsigned short )id;
    txmsg.size = ( unsigned short )size;
    // アドレス変換(論理->物理)
    txmsg.txdadr = ( unsigned long )PPCDEV_TXDRAM_BASE + ( ( unsigned long )buf - ( unsigned long )txdram_ioaddr );

    copy_len = ( int )__ppcdev_write( fd, ( const char * )&txmsg, PPCDEV_TXMSGW, PPCDEV_KAPI_SYNC );
    return copy_len;
}


/**
 * @brief  コールバック関数登録[Kernel API]
 * @param  [in]  fd         受信キューインデックス
 * @param  [in]  callback   コールバック関数
 */
void ppcdev_kreqirq( int fd, ppcdev_irq_handler_t callback )
{
    switch ( fd ) {
    case _ETH:
    case _DMA:
        ppcdev_cb[ fd ] = callback;
        break;
    default:
        break;
    }
}


/**
 * @brief  デバッグ情報取得/デバイス制御[Kernel API]
 * @param  [in]  fd         受信キューインデックス
 * @param  [in]  cmd        コマンド
 * @param  [in]  arg        引数、戻り値ポインタ
 * @return エラーコード
 */
int ppcdev_kioctl( int fd, unsigned int cmd, unsigned long arg )
{
    return 0;
}

/* End of ppcdev_kapi.c */
