/**
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 * @file
 * @brief   CPU間通信D/Dメイン処理
 */
#include <asm/uaccess.h>    /* copy_from_user, copy_to_user */
#include <asm/io.h>         /* ioremap, iounmap */
#include <asm/irq.h>        /* request_irq, free_irq */
#include <linux/slab.h>     /* kmalloc, kfree */
#include <linux/errno.h>
#include <linux/fs.h>       /* inode, file, file_operations */
#include <linux/kernel.h>   /* printk() */
#include <linux/module.h>
#include <linux/ioctl.h>    /* ioctl */
#include <linux/poll.h>     /* POLLIN ... */
#include <linux/wait.h>     /* wait_queue_head_t */
#include <linux/sched.h>    /* wait_event_interruptible, wake_up_interruptible */
#include <linux/timer.h>    /* init_timer, add_timer, del_timer_sync */
#include <linux/spinlock.h>
#include <linux/mutex.h>
#include <linux/kthread.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>

#include "ppcirq.h"
#include "ppcdev_def.h"
#include "ppcdev_util.h"


// 外部変数
extern int   devmajor;
extern char *devname;


// グローバル変数
void __iomem *srbase_ioaddr;             // SRAMベースアドレス
void __iomem *txsram_ioaddr;             // 送信SRAMアドレス
void __iomem *rxsram_ioaddr;             // 受信SRAMアドレス
void __iomem *drbase_ioaddr;             // DRAMベースアドレス
void __iomem *txdram_ioaddr;             // 送信DRAMアドレス
void __iomem *rxdram_ioaddr;             // 受信DRAMアドレス
int ppcirq_fd;                           // PPC割り込みD/Dディスクリプタ
ppcdev_work_t ppcdev_work[ _MAXQ ];      // ワークキュー配列
ppcdev_irq_handler_t ppcdev_cb[ _MAXQ ]; // コールバック配列
spinlock_t ppcdev_qptr_spinlock;         // 送受信キュー構造体配列ロック
struct task_struct *ppcdev_kth;          // カーネルスレッド
int ppcdev_retry;                        // 送信リトライ回数 デフォルト 1回
DEFINE_MUTEX( ppcdev_tx_mutex );         // SMC/Ether送信ミューテックス
DEFINE_MUTEX( ppcdev_kmalloc_mutex );    // kmallocミューテックス
DEFINE_MUTEX( ppcdev_kfree_mutex );      // kfreeミューテックス
DEFINE_MUTEX( ppcdev_retry_mutex );      // リトライミューテックス
static wait_queue_head_t ppcdev_irack_q; // 割り込み応答ウェイト
static int ppcdev_irack_flag;            // 割り込み応答フラグ
DEFINE_MUTEX( ppcdev_irack_mutex );      // 割り込み応答フラグミューテックス

// プロトタイプ宣言
int __ppcdev_open( int qid, int id );
int __ppcdev_release( int qid );
ssize_t __ppcdev_read( int qid, char *buf, size_t count, int tmout );
ssize_t __ppcdev_write( int qid, const char *buf, size_t count, int tmout );
void *__ppcdev_kmalloc( size_t size );
void __ppcdev_kfree( const void *buf );
static int __ppcdev_tx_msg( char *buf );
static int ppcdev_txq2sram( void *arg );
static irqreturn_t ppcdev_sram2rxq( int irq, void *dev_id );
static void ppcdev_work_handler( struct work_struct *work );
static void ppcdev_irack_cb( void *arg );


/**
 * @brief  デバイスオープン処理<br>
 *         送受信キューの作成
 * @param  [in] qid  キュー番号
 * @param  [in] id   送信ID
 * @return キュー番号      成功<br>
 *         エラー          失敗
 */
int __ppcdev_open( int qid, int id )
{
    if ( qid >= _MAXQ )
        return -EINVAL;

    spin_lock( &ppcdev_qptr_spinlock );

    if ( ppcdev_que[ qid ] != NULL )
        return -EBUSY;

    // 受信キュー作成
    ppcdev_que[ qid ] = ( ppcdev_que_t * ) kmalloc( sizeof( ppcdev_que_t ), GFP_KERNEL );
    init_que( qid, id );
    INIT_WORK( &ppcdev_work[ qid ].work, ppcdev_work_handler );

    // 送信キュー作成(初回のみ)
    if ( ppcdev_que[ _TX ] == NULL ) {
        ppcdev_que[ _TX ] = ( ppcdev_que_t * ) kmalloc( sizeof( ppcdev_que_t ), GFP_KERNEL );
        init_que( _TX, 0 );
        iowrite8( ( unsigned char )_TRUE, ( txsram_ioaddr + 12 ) ); // SRAM通信許可

        // 割り込みドライバ開始
        ppcirq_fd = ppcirq_kopen( PPCIRQ_PIL_NWA, PPC_IRC1DINT0_MASK, PPC_IRC1DINT0_OFS, PPC_IRC2DINT1_OFS );
        ppcirq_ksetcb( ppcirq_fd, ppcdev_sram2rxq , ppcdev_irack_cb );
        ppcirq_kenable( ppcirq_fd );

        init_waitqueue_head( &ppcdev_irack_q );
        ppcdev_kth = kthread_run( ppcdev_txq2sram, NULL, "ppcdev_txq2sram" );
    }

    spin_unlock( &ppcdev_qptr_spinlock );

    return qid;
}


/**
 * @brief  デバイスクローズ処理<br>
 *         送受信キューの削除
 * @param  [in] qid  キュー番号
 * @return 0               成功<br>
 *         エラー          失敗
 */
int __ppcdev_release( int qid )
{
    int i;
    int last = _TRUE;

    if ( qid >= _MAXQ )
        return -EINVAL;

    spin_lock( &ppcdev_qptr_spinlock );

    if ( ppcdev_que[ qid ] == NULL )
        return -EFAULT;

    // 最終release判定
    for ( i = _NWA; i < _MAXQ; i++ ) {
        if ( i == qid )
            continue;
        else
            last &= ( ppcdev_que[ i ] == NULL );
    }

    // 最終release処理
    if ( last == _TRUE ) {
        // SRAM通信禁止
        iowrite8( ( unsigned char )_FALSE, ( txsram_ioaddr + 12 ) );

        // 割り込みドライバ停止
        ppcirq_kdisable( ppcirq_fd );
        ppcirq_kdelcb( ppcirq_fd );
        ppcirq_kclose( ppcirq_fd );

        // 送信キュー削除 kthreadを停止
        kthread_stop( ppcdev_kth );
        kfree( ppcdev_que[ _TX ] );
        ppcdev_que[ _TX ] = NULL;
    }

    // 受信キュー削除
    kfree( ppcdev_que[ qid ] );
    ppcdev_que[ qid ] = NULL;

    spin_unlock( &ppcdev_qptr_spinlock );

    return 0;
}


/**
 * @brief  受信キューPOP処理
 * @param  [in]  qid       キュー番号
 * @param  [out] buf       リードデータ
 * @param  [in]  count     リードサイズ
 * @param  [in]  tmout     タイムアウト指定
 * @return リードバイトサイズ    成功<br>
 *         エラー                失敗
 */
ssize_t __ppcdev_read( int qid, char* buf, size_t count, int tmout )
{
    int tmouterr;
    int copy_len;
    int grp;
    int gwp;
    int usage;
    int i;
    int rp;

    // 受信キューにデータが無ければタイムアウトまで待つ
    while ( is_empty( qid ) ) {
        // ノンブロッキング
        if ( tmout == 0 ) {
            return -EAGAIN;
        }

        // スリープ
        if ( tmout > 0 ) {
            // タイムアウト
            tmouterr = wait_event_interruptible_timeout( ppcdev_que[ qid ]->read_q, !is_empty( qid ), tmout );
            if ( tmouterr < 0 )
                return -ERESTARTSYS;
            else if ( tmouterr == 0 )
                return -EAGAIN;
        } else {
            // ブロッキング
            if ( wait_event_interruptible( ppcdev_que[ qid ]->read_q, !is_empty( qid ) ) )
                return -ERESTARTSYS;
        }
    }

    // キューの使用量を算出
    grp = get_rp( qid );
    gwp = get_wp( qid );

    if ( is_empty( qid ) )
        usage = 0;
    else if ( is_full( qid ) )
        usage = PPCDEV_MAXBUF;
    else if( grp < gwp )
        usage = gwp - grp;
    else
        usage = gwp + PPCDEV_MAXBUF - grp;

    // コピーサイズがキューのサイズを超えないようにクリップ
    if ( count > usage )
        copy_len = usage;
    else
        copy_len = count;

    // 受信メッセージをキューから取り出す
    for ( i = 0; i < copy_len; i += PPCDEV_TXMSGW ) {
        rp = grp + i;
        if ( rp >= PPCDEV_MAXBUF )
            rp = 0;

        if ( qid == _ETH ) { // 受信メッセージをカーネル空間にコピー
            memcpy( buf, &ppcdev_que[ qid ]->msgq[ rp ], PPCDEV_TXMSGW );
        }
        else { // 受信メッセージをユーザー空間にコピー
            if ( copy_to_user( buf, &ppcdev_que[ qid ]->msgq[ rp ], PPCDEV_TXMSGW ) ) {
                printk( KERN_DEBUG "%s : (%s) copy_to_user failed\n", devname, __func__ );
                return -EFAULT;
            }
        }
    }

    // ポインタを更新
    set_rp( qid, PPCDEV_TXMSGW );

    return copy_len;
}

/**
 * @brief  送信キューPUSH処理
 * @param  [in]  qid       キュー番号
 * @param  [in]  buf       ライトデータ
 * @param  [in]  count     ライトサイズ
 * @param  [in]  tmout     タイムアウト指定
 * @return ライトバイトサイズ    成功<br>
 *         エラー                失敗
 */
ssize_t __ppcdev_write( int qid, const char* buf, size_t count, int tmout )
{
    int tmouterr;
    int copy_len;
    int grp;
    int gwp;
    int blank;
    int i;
    int wp;

    // ミューテックスの獲得 失敗ならエラーを返して終了
    if ( mutex_lock_interruptible( &ppcdev_tx_mutex ) ) {
        printk( KERN_DEBUG "%s : (%s) mutex_lock_interruptible failed 1\n", devname, __func__ );
        return -ERESTARTSYS;
    }

    // 送信キューに空きが無ければタイムアウトまで待つ
    while ( is_full( _TX ) ) {
        mutex_unlock( &ppcdev_tx_mutex );

        // ノンブロッキング
        if ( tmout == 0 ) {
            return -EAGAIN;
        }

        // スリープ
        if ( tmout > 0 ) {
            // タイムアウト
            tmouterr = wait_event_interruptible_timeout( ppcdev_que[ _TX ]->write_q, !is_full( _TX ), tmout );
            if ( tmouterr < 0 )
                return -ERESTARTSYS;
            else if ( tmouterr == 0 )
                return -EAGAIN;
        } else {
            // ブロッキング
            if ( wait_event_interruptible( ppcdev_que[ _TX ]->write_q, !is_full( _TX ) ) )
                return -ERESTARTSYS;
        }

        if ( mutex_lock_interruptible( &ppcdev_tx_mutex ) ) {
            printk( KERN_DEBUG "%s : (%s) mutex_lock_interruptible for write failed 2\n", devname, __func__ );
            return -ERESTARTSYS;
        }
    }

    // キューの空き容量を算出
    gwp = get_wp( _TX );
    grp = get_rp( _TX );

    if ( is_full( _TX ) )
        blank = 0;
    else if ( is_empty( _TX ) )
        blank = PPCDEV_MAXBUF;
    else if( gwp < grp )
        blank = grp - gwp;
    else
        blank = grp + PPCDEV_MAXBUF - gwp;

    // コピーサイズがキューの空き容量を超えないようにクリップ
    if ( count > blank )
        copy_len = blank;
    else
        copy_len = count;

    // 送信メッセージをキューにコピー
    for ( i = 0; i < copy_len; i += PPCDEV_TXMSGW ) {
        wp = gwp + i;
        if ( wp >= PPCDEV_MAXBUF )
            wp = 0;

        if ( qid == _NWA ) { // 送信メッセージをユーザー空間からコピー
            if ( copy_from_user( &ppcdev_que[ _TX ]->msgq[ wp ], buf, PPCDEV_TXMSGW ) ) {
                mutex_unlock( &ppcdev_tx_mutex );
                printk( KERN_DEBUG "%s : (%s) copy_from_user failed\n", devname, __func__ );
                return -EFAULT;
            }
        }
        else if ( qid == _ETH || qid == _DMA ) { // 送信メッセージをカーネル空間からコピー
            memcpy( &ppcdev_que[ _TX ]->msgq[ wp ], buf, PPCDEV_TXMSGW );
        }
        else {
            return -EINVAL;
        }
    }

#if defined(CONFIG_PPCDEV_DIRECT_SEND)
    // キューが空であれば、kthreadをバイパスして送信処理
    if ( blank == PPCDEV_MAXBUF && copy_len == PPCDEV_TXMSGW ) {
        if ( __ppcdev_tx_msg( &ppcdev_que[ _TX ]->msgq[ gwp ] ) == _SUCCESS ) {
            mutex_unlock( &ppcdev_tx_mutex );
            return copy_len;
        }
    }
#endif	/* CONFIG_PPCDEV_DIRECT_SEND */

    // ポインタを更新
    set_wp( _TX, copy_len );

    // ミューテックスを開放
    mutex_unlock( &ppcdev_tx_mutex );

    // 送信キューのリード待ちを解除
    wake_up_interruptible( &ppcdev_que[ _TX ]->read_q );

    return copy_len;
}


/**
 * @brief  DRAM領域確保
 * @param  [in]  size       メモリブロックサイズ
 * @return カーネル仮想アドレス
 */
void *__ppcdev_kmalloc( size_t size )
{
    int i;
    volatile ppcdev_memh_t *memh;
    void *txmem = NULL;

    if ( mutex_lock_interruptible( &ppcdev_kmalloc_mutex ) ) {
        printk( KERN_DEBUG "%s : (%s) mutex_lock_interruptible failed\n", devname, __func__ );
        return NULL;
    }

    // 共有DRAMから空きを検索
    // リモートでは後半の8個をユーザ空間で使用するため
    // ppcdevではインデックス0~7を使用する
    for ( i = 0; i < PPCDEV_MEMSET / 2; i++ ) {
        memh = (void*)((unsigned long)txdram_ioaddr + PPCDEV_MEMBLK * i);

        if ( memh->en ) { // 使用中であれば次の領域へ
            continue;
        }
        else { // 未使用であれば、使用中にして検索終了
            memh->en = 1;
            memh->size = ( unsigned long )size;
            txmem = (void*)((unsigned long)memh + sizeof( ppcdev_memh_t ));
            break;
        }
    }

    mutex_unlock( &ppcdev_kmalloc_mutex );
    return txmem;
}


/**
 * @brief  DRAM領域開放
 * @param  [in]  buf        物理アドレス
 */
void __ppcdev_kfree( const void *buf )
{
    unsigned long head_en;
    unsigned long page_addr;
    unsigned long page_offset;

    if ( mutex_lock_interruptible( &ppcdev_kfree_mutex ) ) {
        printk( KERN_DEBUG "%s : (%s) mutex_lock_interruptible failed\n", devname, __func__ );
    }
    else {
        // メモリヘッダー Enableアドレス
        head_en = ( unsigned long )buf - sizeof( ppcdev_memh_t ) + sizeof( long );

        (void)page_addr;	/* unused */
        page_offset = head_en - (unsigned long)PPCDEV_RXDRAM_BASE;
        iowrite32( _FALSE, rxdram_ioaddr + page_offset );

        mutex_unlock( &ppcdev_kfree_mutex );
    }
}


/**
 * @brief  メッセージ送信
 * @param  [in]  buf        メッセージポインタ
 * @return 送信結果
 */
static int __ppcdev_tx_msg( char *buf )
{
    memcpy_toio( txsram_ioaddr, buf, PPCDEV_TXMSGW );

    // 割り込み応答待ちフラグをクリア
    mutex_lock( &ppcdev_irack_mutex );
    ppcdev_irack_flag = 0;
    mutex_unlock( &ppcdev_irack_mutex );

    // 割り込み送信
    ppcirq_ksend( ppcirq_fd );
    // 割り込み応答待ち
    if ( !ppcdev_irack_flag )
        wait_event_interruptible( ppcdev_irack_q, ppcdev_irack_flag );

    // 送信結果をリード
    return ioread32( rxsram_ioaddr + 8 );
}


/**
 * @brief  送信キューデータSRAM転送[Kernel thread]
 * @param  [in]  arg        未使用
 * @return 0
 */
static int ppcdev_txq2sram( void *arg )
{
    int rp;
    volatile int txresult;
    volatile int retry_th;
    volatile int retry_num;

    retry_num = 0;

    while ( !kthread_should_stop() ) {
        // リトライ回数を取得
        if ( mutex_lock_interruptible( &ppcdev_retry_mutex ) )
            continue;
        retry_th = ppcdev_retry;
        mutex_unlock( &ppcdev_retry_mutex );

        if ( is_empty( _TX ) ) {
            // スリープ
            wait_event_interruptible( ppcdev_que[ _TX ]->read_q, ( !is_empty( _TX ) || kthread_should_stop() ) );
        }
        else {
            // SRAMアドレスにキューの内容を転送
            rp = get_rp( _TX );
            txresult = __ppcdev_tx_msg( &ppcdev_que[ _TX ]->msgq[ rp ] );

            /* ********************************** */
            /* 受信エラー時の処理を入れる必要あり */
            /* ********************************** */
            // 成功->リードポインタ更新
            // リトライ->リードポインタ更新を停止
            // その他->カーネルパニック??

            // リードポインタを更新
            if ( txresult == _SUCCESS ) {
                set_rp( _TX, PPCDEV_TXMSGW );
                retry_num = 0;
            }
            else if ( txresult == _RETRY && retry_num < retry_th ) {
                retry_num ++;
            }
            else {
                set_rp( _TX, PPCDEV_TXMSGW );
                retry_num = 0;
            }
        }
    }

    return 0;
}


/**
 * @brief  SRAMデータ受信キュー転送[割り込みハンドラー実体]
 * @param  [in]  irq        未使用
 * @param  [in]  dev_id     未使用
 * @return 割り込み受理結果
 */
static irqreturn_t ppcdev_sram2rxq( int irq, void *dev_id )
{
    int i;
    int wp;
    int qid = -1;
    ppcdev_msgh_t rxmsgh;
    int rxresult;

    // SRAM送信データを読み込み
    memcpy_fromio( &rxmsgh, rxsram_ioaddr, sizeof( ppcdev_msgh_t ) );

    // IDによる受信キューの振り分け
    for ( i = _TX + 1; i < _MAXQ; i++ ) {
        if ( ppcdev_que[ i ] != NULL && rxmsgh.id == ppcdev_que[ i ]->id ) {
                qid = i;
                break;
        }
    }

    // 受信結果
    if ( qid < 0 ) {
        // IDが不正
        printk( KERN_DEBUG "%s : (%s) illegal ID\n", devname, __func__ );
        rxresult = _FATAL;
    }
    else if ( is_full( qid ) ) {
        // キューに空きが無ければリトライ要求
        printk( KERN_DEBUG "%s : (%s) que full retry\n", devname, __func__ );
        rxresult = _RETRY;
    }
    else {
        rxresult = _SUCCESS;
    }

    // 受信結果をSRAMにセット
    iowrite8( ( unsigned char )rxresult, txsram_ioaddr + PPCDEV_TXMSGW );

    // 受信に失敗したら、以下の処理をスキップ
    if ( rxresult != _SUCCESS )
        return IRQ_HANDLED;

    // メッセージ転送
    wp = get_wp( qid );
    memcpy( &ppcdev_que[ qid ]->msgq[ wp ], ( void * )&rxmsgh, PPCDEV_TXMSGW );
    set_wp( qid, PPCDEV_TXMSGW );

    // ボトムハーフ処理の振り分け
    switch ( qid ) {
    case _NWA: // 受信キューのリードウェイト解除
        wake_up_interruptible( &ppcdev_que[ qid ]->read_q );
        break;
    case _ETH: // ワークを待ち行列に追加
    case _DMA:
        ppcdev_work[ qid ].qid = qid;
        if ( !work_pending( ( struct work_struct * )&ppcdev_work[ qid ] ) )
            schedule_work( ( struct work_struct * )&ppcdev_work[ qid ] );
        break;
    default:
        printk( KERN_DEBUG "%s : (%s) not defined bottom-half [qid=%d]\n", devname, __func__, qid );
        break;
    }

    return IRQ_HANDLED;
}


/**
 * @brief  コールバック[Work Queie]
 * @param  [in]  work       ワークキュー構造体
 */
static void ppcdev_work_handler( struct work_struct *work )
{
    ppcdev_work_t *exwork;
    ppcdev_msgh_t rxmsgh;
    unsigned long head;
    unsigned long size;
    unsigned long offset;
    void *laddr;

    exwork = ( ppcdev_work_t * )work;

    if ( ppcdev_cb[ exwork->qid ] == NULL ) {
        // コールバックが登録されていなければ何もしない
        printk( KERN_DEBUG "%s : (%s) callback is null\n", devname, __func__ );
    }
    else {
        // 受信キューからデータをPOP
        do {
            if ( __ppcdev_read( exwork->qid, ( char * )&rxmsgh, PPCDEV_TXMSGW, PPCDEV_KAPI_SYNC ) != PPCDEV_TXMSGW ) {
                break;
            }
            else {
                // DRAMアドレス head/size計算
                head = ( unsigned long )rxmsgh.txdadr - sizeof( ppcdev_memh_t );
                offset = head & ( PPCDEV_PAGESIZE - 1 );
                size = offset + sizeof( ppcdev_memh_t ) + rxmsgh.size;

                // DRAMをカーネル空間にマッピング
                laddr = ioremap_nocache( head & ~( PPCDEV_PAGESIZE - 1 ), size );

                // コールバック呼び出し
                ( *ppcdev_cb[ exwork->qid ] )( exwork->qid, rxmsgh.id, laddr + offset + sizeof( ppcdev_memh_t ), ( size_t )rxmsgh.size );

                // DRAMを開放
                iowrite32( _FALSE, laddr + offset + sizeof( long ) );
                iounmap( laddr );
            }
        } while ( 1 );
    }
}


/**
 * @brief  割り込み送信コールバック
 * @param  [in] arg  未使用
 */
static void ppcdev_irack_cb( void *arg )
{
    // 割り込み応答待ちフラグをセット
    mutex_lock( &ppcdev_irack_mutex );
    ppcdev_irack_flag = 1;
    mutex_unlock( &ppcdev_irack_mutex );

    // 割り込み応答待ち解除
    wake_up_interruptible( &ppcdev_irack_q );
}

/* End of ppcdev_main.c */
