/**
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 * @file
 * @brief   CPU間通信D/D システムコール
 */
#include <asm/uaccess.h>    /* copy_from_user, copy_to_user */
#include <asm/io.h>         /* ioremap, iounmap */
#include <linux/errno.h>
#include <linux/fs.h>       /* inode, file, file_operations */
#include <linux/kernel.h>   /* printk() */
#include <linux/module.h>
#include <linux/ioctl.h>    /* ioctl */
#include <linux/poll.h>     /* POLLIN ... */
#include <linux/wait.h>     /* wait_queue_head_t */
#include <linux/sched.h>    /* wait_event_interruptible, wake_up_interruptible */

#include "ppcirq.h"
#include "ppcdev_main.h"
#include "ppcdev_util.h"
#include "ppcdev_ioctl.h"


// モジュール定義
MODULE_AUTHOR( "Panasonic" );
MODULE_DESCRIPTION( "ppc character device driver." );
MODULE_LICENSE( "Proprietary" );

int   devmajor = 60;
char *devname  = "ppcdev";


// グローバル変数
static int ppcdev_txtmout;            // 送信タイムアウト
static int ppcdev_rxtmout;            // 受信タイムアウト
DEFINE_MUTEX( ppcdev_txtmout_mutex ); // 送信タイムアウトミューテックス
DEFINE_MUTEX( ppcdev_rxtmout_mutex ); // 送信タイムアウトミューテックス


// プロトタイプ宣言
static int ppcdev_open( struct inode *inode, struct file *filp );
static int ppcdev_release( struct inode *inode, struct file *filp );
static ssize_t ppcdev_read( struct file *filp, char *buf, size_t count, loff_t *pos );
static ssize_t ppcdev_write( struct file *filp, const char *buf, size_t count, loff_t *pos );
static unsigned int ppcdev_poll( struct file *filp, poll_table *wait );
static int ppcdev_ioctl( struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg );
int init_module( void );
void cleanup_module( void );


// システムコール登録
static struct file_operations ppcdev_fops = 
{
owner   : THIS_MODULE,
read    : ppcdev_read,
write   : ppcdev_write,
poll    : ppcdev_poll,
ioctl   : ppcdev_ioctl,
open    : ppcdev_open,
release : ppcdev_release,
};


/**
 * @brief  デバイスオープン[Systemcall]
 * @param  [in]  inode      iノード
 * @param  [in]  filp       ファイル構造体
 * @return 0                     成功
 */
static int ppcdev_open( struct inode *inode, struct file *filp )
{
    filp->private_data = &inode->i_rdev;
    __ppcdev_open( _NWA, 0 );
    return 0;
}


/**
 * @brief  デバイスクローズ[Systemcall]
 * @param  [in]  inode      iノード
 * @param  [in]  filp       ファイル構造体
 * @return 0                     成功
 */
static int ppcdev_release( struct inode *inode, struct file *filp )
{
    __ppcdev_release( _NWA );
    return 0;
}


/**
 * @brief  受信バッファリード[Systemcall]
 * @param  [in]  filp       ファイル構造体
 * @param  [out] buf        データ格納バッファ
 * @param  [in]  count      リードバイト数
 * @param  [out] pos        オフセット
 * @return リードバイト数
 */
static ssize_t ppcdev_read( struct file *filp, char *buf, size_t count, loff_t *pos )
{
    int copy_len;
    int tmout;

    if ( mutex_lock_interruptible( &ppcdev_rxtmout_mutex ) ) {
        printk( KERN_DEBUG "%s : (%s) mutex_lock_interruptible failed\n", devname, __func__ );
        return -EFAULT;
    }
    tmout = ppcdev_rxtmout;
    mutex_unlock( &ppcdev_rxtmout_mutex );
    copy_len = __ppcdev_read( _NWA, buf, count, tmout );
    *pos += copy_len; // オフセットの更新

    return copy_len;
}


/**
 * @brief  送信バッファライト[Systemcall]
 * @param  [in]  filp       ファイル構造体
 * @param  [in]  buf        データ格納バッファ
 * @param  [in]  count      ライトバイト数
 * @param  [out] pos        オフセット
 * @return ライトバイト数
 */
static ssize_t ppcdev_write( struct file *filp, const char *buf, size_t count, loff_t *pos )
{
    ssize_t copy_len;
    int tmout;

    if ( mutex_lock_interruptible( &ppcdev_txtmout_mutex ) ) {
        printk( KERN_DEBUG "%s : (%s) mutex_lock_interruptible failed\n", devname, __func__ );
        return -EFAULT;
    }
    tmout = ppcdev_txtmout;
    mutex_unlock( &ppcdev_txtmout_mutex );
    copy_len = __ppcdev_write( _NWA, buf, count, tmout );
    *pos += copy_len; // オフセットの更新

    return copy_len;
}


/**
 * @brief  送受信バッファ待ち状態[Systemcall]
 * @param  [in]  filp       ファイル構造体
 * @param  [in]  wait       ポールテーブル
 * @return ポール状態
 */
static unsigned int ppcdev_poll( struct file *filp, poll_table *wait )
{
    unsigned int retmask = 0;
    int full_tx;
    int empty_rx;

    poll_wait( filp, &ppcdev_que[ _TX ]->read_q,  wait );
    poll_wait( filp, &ppcdev_que[ _NWA ]->write_q, wait );
    poll_wait( filp, &ppcdev_que[ _ETH ]->read_q,  wait );

    full_tx  = is_full( _TX );
    empty_rx = is_empty( _NWA );

    if ( empty_rx != 0 ) {
        retmask |= ( POLLIN  | POLLRDNORM );
    }

    if ( full_tx != 0 ) {
        retmask |= ( POLLOUT | POLLWRNORM );
    }

    return retmask;
}


/**
 * @brief  デバイス制御[Systemcall]
 * @param  [in]  inode      iノード
 * @param  [in]  filp       ファイル構造体
 * @param  [in]  cmd        コマンド
 * @param  [io]  arg        引数、戻り値ポインタ
 * @return 0                成功<br>
 *         エラー           失敗
 */
static int ppcdev_ioctl( struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg )
{
    int ret = 0;
    volatile int arg_size;
    volatile unsigned long arg_addr;
    unsigned long poffset;

    switch( cmd ) {
    case PPCDEV_IOCTL_IRQENABLE : // 割り込み許可
        ppcirq_kenable( ppcirq_fd );
        break;

    case PPCDEV_IOCTL_IRQDISABLE : // 割り込み禁止
        ppcirq_kdisable( ppcirq_fd );
        break;

    case PPCDEV_IOCTL_GETBLOCK : // DRAMメモリブロック獲得
        if ( copy_from_user( ( void * )&arg_size, ( void * )arg, 4 ) ) {
            printk( KERN_DEBUG "%s : (%s) IOCTL_GETBLOCK copy_from_user failed\n", devname, __func__ );
            ret = -EFAULT;
        }
        poffset = ( unsigned long )__ppcdev_kmalloc( arg_size ) - ( unsigned long )txdram_ioaddr;
        arg_addr = ( unsigned long )PPCDEV_TXDRAM_BASE + poffset;
        if ( copy_to_user( ( void * )arg, ( void * )&arg_addr, 4 ) ) {
            printk( KERN_DEBUG "%s : (%s) IOCTL_GETBLOCK copy_to_user failed\n", devname, __func__ );
            ret = -EFAULT;
        }
        break;

    case PPCDEV_IOCTL_FREEBLOCK : // DRAMメモリブロック開放
        __ppcdev_kfree( ( const void * )arg );
        break;

    case PPCDEV_IOCTL_TXTMOUT : // 送信タイムアウト値を設定
        if ( mutex_lock_interruptible( &ppcdev_txtmout_mutex ) ) {
            printk( KERN_DEBUG "%s : (%s) IOCTL_TXTMOUT mutex_lock_interruptible failed\n", devname, __func__ );
            return -ERESTARTSYS;
        }
        ppcdev_txtmout = ( int )arg;
        mutex_unlock( &ppcdev_txtmout_mutex );
        break;

    case PPCDEV_IOCTL_RXTMOUT : // 受信タイムアウト値を設定
        if ( mutex_lock_interruptible( &ppcdev_rxtmout_mutex ) ) {
            printk( KERN_DEBUG "%s : (%s) IOCTL_RXTMOUT mutex_lock_interruptible failed\n", devname, __func__ );
            return -ERESTARTSYS;
        }
        ppcdev_rxtmout = ( int )arg;
        mutex_unlock( &ppcdev_rxtmout_mutex );
        break;

    case PPCDEV_IOCTL_SETRETRY : // リトライ回数を設定
        if ( mutex_lock_interruptible( &ppcdev_retry_mutex ) ) {
            printk( KERN_DEBUG "%s : (%s) IOCTL_SETRETRY mutex_lock_interruptible failed\n", devname, __func__ );
            return -ERESTARTSYS;
        }
        ppcdev_retry = ( int )arg;
        mutex_unlock( &ppcdev_retry_mutex );
        break;

    default:
        ret = -EINVAL;
        break;
        /* POSIX 標準では -ENOTTY らしい */
    }

    return ret;
}


/**
 * @brief  モジュールの初期処理[insmod]
 * @return エラーコード
 */
int init_module( void )
{
    int i;
    ppcdev_memh_t memh;

    if ( register_chrdev( devmajor, devname, &ppcdev_fops ) ) {
        printk( KERN_DEBUG "%s : (%s) register_chrdev failed\n", devname, __func__ );
        return -EBUSY;
    }

    ppcdev_retry = 1;
    spin_lock_init( &ppcdev_qptr_spinlock );
    for ( i = 0; i < _MAXQ; i++ ) {
        ppcdev_que[ i ] = NULL; // 送受信キューの初期化
        ppcdev_cb[ i ] = NULL;  // コールバック初期化
    }

    srbase_ioaddr = ioremap_nocache( PPCDEV_SRAM_BASE, PPCDEV_SRAM_SIZE );      // 共有SRAM領域をマッピング
    txsram_ioaddr = srbase_ioaddr + PPCDEV_TXSRAM_OFFSET; // 送信SRAMオフセット加算
    rxsram_ioaddr = srbase_ioaddr + PPCDEV_RXSRAM_OFFSET; // 受信SRAMオフセット加算

    drbase_ioaddr = ioremap_nocache( PPCDEV_DRAM_BASE, PPCDEV_DRAM_SIZE );      // 共有DRAM領域をマッピング
    txdram_ioaddr = drbase_ioaddr + PPCDEV_TXDRAM_OFFSET; // 送信DRAMオフセット加算
    rxdram_ioaddr = drbase_ioaddr + PPCDEV_RXDRAM_OFFSET; // 受信DRAMオフセット加算

    // DRAM初期化
    for ( i = 0; i < PPCDEV_MEMSET; i++ ) {
        memh.signature = 0xBF;
        memh.en = 0;
        memh.size = 0;
        if ( i == PPCDEV_MEMSET - 1 )
            memh.next = 0; // tail NULL
        else
            memh.next = PPCDEV_TXDRAM_BASE + PPCDEV_MEMBLK * ( i + 1 );
        memcpy_toio( ( txdram_ioaddr + PPCDEV_MEMBLK * i ), &memh, sizeof( ppcdev_memh_t ) );
    }

    // SRAM初期化
    iowrite32( 0, txsram_ioaddr + 8 );
    iowrite32( 0, txsram_ioaddr + 12 );

    // タイムアウト初期化(ブロッキング)
    ppcdev_txtmout = -1;
    ppcdev_rxtmout = -1;

    return 0;
}


/**
 * @brief  モジュールの解放処理[rmmod]
 */
void cleanup_module( void )
{
    unregister_chrdev( devmajor, devname );
    iounmap( srbase_ioaddr ); // 共有SRAM領域アンマップ
    iounmap( drbase_ioaddr ); // 共有DRAM領域アンマップ
}

/* End of ppcdev_syscall.c */
