/**
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 * @file
 * @brief   CPU間通信D/Dキュー操作
 */
#include <asm/uaccess.h>    /* copy_from_user, copy_to_user */
#include <asm/io.h>         /* ioremap, iounmap */
#include <asm/irq.h>        /* request_irq, free_irq */
#include <linux/slab.h>     /* kmalloc, kfree */
#include <linux/errno.h>
#include <linux/fs.h>       /* inode, file, file_operations */
#include <linux/kernel.h>   /* printk() */
#include <linux/module.h>
#include <linux/wait.h>     /* wait_queue_head_t */
#include <linux/sched.h>    /* wait_event_interruptible, wake_up_interruptible */
#include <linux/spinlock.h>
#include <linux/mutex.h>
#include <linux/kthread.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>

#include "ppcdev_def.h"


// 外部変数
extern int   devmajor;
extern char *devname;


// グローバル変数
ppcdev_que_t *ppcdev_que[ _MAXQ ]; // 送受信キュー構造体


// プロトタイプ宣言
void init_que( int qid, int id );
int is_empty( int qid );
int is_full( int qid );
int get_rp( int qid );
int get_wp( int qid );
void set_rp( int qid, int inc );
void set_wp( int qid, int inc );


/**
 * @brief  内部キューの初期化
 * @param  [in] qid  キュー番号
 * @param  [in] id   送信ID
 */
void init_que( int qid, int id )
{
    spin_lock_init( &ppcdev_que[ qid ]->que_spinlock );
    init_waitqueue_head( &ppcdev_que[ qid ]->write_q );
    init_waitqueue_head( &ppcdev_que[ qid ]->read_q );
    ppcdev_que[ qid ]->id = id;
    ppcdev_que[ qid ]->rp = 0;
    ppcdev_que[ qid ]->wp = 0;
    ppcdev_que[ qid ]->empty = 1;
    ppcdev_que[ qid ]->full = 0;
}


/**
 * @brief  エンプティフラグチェック
 * @param  [in] qid  キュー番号
 * @return 1         empty<br>
 *         0         Not empty
 */
int is_empty( int qid )
{
    int empty;
    unsigned long flags;

    if ( ppcdev_que[ qid ] == NULL )
        return -EFAULT;

    spin_lock_irqsave( &ppcdev_que[ qid ]->que_spinlock, flags );
    empty = ppcdev_que[ qid ]->empty;
    spin_unlock_irqrestore( &ppcdev_que[ qid ]->que_spinlock, flags );
    return empty;
}


/**
 * @brief  フルフラグチェック
 * @param  [in] qid  キュー番号
 * @return 1         full<br>
 *         0         Not full
 */
int is_full( int qid )
{
    int full;
    unsigned long flags;

    if ( ppcdev_que[ qid ] == NULL )
        return -EFAULT;

    spin_lock_irqsave( &ppcdev_que[ qid ]->que_spinlock, flags );
    full = ppcdev_que[ qid ]->full;
    spin_unlock_irqrestore( &ppcdev_que[ qid ]->que_spinlock, flags );
    return full;
}


/**
 * @brief  リードポインタ参照
 * @param  [in] qid  キュー番号
 * @return リードポインタ
 */
int get_rp( int qid )
{
    int rp;
    unsigned long flags;

    if ( ppcdev_que[ qid ] == NULL )
        return -EFAULT;

    spin_lock_irqsave( &ppcdev_que[ qid ]->que_spinlock, flags );
    rp = ppcdev_que[ qid ]->rp;
    spin_unlock_irqrestore( &ppcdev_que[ qid ]->que_spinlock, flags );
    return rp;
}

/**
 * @brief  ライトポインタ参照
 * @param  [in] qid  キュー番号
 * @return ライトポインタ
 */
int get_wp( int qid )
{
    int wp;
    unsigned long flags;

    if ( ppcdev_que[ qid ] == NULL )
        return -EFAULT;

    spin_lock_irqsave( &ppcdev_que[ qid ]->que_spinlock, flags );
    wp = ppcdev_que[ qid ]->wp;
    spin_unlock_irqrestore( &ppcdev_que[ qid ]->que_spinlock, flags );
    return wp;
}


/**
 * @brief  リードポインタ更新
 * @param  [in] qid  キュー番号
 * @param  [in] inc  更新バイト数
 */
void set_rp( int qid, int inc )
{
    int next_rp;
    unsigned long flags;

    if ( ppcdev_que[ qid ] == NULL ) {
        printk( KERN_DEBUG "%s : (%s) ppcdev_que pointer null\n", devname, __func__ );
    }
    else {
        spin_lock_irqsave( &ppcdev_que[ qid ]->que_spinlock, flags );
        // update pointer
        next_rp = ppcdev_que[ qid ]->rp + inc;
        if ( next_rp >= PPCDEV_MAXBUF )
            ppcdev_que[ qid ]->rp = next_rp - PPCDEV_MAXBUF;
        else
            ppcdev_que[ qid ]->rp = next_rp;
        // update flag
        ppcdev_que[ qid ]->full = 0;
        if ( ppcdev_que[ qid ]->rp == ppcdev_que[ qid ]->wp )
            ppcdev_que[ qid ]->empty = 1;
        spin_unlock_irqrestore( &ppcdev_que[ qid ]->que_spinlock, flags );
    }
}


/**
 * @brief  ライトポインタ更新
 * @param  [in] qid  キュー番号
 * @param  [in] inc  更新バイト数
 */
void set_wp( int qid, int inc )
{
    int next_wp;
    unsigned long flags;

    if ( ppcdev_que[ qid ] == NULL ) {
        printk( KERN_DEBUG "%s : (%s) ppcdev_que pointer null\n", devname, __func__ );
    }
    else {
        spin_lock_irqsave( &ppcdev_que[ qid ]->que_spinlock, flags );
        // update pointer
        next_wp = ppcdev_que[ qid ]->wp + inc;
        if ( next_wp >= PPCDEV_MAXBUF )
            ppcdev_que[ qid ]->wp = next_wp - PPCDEV_MAXBUF;
        else
            ppcdev_que[ qid ]->wp = next_wp;
        // update flag
        ppcdev_que[ qid ]->empty = 0;
        if ( ppcdev_que[ qid ]->wp == ppcdev_que[ qid ]->rp )
            ppcdev_que[ qid ]->full = 1;
        spin_unlock_irqrestore( &ppcdev_que[ qid ]->que_spinlock, flags );
    }
}

/* End of ppcdev_util.c */
