/**
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 * @file
 * @brief   CPU間通信割り込みD/D
 */
#include <asm/io.h>         /* ioremap, iounmap */
//#include <asm/irq.h>        /* request_irq, free_irq */
#include <linux/slab.h>     /* kmalloc, kfree */
#include <linux/errno.h>
#include <linux/fs.h>       /* inode, file, file_operations */
#include <linux/kernel.h>   /* printk() */
#include <linux/module.h>
#include <linux/wait.h>     /* wait_queue_head_t */
#include <linux/sched.h>    /* wait_event_interruptible, wake_up_interruptible */
#include <linux/spinlock.h>
#include <linux/kthread.h>
#include <linux/interrupt.h>

#include "ppcirq.h"


// モジュール定義
MODULE_AUTHOR( "Panasonic" );
MODULE_DESCRIPTION( "ppc interrupt character device driver." );
MODULE_LICENSE( "Proprietary" );

static int   devmajor = 62;
static char *devname  = "ppcirq";


// マクロ定義
#define PPCIRQ_IRQC_BASE 0xFFFFE000 // 割り込み制御レジスタ   ベースアドレス
#define PPCIRQ_IRQC_STAT 0x190      // 割り込み状態レジスタ   ページ内オフセット
#define PPCIRQ_IRQC_MASK 0x194      // 割り込みマスクレジスタ ページ内オフセット
#define PPCIRQ_SWIR_BASE 0xFFFFF000 // SW割り込みレジスタ     ベースアドレス
#define PPCIRQ_IRQNUM    18         // 割り込みグループ番号
#define PPCIRQ_PAGESIZE  4096       // ページサイズ 4[KByte]
#define PPCIRQ_TXPOLLING 100        // 割り込みクリア待ちポーリング回数


// 構造体/共用体/型定義
enum ppcirq_valid_t { _INVALID = 0, _VALID }; // 有効/無効
enum ppcirq_irdir_t { _RX = 0, _TX, _NX };    // 送受信の方向(優先度順)
typedef struct ppcirq_irlist_t {
    int valid;                 // エントリ有効フラグ
    int irq;                   // 割り込み送信リクエスト
    int mask;                  // 受信割り込みマスク
    void __iomem *swir[ _NX ]; // SW割り込みレジスタ
    ppcirq_rxcb_t callback_rx; // 受信割り込みコールバック
    ppcirq_txcb_t callback_tx; // 送信割り込みコールバック
    spinlock_t irq_lock;       // スピンロック
} ppcirq_irlist_t;


// グローバル変数
static void __iomem *ppcirq_irqc_base; // 割り込み制御レジスタ   仮想アドレス
static void __iomem *ppcirq_swir_base; // SW割り込みレジスタ     仮想アドレス
static void __iomem *ppcirq_irqc_stat; // 割り込み状態レジスタ   仮想アドレス
static void __iomem *ppcirq_irqc_mask; // 割り込みマスクレジスタ 仮想アドレス
static spinlock_t ppcirq_irqc_lock;    // 割り込み制御レジスタスピンロック
static wait_queue_head_t ppcirq_tx_q;  // 割り込み送信ウェイト
static ppcirq_irlist_t ppcirq_irlist[ PPCIRQ_PIL_END ]; // 優先順位付き割り込みリスト配列


// プロトタイプ宣言
static irqreturn_t ppcirq_rxhandler( int irq, void *dev_id ); // 割り込みハンドラー
int init_module( void );         // ロード処理
void cleanup_module( void );     // アンロード処理


// システムコール登録
static struct file_operations ppcirq_fops = 
{
owner   : THIS_MODULE,
};


// シンボルテーブル登録
EXPORT_SYMBOL( ppcirq_kopen );
EXPORT_SYMBOL( ppcirq_kclose );
EXPORT_SYMBOL( ppcirq_kenable );
EXPORT_SYMBOL( ppcirq_kdisable );
EXPORT_SYMBOL( ppcirq_ksetcb );
EXPORT_SYMBOL( ppcirq_kdelcb );
EXPORT_SYMBOL( ppcirq_ksend );
EXPORT_SYMBOL( ppcirq_ksend_na );


/**
 * @brief  割り込みリスト作成[Kernel API]
 * @param  [in]  priority  優先順位
 * @param  [in]  mask      割り込みマスク
 * @param  [in]  rxofs     受信割り込みオフセットアドレス
 * @param  [in]  txofs     送信割り込みオフセットアドレス
 * @return インデックス
 */
int ppcirq_kopen( int priority, int mask, int rxofs, int txofs )
{
    unsigned long flags;

    if ( priority < 0 && priority >= PPCIRQ_PIL_END )
        return -EINVAL;

    spin_lock_irqsave( &ppcirq_irlist[ priority ].irq_lock, flags );
    /* critical section */
    if ( ppcirq_irlist[ priority ].valid == _VALID ) {
        /* error */
        spin_unlock_irqrestore( &ppcirq_irlist[ priority ].irq_lock, flags );
        return -EBUSY;
    }

    // 割り込みリスト登録
    ppcirq_irlist[ priority ].valid = _VALID;
    ppcirq_irlist[ priority ].mask  = mask;
    ppcirq_irlist[ priority ].swir[ _RX ] = ppcirq_swir_base + rxofs;
    ppcirq_irlist[ priority ].swir[ _TX ] = ppcirq_swir_base + txofs;
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ priority ].irq_lock, flags );

    return priority;
}


/**
 * @brief  割り込みリスト削除[Kernel API]
 * @param  [in]  fd         受信キューインデックス
 * @return エラーコード
 */
int ppcirq_kclose( int fd )
{
    unsigned long flags;

    if ( fd < 0 && fd >= PPCIRQ_PIL_END )
        return -EFAULT;

    spin_lock_irqsave( &ppcirq_irlist[ fd ].irq_lock, flags );
    /* critical section */
    ppcirq_irlist[ fd ].valid = _INVALID; // エントリ無効
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ fd ].irq_lock, flags );

    return 0;
}


/**
 * @brief  CPU間通信開始[Kernel API]
 * @param  [in]  fd        リストインデックス
 * @return エラーコード
 */
int ppcirq_kenable( int fd )
{
    unsigned long flags;
    int mask;

    if ( fd < 0 && fd >= PPCIRQ_PIL_END )
        return -EFAULT;

    spin_lock_irqsave( &ppcirq_irlist[ fd ].irq_lock, flags );
    /* critical section */
    mask = ppcirq_irlist[ fd ].mask;
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ fd ].irq_lock, flags );

    spin_lock( &ppcirq_irqc_lock );
    /* critical section */
    // 受信割り込み許可
    mask |= ioread32( ppcirq_irqc_mask );
    iowrite32( mask, ppcirq_irqc_mask );
    /* end */
    spin_unlock( &ppcirq_irqc_lock );

    return 0;
}


/**
 * @brief  CPU間通信停止[Kernel API]
 * @param  [in]  fd         リストインデックス
 * @return エラーコード
 */
int ppcirq_kdisable( int fd )
{
    unsigned long flags;
    int mask;

    if ( fd < 0 && fd >= PPCIRQ_PIL_END )
        return -EFAULT;

    spin_lock_irqsave( &ppcirq_irlist[ fd ].irq_lock, flags );
    /* critical section */
    mask = ~ppcirq_irlist[ fd ].mask;
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ fd ].irq_lock, flags );

    spin_lock( &ppcirq_irqc_lock );
    /* critical section */
    // 受信割り込み禁止
    mask &= ioread32( ppcirq_irqc_mask );
    iowrite32( mask, ppcirq_irqc_mask );
    /* end */
    spin_unlock( &ppcirq_irqc_lock );

    return 0;
}


/**
 * @brief  割り込みコールバック登録[Kernel API]
 * @param  [in]  fd         リストインデックス
 * @param  [in]  rxcb       受信割り込みコールバック
 * @param  [in]  txcb       送信割り込みコールバック
 * @return エラー
 */
int ppcirq_ksetcb( int fd, ppcirq_rxcb_t rxcb, ppcirq_txcb_t txcb )
{
    unsigned long flags;

    spin_lock_irqsave( &ppcirq_irlist[ fd ].irq_lock, flags );
    /* critical section */
    ppcirq_irlist[ fd ].callback_rx = rxcb;
    ppcirq_irlist[ fd ].callback_tx = txcb;
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ fd ].irq_lock, flags );

    return 0;
}


/**
 * @brief  割り込みコールバック削除[Kernel API]
 * @param  [in]  flag       割り込み要因フラグ
 * @param  [in]  handler    割り込みハンドラー
 * @return エラー
 */
int ppcirq_kdelcb( int fd )
{
    unsigned long flags;

    spin_lock_irqsave( &ppcirq_irlist[ fd ].irq_lock, flags );
    /* critical section */
    ppcirq_irlist[ fd ].callback_rx = NULL;
    ppcirq_irlist[ fd ].callback_tx = NULL;
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ fd ].irq_lock, flags );

    return 0;
}


/**
 * @brief  割り込み送信ハンドシェイク型[Kernel API]
 * @param  [in]  fd         割り込みリストインデックス
 * @return エラーコード
 */
int ppcirq_ksend( int fd )
{
    unsigned long flags;
    volatile int i;
    volatile int swir;
    volatile void __iomem *swir_ptr;
    volatile ppcirq_txcb_t callback;

    spin_lock_irqsave( &ppcirq_irlist[ fd ].irq_lock, flags );
    /* critical section */
    swir_ptr = ppcirq_irlist[ fd ].swir[ _TX ];
    callback = ppcirq_irlist[ fd ].callback_tx;
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ fd ].irq_lock, flags );

    iowrite32( _VALID, swir_ptr ); // 割り込みアサート

    // 割り込みクリア ポーリング回数待ち
    for ( i = 0; i < PPCIRQ_TXPOLLING; i++ ) {
        swir = ioread32( swir_ptr );
        if ( !( swir & _VALID ) )
            break;
    }

    // 割り込みクリア 無限ループ待ち
    while ( swir & _VALID ) {
        swir = ioread32( swir_ptr );
        if ( !( swir & _VALID ) )
            break;
        schedule(); // ディスパッチ
    }

    // 割り込み送信完了コールバック
    if ( callback != NULL )
        callback( NULL );

    return 0;
}


/**
 * @brief  割り込み送信突き放し型[Kernel API]
 * @param  [in]  fd         割り込みリストインデックス
 * @return エラーコード
 */
int ppcirq_ksend_na( int fd )
{
    unsigned long flags;
    volatile void __iomem *swir_ptr;
    volatile ppcirq_txcb_t callback;

    spin_lock_irqsave( &ppcirq_irlist[ fd ].irq_lock, flags );
    /* critical section */
    swir_ptr = ppcirq_irlist[ fd ].swir[ _TX ];
    callback = ppcirq_irlist[ fd ].callback_tx;
    /* end */
    spin_unlock_irqrestore( &ppcirq_irlist[ fd ].irq_lock, flags );

    iowrite32( _VALID, swir_ptr ); // 割り込みアサート
    if ( callback != NULL )
        callback( NULL );          // 割り込み送信完了コールバック

    return 0;
}


/**
 * @brief  割り込みハンドラーエントリー関数
 * @param  [in]  irq        未使用
 * @param  [in]  dev_id     未使用
 * @return 割り込み受理結果
 */
static irqreturn_t ppcirq_rxhandler( int irq, void *dev_id )
{
    volatile int i;
    volatile int stat;
    volatile int mask;
    volatile int valid;
    unsigned long flags;
    irqreturn_t ret = IRQ_NONE;

    // 割り込み要因チェック
    stat = ioread32( ppcirq_irqc_stat );

    for ( i = 0; i < PPCIRQ_PIL_END; i++ ) {
        spin_lock_irqsave( &ppcirq_irlist[ i ].irq_lock, flags );
        /* critical section */
        mask  = ppcirq_irlist[ i ].mask;
        valid = ppcirq_irlist[ i ].valid;
        /* end */
        spin_unlock_irqrestore( &ppcirq_irlist[ i ].irq_lock, flags );
        if ( valid && ( stat & mask ) )
            break;
        else if ( i == PPCIRQ_PIL_END - 1 )
            return IRQ_NONE;
    }

    // コールバック呼び出し
    if ( !ppcirq_irlist[ i ].callback_rx ) {
        printk( KERN_DEBUG "%s : (%s) callback is not set.", devname, __func__ );
        ret = IRQ_HANDLED;
    } else {
        ret = ppcirq_irlist[ i ].callback_rx( irq, dev_id );
    }

    // 割り込みクリア
    iowrite32( _INVALID, ppcirq_irlist[ i ].swir[ _RX ] );
    iowrite32( mask, ppcirq_irqc_stat );

    return ret;
}


/**
 * @brief  モジュールの初期処理[insmod]
 * @return エラーコード
 */
int init_module( void )
{
    int i;

    // ロック/ウェイトキュー初期化
    init_waitqueue_head( &ppcirq_tx_q );
    spin_lock_init( &ppcirq_irqc_lock );

    // 割り込みリスト初期化
    for ( i = 0; i < PPCIRQ_PIL_END; i++ ) {
        ppcirq_irlist[ i ].valid = _INVALID;
        ppcirq_irlist[ i ].callback_rx = NULL;
        ppcirq_irlist[ i ].callback_tx = NULL;
        spin_lock_init( &ppcirq_irlist[ i ].irq_lock );
    }

    // レジスタマッピング
    ppcirq_irqc_base = ioremap_nocache( PPCIRQ_IRQC_BASE, PPCIRQ_PAGESIZE );
    ppcirq_swir_base = ioremap_nocache( PPCIRQ_SWIR_BASE, PPCIRQ_PAGESIZE );

    ppcirq_irqc_stat = ppcirq_irqc_base + PPCIRQ_IRQC_STAT;
    ppcirq_irqc_mask = ppcirq_irqc_base + PPCIRQ_IRQC_MASK;

    // デバイス登録
    if ( register_chrdev( devmajor, devname, &ppcirq_fops ) ) {
        printk( KERN_DEBUG "%s : (%s) register_chrdev failed\n", devname, __func__ );
        return -EBUSY;
    }

    // 割り込みハンドラー登録
    if ( request_irq( PPCIRQ_IRQNUM, ppcirq_rxhandler, 0, devname, NULL ) ) {
        printk( KERN_DEBUG "%s : (%s) request_irq failed\n", devname, __func__ );
        unregister_chrdev( devmajor, devname ); // デバイス削除
        return -EBUSY;
    }

    return 0;
}

/**
 * @brief  モジュールの解放処理[rmmod]
 */
void cleanup_module( void )
{
    // 割り込みハンドラー抹消
    free_irq( PPCIRQ_IRQNUM, NULL );

    // デバイス削除
    unregister_chrdev( devmajor, devname );

    // レジスタマッピング解除
    iounmap( ppcirq_irqc_base );
    iounmap( ppcirq_swir_base );
}

/* End of ppcirq_main.c */
