/**
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 * @file
 * @brief	KAO-CPU間通信D/D
 */
#include <asm/io.h>			/* ioremap, iounmap */
#include <linux/errno.h>
#include <linux/fs.h>		/* inode, file, file_operations */
#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>
#include <linux/ioctl.h>	/* ioctl */
#include <linux/poll.h>		/* POLLIN ... */
#include <linux/wait.h>		/* wait_queue_head_t */
#include <linux/sched.h>	/* wait_event_interruptible, wake_up_interruptible */
#include <linux/timer.h>	/* init_timer, add_timer, del_timer_sync */
#include <linux/spinlock.h>
#include <linux/mutex.h>
#include <linux/interrupt.h>

#include "ppcirq.h"
#include "ppckao_ioctl.h"


// モジュール定義
MODULE_AUTHOR( "Panasonic" );
MODULE_DESCRIPTION( "ppc for kao character device driver." );
MODULE_LICENSE( "Proprietary" );

static int	 devmajor = 61;
static char	*devname  = "ppckao";


// グローバル変数
static int ppckao_txtmout;
static int ppckao_rxtmout;
static int ppckao_irqflag;
static int ppcirq_fd;
static wait_queue_head_t ppckao_read_q;
DEFINE_MUTEX( ppckao_txtmout_mutex );		// 送信タイムアウトミューテックス
DEFINE_MUTEX( ppckao_rxtmout_mutex );		// 送信タイムアウトミューテックス
static spinlock_t ppckao_irqflag_spinlock;	// 割り込みコントローラスピンロック


// プロトタイプ宣言
static int ppckao_open( struct inode *inode, struct file *filp );
static int ppckao_release( struct inode *inode, struct file *filp );
static ssize_t ppckao_read( struct file *filp, char *buf, size_t count, loff_t *pos );
static ssize_t ppckao_write( struct file *filp, const char *buf, size_t count, loff_t *pos );
static int ppckao_ioctl( struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg );
static irqreturn_t ppckao_handler( int irq, void *dev_id );
int init_module( void );
void cleanup_module( void );


// システムコール登録
static struct file_operations ppckao_fops = 
{
owner	: THIS_MODULE,
read	: ppckao_read,
write	: ppckao_write,
ioctl	: ppckao_ioctl,
open	: ppckao_open,
release	: ppckao_release,
};


/**
 * @brief  デバイスオープン[Systemcall]
 * @param  [in]  inode		iノード
 * @param  [in]  filp		ファイル構造体
 * @return 0					 成功
 */
static int ppckao_open( struct inode *inode, struct file *filp )
{
	// 初期化
	filp->private_data = &inode->i_rdev;
	ppckao_irqflag = 0;
	ppckao_txtmout = -1;
	ppckao_rxtmout = -1;
	spin_lock_init( &ppckao_irqflag_spinlock );

	// 割り込みD/D
	ppcirq_fd = ppcirq_kopen( PPCIRQ_PIL_KAO, PPC_IRC0DINT1_MASK, PPC_IRC0DINT1_OFS, PPC_IRC1DINT1_OFS );
	ppcirq_ksetcb( ppcirq_fd, ppckao_handler, NULL );
	ppcirq_kenable( ppcirq_fd );

	return 0;
}


/**
 * @brief  デバイスクローズ[Systemcall]
 * @param  [in]  inode		iノード
 * @param  [in]  filp		ファイル構造体
 * @return 0					 成功
 */
static int ppckao_release( struct inode *inode, struct file *filp )
{
	// 割り込みD/D
	ppcirq_kdisable( ppcirq_fd );
	ppcirq_kdelcb( ppcirq_fd );
	ppcirq_kclose( ppcirq_fd );

	return 0;
}


/**
 * @brief  開始割り込み待ち[Systemcall]
 * @param  [in]  filp		ファイル構造体
 * @param  [out] buf		データ格納バッファ
 * @param  [in]  count		リードバイト数
 * @param  [out] pos		オフセット
 * @return リードバイト数
 */
static ssize_t ppckao_read( struct file *filp, char *buf, size_t count, loff_t *pos )
{
	int tmout;
	int tmouterr;
	unsigned long flags;

	if ( mutex_lock_interruptible( &ppckao_rxtmout_mutex ) ) {
		printk( KERN_DEBUG "%s : (%s) mutex_lock_interruptible failed\n", devname, __func__ );
		return -EFAULT;
	}
	tmout = ppckao_rxtmout;
	mutex_unlock( &ppckao_rxtmout_mutex );

	// 割り込み待ち
	if ( !ppckao_irqflag ) {
		if ( tmout == 0 ) {
			printk( KERN_DEBUG "%s : (%s) NONBLOCK\n", devname, __func__ );
			return -EFAULT;
		}

		if ( tmout > 0 ) {
			tmouterr = wait_event_interruptible_timeout( ppckao_read_q, ppckao_irqflag, tmout );
			if ( tmouterr < 0 )
				return -ERESTARTSYS;
			else if ( tmouterr == 0 )
				return -EAGAIN;
		}
		else {
			if ( wait_event_interruptible( ppckao_read_q, ppckao_irqflag ) )
				return -ERESTARTSYS;
		}
	}

	spin_lock_irqsave( &ppckao_irqflag_spinlock, flags );
	ppckao_irqflag = 0;
	spin_unlock_irqrestore( &ppckao_irqflag_spinlock, flags );

	return 0;
}


/**
 * @brief  完了割り込み通知[Systemcall]
 * @param  [in]  filp		ファイル構造体
 * @param  [in]  buf		データ格納バッファ
 * @param  [in]  count		ライトバイト数
 * @param  [out] pos		オフセット
 * @return ライトバイト数
 */
static ssize_t ppckao_write( struct file *filp, const char *buf, size_t count, loff_t *pos )
{
	// 完了割り込み送信
#ifdef PPCKAO_INTNA
	ppcirq_ksend_na( ppcirq_fd );
#else
	ppcirq_ksend( ppcirq_fd );
#endif
	return 0;
}

/**
 * @brief  デバイス制御[Systemcall]
 * @param  [in]  inode		iノード
 * @param  [in]  filp		ファイル構造体
 * @param  [in]  cmd		コマンド
 * @param  [io]	 arg		引数、戻り値ポインタ
 * @return 0				成功<br>
 *         エラー			失敗
 */
static int ppckao_ioctl( struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg )
{
	int ret = 0;

//	printk( KERN_DEBUG "%s : (%s) called\n", devname, __func__ );

	switch( cmd ) {
	case PPCKAO_IOCTL_TXTMOUT :
		// 完了割り込みタイムアウト値を設定
//		printk( KERN_DEBUG "%s : (%s) IOCTL_TXTMOUT\n", devname, __func__ );
		if ( mutex_lock_interruptible( &ppckao_txtmout_mutex ) ) {
			printk( KERN_DEBUG "%s : (%s) IOCTL_TXTMOUT mutex_lock_interruptible failed\n", devname, __func__ );
			return -ERESTARTSYS;
		}
		ppckao_txtmout = ( int )arg;
		mutex_unlock( &ppckao_txtmout_mutex );
		break;

	case PPCKAO_IOCTL_RXTMOUT :
		// 開始割り込みタイムアウト値を設定
//		printk( KERN_DEBUG "%s : (%s) IOCTL_RXTMOUT\n", devname, __func__ );
		if ( mutex_lock_interruptible( &ppckao_rxtmout_mutex ) ) {
			printk( KERN_DEBUG "%s : (%s) IOCTL_RXTMOUT mutex_lock_interruptible failed\n", devname, __func__ );
			return -ERESTARTSYS;
		}
		ppckao_rxtmout = ( int )arg;
		mutex_unlock( &ppckao_rxtmout_mutex );
		break;

	default:
		ret = -EINVAL;
		break;
		/* POSIX 標準では -ENOTTY らしい */
	}

	return ret;
}


/**
 * @brief  開始割り込みフラグセット[割り込みハンドラー]
 * @param  [in]  irq		未使用
 * @param  [in]  dev_id		未使用
 * @return 割り込み受理結果
 */
static irqreturn_t ppckao_handler( int irq, void *dev_id )
{
	unsigned long flags;

	// 割り込み受信フラグセット
	spin_lock_irqsave( &ppckao_irqflag_spinlock, flags );
	ppckao_irqflag = 1;
	spin_unlock_irqrestore( &ppckao_irqflag_spinlock, flags );

	// リードウェイト解除
	wake_up_interruptible( &ppckao_read_q );

	return IRQ_HANDLED;
}


/**
 * @brief  モジュールの初期処理[insmod]
 * @return エラーコード
 */
int init_module( void )
{
	if ( register_chrdev( devmajor, devname, &ppckao_fops ) ) {
		printk( KERN_DEBUG "%s : (%s) register_chrdev failed\n", devname, __func__ );
		return -EBUSY;
	}

	init_waitqueue_head( &ppckao_read_q );
	return 0;
}

/**
 * @brief  モジュールの解放処理[rmmod]
 */
void cleanup_module( void )
{
	unregister_chrdev( devmajor, devname );
}

/* End of ppckao_main.c */
