/* ============================================================================ */
/*    include files                                                             */
/* ============================================================================ */
#include <asm/uaccess.h>    /* copy_from_user, copy_to_user */
#include <asm/page.h>       /* copy_from_user, copy_to_user */
#include <linux/errno.h>
#include <linux/fs.h>       /* inode, file, file_operations */
#include <linux/ioctl.h>    /* ioctl */
#include <linux/kernel.h>   /* printk */
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/wait.h>     /* wait_queue_head_t */
#include <linux/sched.h>    /* wait_event_interruptible, wake_up_interruptible */
#include <linux/interrupt.h>
#include <linux/irq.h>      /* IRQ macros*/
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/cacheflush.h>

#include <asm/xdmac_ioctl.h>
#include <asm/xdmac_drv.h>


/* ============================================================================ */
/*    macros                                                                    */
/* ============================================================================ */
/* hardware specification */
#define XDMAC_NUM_OF_CH    6  // number of XDMA logical channels
#define XDMAC_IRQ_NUMBER  42  // IRQ number of XDMAC

/* software specification */
#define XDMAC_MAX_OPEN     1            // number of being able to open each device file
#define XDMAC_VERSION      "1.00"        // device driver version
#define XDMAC_RELEASE_DAY  "2012-07-31" // release day

/* switch */
#define XDMAC_DBG_PRINT    0  // 0:debug print off, 1:debug print on

#if XDMAC_DBG_PRINT
	#define ddbg_printk(args...)    printk(args)
#else
	#define ddbg_printk(args...)   
#endif

/* XDMAC addresses */
#define XDMAC_BASE             0x612D0000 // XDMAC physical base address
#define XDMAC_IO_SIZE          0x6000     // XDMAC IO space size
#define XDMAC_CH_OFFSET        0x100      // offset address between each channel
#define XDMAC_CTR1_OFFSET      0x00       // offset address of channel control register 1
#define XDMAC_TFA_OFFSET       0x08       // offset address of transfer factor register
#define XDMAC_SADM_OFFSET      0x10       // offset address of source addressing mode register
#define XDMAC_DADM_OFFSET      0x14       // offset address of destination addressing mode register
#define XDMAC_SAD_OFFSET       0x20       // offset address of source address register
#define XDMAC_DAD_OFFSET       0x24       // offset address of destination address register
#define XDMAC_ITS_OFFSET       0x28       // offset address of cycle-stealing transfer-size register
#define XDMAC_TNUM_OFFSET      0x2C       // offset address of transfer count register
#define XDMAC_TSS_OFFSET       0x30       // offset address of startup and stop request register
#define XDMAC_IEN_OFFSET       0x34       // offset address of interrupt enable register
#define XDMAC_STAT_OFFSET      0x40       // offset address of status register
#define XDMAC_IR_OFFSET        0x44       // offset address of interrupt factor register
#define XDMAC_ID_OFFSET        0x48       // offset address of interrupt detect register

#define XDMAC_SQS0_OFFSET      0x2000     // offset address of rectangular X direction transfer size register 0
#define XDMAC_SQAD2D0_OFFSET   0x2008     // offset address of rectangular discontinuity address increment 2D register 0
#define XDMAC_SQS1_OFFSET      0x2100     // offset address of rectangular X direction transfer size register 1
#define XDMAC_SQAD2D1_OFFSET   0x2108     // offset address of rectangular discontinuity address increment 2D register 1
#define XDMAC_SQS2_OFFSET      0x2200     // offset address of rectangular X direction transfer size register 2
#define XDMAC_SQAD2D2_OFFSET   0x2208     // offset address of rectangular discontinuity address increment 2D register 2
#define XDMAC_SQS3_OFFSET      0x2300     // offset address of rectangular X direction transfer size register 3
#define XDMAC_SQAD2D3_OFFSET   0x2308     // offset address of rectangular discontinuity address increment 2D register 3

#define XDMAC_PRI_OFFSET       0x4000     // offset address of priority scheme register
#define XDMAC_KCDATA_OFFSET    0x4020     // offset address of key color data register
#define XDMAC_ACTB_OFFSET      0x4100     // offset address of address translation table base register
#define XDMAC_MSTAT_OFFSET     0x5000     // offset address of multiple channel status register
#define XDMAC_MIR_OFFSET       0x5004     // offset address of multiple channel interrupt factor register
#define XDMAC_MID_OFFSET       0x5008     // offset address of multiple channel interrupt detect register

/* XDMAC bit position */
// channel control register 1
#define XDMAC_CTR1_KCEN        18
#define XDMAC_CTR1_KCS         16
// source addressing mode register
#define XDMAC_SADM_STW         24
#define XDMAC_SADM_SE          16
#define XDMAC_SADM_SSQNUM      12
#define XDMAC_SADM_SAM          4
#define XDMAC_SADM_STM          0
// destination addressing mode register
#define XDMAC_DADM_STW         24
#define XDMAC_DADM_SE          16
#define XDMAC_DADM_SSQNUM      12
#define XDMAC_DADM_SAM          4
#define XDMAC_DADM_STM          0
// multiple channel interrupt factor register
#define XDMAC_MIR_ERRF0         8
// interrupt factor register
#define XDMAC_IR_ERRF        0x02
#define XDMAC_IR_ENDF        0x01

/* XDMAC parameters */
// startup and stop request register
#define XDMAC_TSS_TSS_STOP     0x00
#define XDMAC_TSS_TSS_START    0x01
// source addressing mode register
#define XDMAC_SADM_RECT0       0x00
#define XDMAC_SADM_RECT1       0x01
#define XDMAC_SADM_RECT2       0x02
#define XDMAC_SADM_RECT3       0x03
// destination addressing mode register
#define XDMAC_DADM_RECT0       0x00
#define XDMAC_DADM_RECT1       0x01
#define XDMAC_DADM_RECT2       0x02
#define XDMAC_DADM_RECT3       0x03
// interrupt enable register
#define XDMAC_IEN_ERRIEN_EN    0x02
#define XDMAC_IEN_ERRIEN_DIS   0x00
#define XDMAC_IEN_ENDIEN_EN    0x01
#define XDMAC_IEN_ENDIEN_DIS   0x00
// status register
#define XDMAC_STAT_STOPF_STOP  0x02
#define XDMAC_STAT_TENF_START  0x01
// interrupt factor register
#define XDMAC_IR_ERRF_CLR      0x02
#define XDMAC_IR_ENDF_CLR      0x01
// priority scheme register
#define XDMAC_PRI_PRI_FIX      0x00
#define XDMAC_PRI_PRI_RR       0x01


/* ============================================================================ */
/*    enum, structure                                                           */
/* ============================================================================ */
typedef enum {
	irq_none = 0, // there are no IRQs
	irq_cmp  = 1, // transfer was completed 
	irq_err  = 2  // error happened
} EIrqStat;


/* ============================================================================ */
/*    local valiables                                                           */
/* ============================================================================ */
/* device information */
static int   devmajor     = 120;
static char *devname      = "xdmac";
static char *procname_val = "driver/xdmac_val";
static char *procname_reg = "driver/xdmac_reg";
static char *msg          = "module [xdmac.ko]";

/* etc */
static wait_queue_head_t	tWqXdmaIrq[XDMAC_NUM_OF_CH];    // waiting-que flags
static volatile char		cAccessNum[XDMAC_NUM_OF_CH];    // number of accesses
static volatile EStatType	eXdmaStat[XDMAC_NUM_OF_CH];     // XDMA state
static volatile EIrqStat	eXdmacIrqStat[XDMAC_NUM_OF_CH]; // IRQ status
static volatile void __iomem *xdmac_base;
static volatile long		lErrStat[XDMAC_NUM_OF_CH];      // DMAC status after iXdmaStart
static struct semaphore  tSemXdmac[XDMAC_NUM_OF_CH];     // semaphore(each channel)
static struct semaphore  tSemXdmacCmn;                   // semaphore(common)
static struct semaphore  tSemXdmacProcVal;               // semaphore(/proc/driver/xdmac_val)
static struct semaphore  tSemXdmacProcReg;               // semaphore(/proc/dirver/xdmac_reg)


/* ============================================================================ */
/*    function prototype definitions                                            */
/* ============================================================================ */
static irqreturn_t xdmac_irq_isr(int, void *);
static int iXdmaLsiInit(void);
static int iXdmaLsiFin(void);
static int iXdmaChStopPoll(int);


/* ============================================================================ */
/*    sytem call main functions                                                 */
/* ============================================================================ */
/*
 * open()
 */
static int xdmac_open(struct inode* inode, struct file* filp)
{
	int iMinorNo;

	iMinorNo = MINOR(inode->i_rdev); // get the device minor number
	ddbg_printk(KERN_INFO "%s : open(%d)  called\n", msg, iMinorNo);

	if (iMinorNo < 0 || iMinorNo >= XDMAC_NUM_OF_CH) {
		return -ENODEV;
	}

	filp->private_data = &(inode->i_rdev);	// save the minor number

	/* critical section in */
	if (down_interruptible(&tSemXdmac[iMinorNo])) {
		printk(KERN_ERR "%s: down_interruptible error\n", msg); // signal interrupt happened
		return -EINTR;
	}

	/* check how many device files have already been opened */
	if (cAccessNum[iMinorNo] >= (char)XDMAC_MAX_OPEN) {
		/* critical section out */
		up(&tSemXdmac[iMinorNo]);
		return -EBUSY;
	}

	cAccessNum[iMinorNo]++;

	/* critical section out */
	up(&tSemXdmac[iMinorNo]);

	return 0;
}


/*
 * release()
 */
static int xdmac_release(struct inode* inode, struct file* filp)
{
	int iMinorNo;

	iMinorNo = MINOR(inode->i_rdev); // get the device minor number
	ddbg_printk(KERN_INFO "%s : close(%d) called\n", msg, iMinorNo);

	if (iMinorNo < 0 || iMinorNo >= XDMAC_NUM_OF_CH) {
		return -ENODEV;
	}

	/* critical section in */
	if (down_interruptible(&tSemXdmac[iMinorNo])) {
		printk(KERN_ERR "%s: down_interruptible error\n", msg); // signal interrupt happened
		return -EINTR;
	}

	cAccessNum[iMinorNo]--;

	/* critical section out */
	up(&tSemXdmac[iMinorNo]);

	return 0;
}

/* 
 * ioctl()
 */
static long xdmac_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	dev_t			i_rdev;
	int				iMinorNo;
	SDmaParam		tDmaSet;
	SDmaStat		tDmaStat;
	SDmaCfg			tDmaCfg;
	SDmaDstInfo		tDmaDstInfo;
	SDmaRegInfo		tDmaRegInfo;
	int				iRet;

//	ddbg_printk(KERN_INFO "%s : ioctl() called %d\n", msg, cmd);

	i_rdev = *((int *)filp->private_data);

	iMinorNo = MINOR(i_rdev);

	if (iMinorNo < 0 || iMinorNo >= XDMAC_NUM_OF_CH) {
		return -ENODEV;
	}

	switch (cmd) {
		case XDMAC_IOCTL_INIT :
			iRet = iXdmaInit(iMinorNo);
			if (iRet == XDMAC_NG) {
				return -EBUSY;
			}
			break;
		case XDMAC_IOCTL_GET :
			iRet = iXdmaGet(iMinorNo, &tDmaStat);
			if (copy_to_user((void *)arg, &tDmaStat, sizeof(SDmaStat))) {
				return -EFAULT;
			}
//			ddbg_printk(KERN_INFO "/dev/xdmac%d GET[%x]\n", iMinorNo, tDmaStat.eStat);
			break;
		case XDMAC_IOCTL_START :
			if (copy_from_user(&tDmaSet, (void *)arg, sizeof(SDmaParam))) {
				return -EFAULT;
			}
			iRet = iXdmaStart(iMinorNo, &tDmaSet);
			if (iRet == XDMAC_NG) {
				if (lErrStat[iMinorNo] == 0) {
					return -EBUSY;
				} else {
					return -lErrStat[iMinorNo];
				}
			}
			break;
		case XDMAC_IOCTL_CONFIG :
			if (copy_from_user(&tDmaCfg, (void *)arg, sizeof(SDmaCfg))) {
				return -EFAULT;
			}
			iRet = iXdmaCfg(&tDmaCfg);
			if (iRet == XDMAC_NG) {
				return -EBUSY;
			}
			break;
		case XDMAC_IOCTL_DCINV :
			ddbg_printk(KERN_INFO "/dev/xdmac%d DCINV\n", iMinorNo);
			if (copy_from_user(&tDmaDstInfo, (void *)arg, sizeof(SDmaDstInfo))) {
				return -EFAULT;
			}
			iRet = iXdmaDcinv(&tDmaDstInfo);
			break;
		case XDMAC_IOCTL_REGREAD :
			if (copy_from_user(&tDmaRegInfo, (void *)arg, sizeof(SDmaRegInfo))) {
				return -EFAULT;
			}
			tDmaRegInfo.uiData = iXdmaRegRead(tDmaRegInfo.uiAddr);
			if (copy_to_user((void *)arg, &tDmaRegInfo, sizeof(SDmaRegInfo))) {
				return -EFAULT;
			}
			break;
		case XDMAC_IOCTL_REGWRITE :
			if (copy_from_user(&tDmaRegInfo, (void *)arg, sizeof(SDmaRegInfo))) {
				return -EFAULT;
			}
			iXdmaRegWrite(tDmaRegInfo.uiAddr, tDmaRegInfo.uiData);
			break;
		default:
			ddbg_printk(KERN_INFO "%s : IOCTL NG\n", msg);
			return -EINVAL;
	}

	return 0;
}

/* 
 * proc read
 */
static int proc_read(char* page, char **start, off_t offset, int count, int *eof, void *data)
{
	int iLen, iCh;

	ddbg_printk(KERN_INFO "%s : proc read()  called\n", procname_val);

	/* critical section in */
	if (down_interruptible(&tSemXdmacProcVal)) {
		printk(KERN_INFO "%s : down_interruptible for read failed\n", procname_val);
		return -ERESTARTSYS;
	}

	iLen = 0;
	for (iCh = 0; iCh < XDMAC_NUM_OF_CH; iCh++) {
		iLen += sprintf(page+iLen, "Ch%d : cAccessNum    = %d\n", iCh, cAccessNum[iCh]);

		if (eXdmaStat[iCh] == stat_trns) {
			iLen += sprintf(page+iLen, "Ch%d : eXdmaStat     = stat_trns\n", iCh);
		} else if (eXdmaStat[iCh] == stat_stop) {
			iLen += sprintf(page+iLen, "Ch%d : eXdmaStat     = stat_stop\n", iCh);
		} else {
			iLen += sprintf(page+iLen, "Ch%d : eXdmaStat     = stat_err\n", iCh);
		}

		if (eXdmacIrqStat[iCh] == irq_none) {
			iLen += sprintf(page+iLen, "Ch%d : eXdmacIrqStat = irq_none\n", iCh);
		} else if (eXdmacIrqStat[iCh] == irq_cmp) {
			iLen += sprintf(page+iLen, "Ch%d : eXdmacIrqStat = irq_cmp\n", iCh);
		} else {
			iLen += sprintf(page+iLen, "Ch%d : eXdmacIrqStat = irq_err\n", iCh);
		}
		iLen += sprintf(page+iLen, "\n");
	}

	/* critical section out */
	up(&tSemXdmacProcVal);

	*eof = 1;
	return iLen;
}

/* 
 * proc write(not supported)
 */
static int proc_write(struct file* filp, const char* buffer, unsigned long count, void* data)
{
//	int copy_len;

	ddbg_printk(KERN_INFO "%s : proc write()  called\n", procname_val);
	return -ENOSYS;

//	return copy_len;
}

/* 
 * proc read_reg
 */
static int proc_read_reg(char* page, char **start, off_t offset, int count, int *eof, void *data)
{
	int				iLen, iCh;
	unsigned int	uiOffset;

	ddbg_printk(KERN_INFO "%s : proc read()  called\n", procname_reg);

	/* critical section in */
	if (down_interruptible(&tSemXdmacProcReg)) {
		printk(KERN_INFO "%s : down_interruptible for read failed\n", procname_reg);
		return -ERESTARTSYS;
	}

	iLen = 0;
	for (iCh = 0; iCh < XDMAC_NUM_OF_CH; iCh++) {
		uiOffset = iCh * XDMAC_CH_OFFSET;
		iLen += sprintf(page+iLen, "XDMA%dCTR1 = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_CTR1_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dTFA  = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_TFA_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dSADM = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_SADM_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dDADM = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_DADM_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dSAD  = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_SAD_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dDAD  = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_DAD_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dITS  = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_ITS_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dTNUM = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_TNUM_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dTSS  = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_TSS_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dIEN  = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_IEN_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dSTAT = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_STAT_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dIR   = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_IR_OFFSET));
		iLen += sprintf(page+iLen, "XDMA%dID   = 0x%08x\n", iCh, ioread32(xdmac_base + uiOffset + XDMAC_ID_OFFSET));
		iLen += sprintf(page+iLen, "\n");
	}
	iLen += sprintf(page+iLen, "XDMASQS0     = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQS0_OFFSET));
	iLen += sprintf(page+iLen, "XDMASQSAD2D0 = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQAD2D0_OFFSET));
	iLen += sprintf(page+iLen, "XDMASQS1     = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQS1_OFFSET));
	iLen += sprintf(page+iLen, "XDMASQSAD2D1 = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQAD2D1_OFFSET));
	iLen += sprintf(page+iLen, "XDMASQS2     = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQS2_OFFSET));
	iLen += sprintf(page+iLen, "XDMASQSAD2D2 = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQAD2D2_OFFSET));
	iLen += sprintf(page+iLen, "XDMASQS3     = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQS3_OFFSET));
	iLen += sprintf(page+iLen, "XDMASQSAD2D3 = 0x%08x\n", ioread32(xdmac_base + XDMAC_SQAD2D3_OFFSET));
	iLen += sprintf(page+iLen, "\n");

	iLen += sprintf(page+iLen, "XDMAPRI      = 0x%08x\n", ioread32(xdmac_base + XDMAC_PRI_OFFSET));
	iLen += sprintf(page+iLen, "XDMAKCDATA   = 0x%08x\n", ioread32(xdmac_base + XDMAC_PRI_OFFSET));
	iLen += sprintf(page+iLen, "XDMAMSTAT    = 0x%08x\n", ioread32(xdmac_base + XDMAC_MSTAT_OFFSET));
	iLen += sprintf(page+iLen, "XDMAIR       = 0x%08x\n", ioread32(xdmac_base + XDMAC_MIR_OFFSET));
	iLen += sprintf(page+iLen, "XDMAID       = 0x%08x\n", ioread32(xdmac_base + XDMAC_MID_OFFSET));

	/* critical section out */
	up(&tSemXdmacProcReg);

	*eof = 1;
	return iLen;
}

/* 
 * proc write_reg(not supported)
 */
static int proc_write_reg(struct file* filp, const char* buffer, unsigned long count, void* data)
{
//	int copy_len;

	ddbg_printk(KERN_INFO "%s : proc write reg()  called\n", procname_reg);
	return -ENOSYS;

//	return copy_len;
}

static struct file_operations xdmac_fops = 
{
	.owner          = THIS_MODULE,
	.open           = xdmac_open,
	.release        = xdmac_release,
	.unlocked_ioctl = xdmac_ioctl,
};


/*
 * initialize (called at insmod)
 */
int init_module(void)
{
	int iCh;
	struct proc_dir_entry *entry;

	/* proc */
	entry = create_proc_entry(procname_val, 0666, NULL);
	if (entry != NULL) {
		entry->read_proc  = proc_read;
		entry->write_proc = proc_write;
	} else {
		printk(KERN_INFO "%s : create_proc_entry failed\n", procname_val);
		return -EBUSY;
	}
	init_MUTEX(&tSemXdmacProcVal);              // semaphore(proc)

	entry = create_proc_entry(procname_reg, 0666, NULL);
	if (entry != NULL) {
		entry->read_proc  = proc_read_reg;
		entry->write_proc = proc_write_reg;
	} else {
		printk(KERN_INFO "%s : create_proc_entry failed\n", procname_reg);
		return -EBUSY;
	}
	init_MUTEX(&tSemXdmacProcReg);              // semaphore(proc)


	/* semaphore, variables initialization */
	for (iCh = 0; iCh < XDMAC_NUM_OF_CH; iCh++) {
		init_MUTEX(&tSemXdmac[iCh]);           // semaphore(each channel)
		eXdmacIrqStat[iCh] = irq_none;           // IRQ state
		eXdmaStat[iCh]     = stat_stop;          // XDMA state
		cAccessNum[iCh]    = 0   ;               // number of being able to open each device file
		init_waitqueue_head(&tWqXdmaIrq[iCh]); // waiting-que flag
	}
	init_MUTEX(&tSemXdmacCmn);                // semaphore(common)

	/* ioremap */
	xdmac_base = ioremap_nocache(XDMAC_BASE, XDMAC_IO_SIZE);

	ddbg_printk(KERN_INFO "xdma base (%x) \n", (unsigned int)xdmac_base);

	/* LSI initialization */
	iXdmaLsiInit();

	/* register device driver */
	if (register_chrdev(devmajor, devname, &xdmac_fops)) {
		printk(KERN_INFO "%s : register_chrdev failed\n", devname);
		return -EBUSY;
	}

	/* register irq handler */
	if (request_irq(XDMAC_IRQ_NUMBER, xdmac_irq_isr, IRQF_DISABLED | IRQF_VALID, devname, NULL)) {
		unregister_chrdev(devmajor, devname);
		printk(KERN_INFO "%s : request_irq failed\n", devname);
		return -EBUSY;
	}

	ddbg_printk(KERN_INFO "%s : loaded  into kernel\n", msg);
	printk(KERN_INFO "MN2PS XDMAC driver version %s (%s)\n", XDMAC_VERSION, XDMAC_RELEASE_DAY);

	return 0;
}


/*
 * finalize (called at rmmod)
 */
void cleanup_module(void)
{
	/* LSI finalization */
	iXdmaLsiFin();

	/* free irq */
	free_irq(XDMAC_IRQ_NUMBER, NULL);

	/* unregister device driver */
	unregister_chrdev(devmajor, devname);

	/* iounmap */
	iounmap(xdmac_base);

	/* proc */
	remove_proc_entry(procname_val, NULL);
	remove_proc_entry(procname_reg, NULL);

	ddbg_printk(KERN_INFO "%s : removed from kernel\n", msg);
}


/* ============================================================================ */
/*    common functions (used by both XDMAC-DD and other DDs                     */
/* ============================================================================ */
/*
 * $Function : int iXdmaInit(EXdmaCh eCh)
 * $Args     : [IN]  eCh       : logical channel number of XDMA
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : initialize "eCh"th XDMAC
 */
int iXdmaInit(EXdmaCh eCh)
{
	if (eCh < 0 || eCh >= XDMAC_NUM_OF_CH) {
		return XDMAC_NG;
	}

	/* critical section in */
	if (down_interruptible(&tSemXdmac[eCh])) {
		printk(KERN_ERR "%s: down_interruptible error\n", msg); // interrupt happened
		return XDMAC_NG;
	}

	ddbg_printk(KERN_INFO "/dev/xdmac%d INIT\n", eCh);

	// return NG in the case among transfer
	if (eXdmaStat[eCh] == stat_trns) {
		/* critical section out */
		up(&tSemXdmac[eCh]);
		return XDMAC_NG;
	}

	// prohibit interrupts
	iowrite32((XDMAC_IEN_ERRIEN_DIS | XDMAC_IEN_ENDIEN_DIS),
	          (xdmac_base + XDMAC_IEN_OFFSET + (eCh * XDMAC_CH_OFFSET))); 

	// stop DMA
	iXdmaChStopPoll(eCh);

	// permit interrupts
	iowrite32((XDMAC_IEN_ERRIEN_EN | XDMAC_IEN_ENDIEN_EN),
	          (xdmac_base + XDMAC_IEN_OFFSET + (eCh * XDMAC_CH_OFFSET))); 

	// fail safe
//	wake_up_interruptible(&tWqXdmaIrq[eCh]);

	eXdmaStat[eCh]     = stat_stop;
	eXdmacIrqStat[eCh] = irq_none;

	/* critical section out */
	up(&tSemXdmac[eCh]);

	return XDMAC_OK;
}
EXPORT_SYMBOL(iXdmaInit);

/*
 * $Function : int iXdmaStart(EXdmaCh eCh, SDmaParam *ptDmaSet)
 * $Args     : [IN]  eCh       : logical channel number of XDMA
             : [IN]  *tDmaSet  : parameters for DMA transfer 
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : set and start "eCh"th XDMAC
 */
int iXdmaStart(EXdmaCh eCh, SDmaParam *ptDmaSet)
{
	unsigned int uiData, uiAddr;
	unsigned int uiHeight, uiNum;

	if (eCh < 0 || eCh >= XDMAC_NUM_OF_CH) {
		return XDMAC_NG;
	}

	/* clear status */
	lErrStat[eCh] = 0; /* no error */
	
	/* critical section in */
	if (down_interruptible(&tSemXdmac[eCh])) {
		printk(KERN_ERR "%s: down_interruptible error\n", msg); // interrupt happened
		return XDMAC_NG;
	}
	ddbg_printk(KERN_INFO "/dev/xdmac%d START\n", eCh);

	// return NG in the case among transfer
	if (eXdmaStat[eCh] == stat_trns) {
		/* critical section out */
		up(&tSemXdmac[eCh]);
		return XDMAC_NG;
	}

	// channel control register 1
	if (ptDmaSet->cKcen == (char)XDMAC_ON) {
		uiData = (unsigned int)(1 << XDMAC_CTR1_KCEN) | (unsigned int)(ptDmaSet->eKcs << XDMAC_CTR1_KCS);
	} else {
		uiData = 0;
	}

	iowrite32(uiData, (xdmac_base + XDMAC_CTR1_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// transfer factor register
	iowrite32(ptDmaSet->eTfa, (xdmac_base + XDMAC_TFA_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// source addressing mode register
	uiData = (unsigned int)(
		(ptDmaSet->eSrcTw     << XDMAC_SADM_STW) |
		(ptDmaSet->eSrcEndian << XDMAC_SADM_SE)  |
		(ptDmaSet->eSrcAm     << XDMAC_SADM_SAM) |
		(ptDmaSet->eSrcTm     << XDMAC_SADM_STM)
	);
	if (ptDmaSet->eSrcTm == tm_rct) {
		if (eCh == xdmac_ch0) {
			uiData |= (unsigned int)(XDMAC_SADM_RECT0 << XDMAC_SADM_SSQNUM);
		} else if (eCh == xdmac_ch1) {
			uiData |= (unsigned int)(XDMAC_SADM_RECT2 << XDMAC_SADM_SSQNUM);
		}
	}
	iowrite32(uiData, (xdmac_base + XDMAC_SADM_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// destination addressing mode register
	uiData = (unsigned int)(
		(ptDmaSet->eDstTw     << XDMAC_DADM_STW) |
		(ptDmaSet->eDstEndian << XDMAC_DADM_SE)  |
		(ptDmaSet->eDstAm     << XDMAC_DADM_SAM) |
		(ptDmaSet->eDstTm     << XDMAC_DADM_STM)
	);
	if (ptDmaSet->eDstTm == tm_rct) {
		if (eCh == xdmac_ch0) {
			uiData |= (unsigned int)(XDMAC_DADM_RECT1 << XDMAC_DADM_SSQNUM);
		} else if (eCh == xdmac_ch1) {
			uiData |= (unsigned int)(XDMAC_DADM_RECT3 << XDMAC_DADM_SSQNUM);
		}
	}
	iowrite32(uiData, (xdmac_base + XDMAC_DADM_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// source address register
	iowrite32(ptDmaSet->uiSrcAdr, (xdmac_base + XDMAC_SAD_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// destination address register
	iowrite32(ptDmaSet->uiDstAdr, (xdmac_base + XDMAC_DAD_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// cycle-stealing transfer-size register
	iowrite32(ptDmaSet->uiIts, (xdmac_base + XDMAC_ITS_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// transfer count register
	iowrite32(ptDmaSet->uiTnum, (xdmac_base + XDMAC_TNUM_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	// rectangular X direction transfer size register 0
	// rectangular discontinuity address increment 2D register 0
	if (ptDmaSet->eSrcTm == tm_rct) {
		if (eCh == xdmac_ch0) {
			iowrite32(ptDmaSet->uiSqs,    (xdmac_base + XDMAC_SQS0_OFFSET));
			iowrite32(ptDmaSet->uiSqad2d, (xdmac_base + XDMAC_SQAD2D0_OFFSET));
		} else if (eCh == xdmac_ch1) {
			iowrite32(ptDmaSet->uiSqs,    (xdmac_base + XDMAC_SQS2_OFFSET));
			iowrite32(ptDmaSet->uiSqad2d, (xdmac_base + XDMAC_SQAD2D2_OFFSET));
		} else {
			/* critical section out */
			up(&tSemXdmac[eCh]);
			return XDMAC_NG;
		}
	}
	if (ptDmaSet->eDstTm == tm_rct) {
		if (eCh == xdmac_ch0) {
			iowrite32(ptDmaSet->uiSqs,    (xdmac_base + XDMAC_SQS1_OFFSET));
			iowrite32(ptDmaSet->uiSqad2d, (xdmac_base + XDMAC_SQAD2D1_OFFSET));
		} else if (eCh == xdmac_ch1) {
			iowrite32(ptDmaSet->uiSqs,    (xdmac_base + XDMAC_SQS3_OFFSET));
			iowrite32(ptDmaSet->uiSqad2d, (xdmac_base + XDMAC_SQAD2D3_OFFSET));
		} else {
			/* critical section out */
			up(&tSemXdmac[eCh]);
			return XDMAC_NG;
		}
	}

	eXdmaStat[eCh]     = stat_trns;
	eXdmacIrqStat[eCh] = irq_none;

	// startup and stop request register
	iowrite32(XDMAC_TSS_TSS_START, (xdmac_base + XDMAC_TSS_OFFSET + (eCh * XDMAC_CH_OFFSET)));

	if (ptDmaSet->cWaitEvent == XDMAC_ON) {
		if (wait_event_interruptible(tWqXdmaIrq[eCh], 
			((eXdmacIrqStat[eCh] == irq_cmp) || (eXdmacIrqStat[eCh] == irq_err)))) {
			/* critical section out */
			up(&tSemXdmac[eCh]);
			return XDMAC_NG;
		}

		if (eXdmacIrqStat[eCh] == irq_err) {
			/* critical section out */
			up(&tSemXdmac[eCh]);
			lErrStat[eCh] = EIO;
			return XDMAC_NG;
		}
		
		if (ptDmaSet->cInvCache == XDMAC_ON) {
			if (ptDmaSet->eDstTm == tm_nrm) {
				// normal transfer
				purge_cache(ptDmaSet->uiDstAdr, (ptDmaSet->uiIts * ptDmaSet->uiTnum), PURGE_CACHE_D_INV);
				ddbg_printk(KERN_INFO "purge_cache 0x%x - 0x%x\n", 
						ptDmaSet->uiDstAdr, (ptDmaSet->uiDstAdr + (ptDmaSet->uiIts * ptDmaSet->uiTnum)));
			} else {
				// rectangular transfer
				if (ptDmaSet->uiSqs == 0) {
					/* critical section out */
					up(&tSemXdmac[eCh]);
					return XDMAC_NG;
				}
				uiHeight = (ptDmaSet->uiIts * ptDmaSet->uiTnum) /  ptDmaSet->uiSqs;
				uiAddr  = ptDmaSet->uiDstAdr;
				for (uiNum = 0; uiNum < uiHeight; uiNum++) {
					purge_cache(uiAddr, ptDmaSet->uiSqs, PURGE_CACHE_D_INV);
					ddbg_printk(KERN_INFO "purge_cache 0x%x - 0x%x\n", uiAddr, (uiAddr + ptDmaSet->uiSqs));
					uiAddr += ptDmaSet->uiSqs + ptDmaSet->uiSqad2d;
				}
			}
		}
	}

	eXdmacIrqStat[eCh] = irq_none; // clear IRQ

	/* critical section out */
	up(&tSemXdmac[eCh]);

	return XDMAC_OK;
}
EXPORT_SYMBOL(iXdmaStart);

/*
 * $Function : int iXdmaGet(EXdmaCh eCh, SDmaStat *ptDmaStat)
 * $Args     : [IN]  eCh        : logical channel number of XDMA
             : [OUT] *ptDmaStat : status parameters of XDMA
 * $Return   : XDMAC_OK         : succeeded
 *           : XDMAC_NG         : failed
 * $Desc     : get "eCh"th XDMAC status
 */
int iXdmaGet(EXdmaCh eCh, SDmaStat *ptDmaStat)
{
	if (eCh < 0 || eCh >= XDMAC_NUM_OF_CH) {
		return XDMAC_NG;
	}

	ptDmaStat->eStat = eXdmaStat[eCh];

	return XDMAC_OK;
}
EXPORT_SYMBOL(iXdmaGet);

/*
 * $Function : int iXdmaCfg(SDmaCfg *ptDmaCfg)
 * $Args     : [IN]  *tDmaCfg  : config parameters for DMAC
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : configure XDMAC
 */
int iXdmaCfg(SDmaCfg *ptDmaCfg)
{
	/* critical section in */
	if (down_interruptible(&tSemXdmacCmn)) {
		printk(KERN_ERR "%s: down_interruptible error\n", msg); // interrupt happened
		return XDMAC_NG;
	}

	ddbg_printk(KERN_INFO "/dev/xdmac CONFIG\n");

	if (ptDmaCfg->ePri == pri_fix) {
		iowrite32(XDMAC_PRI_PRI_FIX , (xdmac_base + XDMAC_PRI_OFFSET));
		ddbg_printk(KERN_INFO "xdma pri FIX \n");
	} else {
		iowrite32(XDMAC_PRI_PRI_RR , (xdmac_base + XDMAC_PRI_OFFSET));
		ddbg_printk(KERN_INFO "xdma pri Round Robin \n");
	}

	// key color data register
	iowrite32(ptDmaCfg->uiKcdata, (xdmac_base + XDMAC_KCDATA_OFFSET));

	/* critical section out */
	up(&tSemXdmacCmn);

	return XDMAC_OK;
}
EXPORT_SYMBOL(iXdmaCfg);

/*
 * $Function : int iXdmaDcinv(SDmaDstInfo *ptDmaDstInfo)
 * $Args     : [IN]  iCh       : logical channel number of XDMA
             : [IN]  *tDmaCfg  : config parameters for DMAC
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : set and start "iCh"th XDMAC
 */
int iXdmaDcinv(SDmaDstInfo *ptDmaDstInfo)
{
	unsigned int uiAddr, uiHeight, uiNum;

	if (ptDmaDstInfo->eDstTm == tm_nrm) {
		// normal transfer
		purge_cache(ptDmaDstInfo->uiDstAdr, ptDmaDstInfo->uiSize, PURGE_CACHE_D_INV);
		ddbg_printk(KERN_INFO "purge_cache 0x%x - 0x%x\n", 
				ptDmaDstInfo->uiDstAdr, (ptDmaDstInfo->uiDstAdr + ptDmaDstInfo->uiSize));
	} else {
		// rectangular transfer
		uiHeight = ptDmaDstInfo->uiSize /  ptDmaDstInfo->uiSqs;
		uiAddr   = ptDmaDstInfo->uiDstAdr;
		for (uiNum = 0; uiNum < uiHeight; uiNum++) {
			purge_cache(uiAddr, ptDmaDstInfo->uiSqs, PURGE_CACHE_D_INV);
			ddbg_printk(KERN_INFO "purge_cache 0x%x - 0x%x\n", uiAddr, (uiAddr + ptDmaDstInfo->uiSqs));
			uiAddr += ptDmaDstInfo->uiSqs + ptDmaDstInfo->uiSqad2d;
		}
	}

	return XDMAC_OK;
}
EXPORT_SYMBOL(iXdmaDcinv);

/*
 * $Function : int iXdmaRegRead(unsigned int uiAddr)
 * $Args     : [IN]  uiAddr    : register address
 * $Return   : register value  : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : set and start "iCh"th XDMAC
 */
int iXdmaRegRead(unsigned int uiAddr)
{
	unsigned int uiReg;

	uiReg = ioread32((uiAddr + (unsigned int)xdmac_base - XDMAC_BASE));

	return uiReg;
}
EXPORT_SYMBOL(iXdmaRegRead);


/*
 * $Function : int iXdmaRegWrite(unsigned int uiAddr, unsigned int uiData)
 * $Args     : [IN]  uiAddr    : register address
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : set and start "iCh"th XDMAC
 */
int iXdmaRegWrite(unsigned int uiAddr, unsigned int uiData)
{
	iowrite32(uiData, (uiAddr + (unsigned int)xdmac_base - XDMAC_BASE));

	return XDMAC_OK;
}
EXPORT_SYMBOL(iXdmaRegWrite);


/* ============================================================================ */
/*    local functions                                                           */
/* ============================================================================ */
/*
 * irq handler
 */
static irqreturn_t xdmac_irq_isr(int irq, void *dev_id)
{
	int iCh;
	unsigned int uiData, uiCmpFlg;
	EIrqStat      eIrqStat; // IRQ status
	EStatType     eStat;     // XDMA state

	ddbg_printk("XDMAC IRQ\n");

	// check XDMAIR to identify which DMA channel was completed
	uiData = ioread32((xdmac_base + XDMAC_MIR_OFFSET));

	for (iCh = 0; iCh < XDMAC_NUM_OF_CH; iCh++) {
		eStat = stat_trns;
		eIrqStat = irq_none;

		// XDMAIR.ENDF
		uiCmpFlg = 1 << iCh;
		if ((uiData & uiCmpFlg) == uiCmpFlg) {
			eStat    = stat_stop;
			eIrqStat = irq_cmp;
		}

		// XDMAIR.ERRF
		uiCmpFlg = 1 << (iCh + XDMAC_MIR_ERRF0);
		if ((uiData & uiCmpFlg) == uiCmpFlg) {
			eStat    = stat_err;
			eIrqStat = irq_err;
		}

		// wakeup thread and clear IRQ flags
		if (eIrqStat != irq_none) {
			iowrite32((XDMAC_IR_ERRF_CLR | XDMAC_IR_ENDF_CLR), (xdmac_base + XDMAC_IR_OFFSET + (iCh * XDMAC_CH_OFFSET)));
			eXdmaStat[iCh]     = eStat;
			eXdmacIrqStat[iCh] = eIrqStat;
			wake_up_interruptible(&tWqXdmaIrq[iCh]);
		}
	}

	return IRQ_HANDLED;
}

/*
 * $Function : static int iXdmaLsiInit(void)
 * $Args     : none
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : Initialize XDMAC
 */
static int iXdmaLsiInit(void)
{
	int iCh;

	/* stop all DMA channels */
	for (iCh = 0; iCh < XDMAC_NUM_OF_CH; iCh++) {
		// prohibit interrupts
		iowrite32((XDMAC_IEN_ERRIEN_DIS | XDMAC_IEN_ENDIEN_DIS),
		          (xdmac_base + XDMAC_IEN_OFFSET + (iCh * XDMAC_CH_OFFSET))); 

		// stop DMA
		iXdmaChStopPoll(iCh);

		// permit interrupts
		iowrite32((XDMAC_IEN_ERRIEN_EN | XDMAC_IEN_ENDIEN_EN),
		          (xdmac_base + XDMAC_IEN_OFFSET + (iCh * XDMAC_CH_OFFSET))); 
	}

	/* change priority scheme to fix-mode */
	iowrite32(XDMAC_PRI_PRI_FIX , (xdmac_base + XDMAC_PRI_OFFSET));

	return XDMAC_OK;
}

/*
 * $Function : static int iXdmaLsiFin(void)
 * $Args     : none
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : Finalize XDMAC
 */
static int iXdmaLsiFin(void)
{
	int iCh;

	/* stop all DMA channels */
	for (iCh = 0; iCh < XDMAC_NUM_OF_CH; iCh++) {
		// prohibit interrupts
		iowrite32((XDMAC_IEN_ERRIEN_DIS | XDMAC_IEN_ENDIEN_DIS),
		          (xdmac_base + XDMAC_IEN_OFFSET + (iCh * XDMAC_CH_OFFSET))); 

		// stop DMA
		iXdmaChStopPoll(iCh);

		// fail safe
//		wake_up_interruptible(&tWqXdmaIrq[iCh]);
	}

	/* change priority scheme to fix-mode */
	iowrite32(XDMAC_PRI_PRI_FIX , (xdmac_base + XDMAC_PRI_OFFSET));

	return XDMAC_OK;
}

/*
 * $Function : static int iXdmaChStopPoll(int iCh)
 * $Args     : [IN]  iCh       : logical channel number of XDMA
 * $Return   : XDMAC_OK        : succeeded
 *           : XDMAC_NG        : failed
 * $Desc     : Reset XDMAC. Be sure to prohibit IRQ before calling this function
 */
static int iXdmaChStopPoll(int iCh)
{
	/* stop DMA channel */
	iowrite32(XDMAC_TSS_TSS_STOP, (xdmac_base + XDMAC_TSS_OFFSET + (iCh * XDMAC_CH_OFFSET)));

	/* make sure the DMA channel has been stopped */
	while (1) {
		if (ioread32((xdmac_base + XDMAC_STAT_OFFSET + (iCh * XDMAC_CH_OFFSET))) == XDMAC_STAT_STOPF_STOP) {
			break;
		}
	}

	/* clear STOPF flag */
	iowrite32(XDMAC_STAT_STOPF_STOP, (xdmac_base + XDMAC_STAT_OFFSET + (iCh * XDMAC_CH_OFFSET)));

	/* clear IRQ flags */
	iowrite32((XDMAC_IR_ERRF_CLR | XDMAC_IR_ENDF_CLR), (xdmac_base + XDMAC_IR_OFFSET + (iCh * XDMAC_CH_OFFSET)));

	return XDMAC_OK;
}

MODULE_AUTHOR("Panasonic Industrial Device Company");
MODULE_DESCRIPTION("DMA device driver" XDMAC_VERSION);
MODULE_VERSION(XDMAC_VERSION);
MODULE_LICENSE("Proprietary");




