
#ifndef _ICCDRV_H_
#define _ICCDRV_H_

/* check receive fifo rotate */
#define ICC_RECEIVE_ROUND_ROBIN

/* A sleep time (tick) for when a fifo is filled. */
#define ICC_SLEEP_TIME			1

/* The local buffer size. It is used for a data receive. */
#define ICC_LOCALBUF_NUM		8

/* debug option */
/* #define ICC_DEBUG */
#undef ICC_DEBUG

#ifdef ICC_DEBUG
#define ICC_ABORT()						\
do {								\
	iccdrv_set_state(icc_cpu_self(), ICC_STATE_ABORT);	\
	while(1);						\
} while (0)

#define ICC_ASSERT(cond)	\
do {				\
	if (!(cond)) {		\
		ICC_ABORT();	\
	}			\
} while (0)
#else
#define ICC_ABORT()		do { } while (0)
#define ICC_ASSERT(cond)	do { } while (0)
#endif


#include "icc_def.h"
#include "icc_if.h"
#include "icc_hwd.h"
#include "linked_list.h"

typedef struct icc_local_data {
	struct linked_list list;
	icc_packetcb_t packet;
} icc_local_data_t;

#define ICC_LOCAL_DATA_MAX_LEN	\
	(ICC_MESSAGE_MAX_LEN + sizeof(icc_local_data_t))

static inline void *__iccdrv_io(unsigned long addr);

#if defined(__ICC_PVC01V__)
#if defined(ICC_ARM)
#include "iccdrv_pvc01v_arm.h"
#elif defined(ICC_AM32)
#include "iccdrv_pvc01v_am32.h"
#endif
#elif defined(__ICC_PVC02V__)
#include "iccdrv_pvc02v.h"
#endif

#if defined(ICC_LINUX)
#include "iccdrv_linux.h"
#elif defined(ICC_ITRON) || defined(ICC_OSLESS)
#include "iccdrv_itron.h"
#endif

static inline int iccdrv_set_idtable(unsigned short id)
{
	icc_basecb_t *cb;
	uint32_t bit;
	int cpu, i;

	cb = (icc_basecb_t *)__iccdrv_io(ICC_CB_TOP_ADDR);

	cpu = icc_id_owner(id);
	if (cpu < 0 || cpu >= ICC_CPU_NUM) {
		return ICC_E_PAR;
	}

	i = id & ~ICC_IDBASE_MASK;
	bit = 1 << (i % 32);

	cb->idtable[cpu][i / 32] |= bit;

	return 0;
}

static inline int iccdrv_clr_idtable(unsigned short id)
{
	icc_basecb_t *cb;
	uint32_t bit;
	int cpu, i;

	cb = (icc_basecb_t *)__iccdrv_io(ICC_CB_TOP_ADDR);

	cpu = icc_id_owner(id);
	if (cpu < 0 || cpu >= ICC_CPU_NUM) {
		return ICC_E_PAR;
	}

	i = id & ~ICC_IDBASE_MASK;
	bit = 1 << (i % 32);

	cb->idtable[cpu][i / 32] &= ~bit;
	return 0;
}

static inline int iccdrv_check_idtable(uint16_t id)
{
	icc_basecb_t *cb;
	uint32_t bit;
	int cpu, i;
	int ret;

	cb = (icc_basecb_t *)__iccdrv_io(ICC_CB_TOP_ADDR);

	cpu = icc_id_owner(id);
	if (cpu < 0 || cpu >= ICC_CPU_NUM) {
		return ICC_E_PAR;
	}

	i = id & ~ICC_IDBASE_MASK;
	bit = 1 << (i % 32);

	ret = (cb->idtable[cpu][i / 32] & bit)	\
		? 0 : ICC_E_NOLINK;

	return ret;
}

static inline int iccdrv_set_state(int cpu, uint8_t state)
{
	icc_basecb_t *cb;

	cb = (icc_basecb_t *)__iccdrv_io(ICC_CB_TOP_ADDR);
	cb->state[cpu] = state;

	return 0;
}

extern int iccdrv_receive_data(int irq);
extern int iccdrv_do_receive_data(icc_local_data_t *buf);
extern int iccdrv_read_wait(unsigned short rcvid, icc_time_t timeout);
extern int iccdrv_sleep(icc_time_t time);

extern void* icc_localbuf_alloc(void);
extern int icc_localbuf_free(void* p);
extern void icc_localbuf_add_list(icc_local_data_t *buf);
extern void icc_localbuf_remove_list(icc_local_data_t *buf);
extern icc_local_data_t *icc_get_local_data(unsigned short rcvid);

#endif /* _ICCDRV_H_ */
