
/*****************************************************************************
 * Inter CPU Communication Driver
 *****************************************************************************/
#include "iccdrv.h"

/*
 * For access to the FIFO control block member.
 *
 *  icc_fifocb_t
 *   oooo: length
 *   oo--: roffset
 *   --oo: woffset
 *   oooo:  icc_fifo_start() --
 *       :   |                |--> icc_fifo_offset()
 *       :   |              p--
 *       :   |
 *   oooo:   |--> icc_fifo_length()
 *          icc_fifo_end()
 */
static inline void *icc_fifo_start(const icc_fifocb_t *fifo)
{
	return (void *)((unsigned char*)fifo + sizeof(icc_fifocb_t));
}

static inline void *icc_fifo_end(const icc_fifocb_t *fifo)
{
	return (void*)((unsigned char*)fifo + fifo->length);
}
static inline int icc_fifo_length(const icc_fifocb_t *fifo)
{
	return fifo->length - sizeof(icc_fifocb_t);
}

static inline void *icc_fifo_raddr(const icc_fifocb_t *fifo)
{
	return (void*)((unsigned char*)icc_fifo_start(fifo) + fifo->roffset);
}

static inline void *icc_fifo_waddr(const icc_fifocb_t *fifo)
{
	return (void*)((unsigned char*)icc_fifo_start(fifo) + fifo->woffset);
}

static inline int icc_fifo_offset(const icc_fifocb_t *fifo, const void *p)
{
	return (int)((unsigned char*)p - (unsigned char*)icc_fifo_start(fifo));
}

static void* icc_fifo_add_addr(const icc_fifocb_t *fifo, const void *base, int len)
{
	void *p = (void*)((unsigned char*)base + len);
	if (p < icc_fifo_end(fifo)) {
		return p;
	}
	return (void*)((unsigned char*)p - icc_fifo_length(fifo));
}

/*
 * When a "pos" is the range of "start" to "end", return 1
 *    s     e
 *  p |-----| p     return 0;
 *    p--p--p       return 1;
 *
 *    e     s
 * ---|  p  |---    return 0;
 * -p-p     p-p-    return 1;
*/
static inline int icc_check_range(void *start, void *end,
				  void *pos)
{
	if ((start <= end) && ((end < pos) || (pos < start))) {
		return 0;
	} else if ((end < pos) && (pos < start)) {
		return 0;
	}
	return 1;
}

static void *icc_write_fifo(const icc_fifocb_t *fifo, void *dst, const void *src, int length)
{
	void *next;
	int tmplen;

#ifdef ICC_DEBUG
	ICC_ASSERT(length >= 0);
#endif
	if (length <= 0) {
		return (void*)dst;
	}
	next = icc_fifo_add_addr(fifo, dst, length);
	if ((dst < next) || (next == icc_fifo_start(fifo))) {
		if (copy_to_iccfifo(dst, src, length)) {
			return NULL;
		}
	} else {
		tmplen = (int)((unsigned char*)icc_fifo_end(fifo) - (unsigned char*)dst);
		if (copy_to_iccfifo(dst, src, tmplen)) {
			return NULL;
		}
		if (copy_to_iccfifo(icc_fifo_start(fifo), (void*)((unsigned char*)src + tmplen), length - tmplen)) {
			return NULL;
		}
	}
	return next;
}

static void *icc_read_fifo(const icc_fifocb_t *fifo, void *dst, const void *src, int length)
{
	void *next;
	int tmplen;

#ifdef ICC_DEBUG
	ICC_ASSERT(length >= 0);
#endif
	if (length <= 0) {
		return (void*)src;
	}
	next = icc_fifo_add_addr(fifo, src, length);
	if ((src < next) || (next == icc_fifo_start(fifo))) {
		if (copy_from_iccfifo(dst, src, length)) {
			return NULL;
		}
	} else {
		tmplen = (int)((unsigned char*)icc_fifo_end(fifo) - (unsigned char*)src);
		if (copy_from_iccfifo(dst, src, tmplen)) {
			return NULL;
		}
		if (copy_from_iccfifo((void*)((unsigned char*)dst + tmplen), icc_fifo_start(fifo), length - tmplen)) {
			return NULL;
		}
	}
	return next;
}

/*
 * The functions for a data send to FIFO.
 */
static int icc_write_packet(icc_fifocb_t *fifo, icc_packetcb_t *packet)
{
	void *p, *next;

	p = icc_fifo_waddr(fifo);
	next = icc_fifo_add_addr(fifo, p, sizeof(icc_packetcb_t) + packet->length);
	if (icc_check_range(icc_fifo_raddr(fifo), p, next)) {
		return ICC_E_BUSY;
	}

	if ((p = icc_write_fifo(fifo, p, (void*)packet,
				sizeof(icc_packetcb_t) + packet->length)) == NULL) {
		return ICC_E_FAULT;
	}

#ifdef ICC_DEBUG
	ICC_ASSERT(next == p);
	ICC_ASSERT(icc_fifo_offset(fifo, p) >= 0);
#endif
	fifo->woffset = (uint16_t)icc_fifo_offset(fifo, p);

	return packet->length;
}

int icc_core_write(icc_packetcb_t *packet, int timeout)
{
	struct iccdrv_info_struct *iccdrv = current_iccdrv();
	icc_fifocb_t* fifo;
	uint16_t dstid;
	uint16_t length;
	int lefttime;
	int retval;
	unsigned long flags;

	dstid = packet->dstid;
	length = packet->length;

	if (icc_check_id_range(dstid)) {
		retval = ICC_E_PAR;
		goto out;
	}
	if ((timeout < 0) && (timeout != ICC_FOREVER)) { /* pgr1223 */
		retval = ICC_E_PAR;
		goto out;
	}
#if ICC_MESSAGE_PENDING_MODE == 0
	if (iccdrv_check_idtable(dstid)) {
		retval = ICC_E_NOLINK;
		goto out;
	}
#endif
	fifo = icc_get_sendfifo(icc_id_owner(dstid));
	if (fifo == NULL) {
		return ICC_E_FAULT;
	}
	/*
	 * If a packet length is not 4 byte multiple, round up.
	 */
	if (length & 0x3) {
		length = (length + 0x3) & ~0x3;
		packet->length = length;
	}

	if (length > (fifo->length - sizeof(icc_fifocb_t) - sizeof(uint32_t))) {
		return ICC_E_PAR;
	}
	lefttime = timeout;
	icc_local_lock(&iccdrv->fifolock, &flags);
	for (;;) {
		retval = icc_write_packet(fifo, packet);
		if (retval >= 0) {
			/* write succeed. retval is write size */
			icc_raise_irq(dstid);
			break;
		} else if (retval != ICC_E_BUSY) {
			/* retval is ICC_E_FAULT */
			break;
		}
		/* retval is ICC_E_BUSY */
		if (timeout == 0) {
			retval = ICC_E_NOMEM;
			break;
		} else if (timeout != ICC_FOREVER) {
			if (lefttime <= 0) {
				retval = ICC_E_NOMEM;
				break;
			}
			lefttime -= icc_time_to_msecs(ICC_SLEEP_TIME);
		}
		icc_local_unlock(&iccdrv->fifolock, &flags);
		iccdrv_sleep(ICC_SLEEP_TIME);
		icc_local_lock(&iccdrv->fifolock, &flags);
		/* retry write packet until succeed, or timeout */
	}
	icc_local_unlock(&iccdrv->fifolock, &flags);
out:
	return retval;
}


/*
 * Local buffers for store receive data.
 */

static inline int icc_localbuf_is_used(int i)
{
	return current_iccdrv()->bufmap[i / 32] & (1 << (i % 32));
}

static inline void icc_localbuf_setbit(int i)
{
	current_iccdrv()->bufmap[i / 32] |= (1 << (i % 32));
}

static inline void icc_localbuf_clrbit(int i)
{
	current_iccdrv()->bufmap[i / 32] &= ~(1 << (i % 32));
}

void* icc_localbuf_alloc(void)
{
	struct iccdrv_info_struct *iccdrv = current_iccdrv();
	int i;
	unsigned long flags;
	void *p = NULL;

	icc_local_lock(&iccdrv->buflock, &flags);
	for (i = 0; i < ICC_LOCALBUF_NUM; i++) {
		if (!icc_localbuf_is_used(i)) {
			icc_localbuf_setbit(i);
			p = &iccdrv->localbuf[i][0];
			break;
		}
	}
	if (i == ICC_LOCALBUF_NUM) {
		/* All buffer is used */
		iccdrv->bufflag |= ICCDRV_BUF_OVERFLOW;
	}
	icc_local_unlock(&iccdrv->buflock, &flags);
	return p;
}

int icc_localbuf_free(void* p)
{
	struct iccdrv_info_struct *iccdrv = current_iccdrv();
	unsigned long flags;
	unsigned long offset;
	int i;
	int retval = 0;

	icc_local_lock(&iccdrv->buflock, &flags);
	offset = (unsigned long)p - (unsigned long)&iccdrv->localbuf[0][0];
	i = offset / (ICC_MESSAGE_MAX_LEN + sizeof(icc_local_data_t));
#ifdef ICC_DEBUG
	if (offset % (ICC_MESSAGE_MAX_LEN + sizeof(icc_local_data_t))) {
		ICC_ABORT();
	}
#endif
	if (i < 0 || i >= ICC_LOCALBUF_NUM) {
		retval = ICC_E_PAR;
		goto out;
	}
	icc_localbuf_clrbit(i);
	if (iccdrv->bufflag & ICCDRV_BUF_OVERFLOW) {
		iccdrv->bufflag &= ~ICCDRV_BUF_OVERFLOW;
		icc_raise_irq_self();
	}
out:
	icc_local_unlock(&iccdrv->buflock, &flags);
	return retval;
}

void icc_localbuf_add_list(icc_local_data_t *buf)
{
	struct iccdrv_info_struct *iccdrv = current_iccdrv();
	unsigned long flags;
	icc_local_lock(&iccdrv->buflock, &flags);
	add_list(&iccdrv->read_list, &buf->list);
	icc_local_unlock(&iccdrv->buflock, &flags);
}

void icc_localbuf_remove_list(icc_local_data_t *buf)
{
	struct iccdrv_info_struct *iccdrv = current_iccdrv();
	unsigned long flags;
	icc_local_lock(&iccdrv->buflock, &flags);
	remove_list(&buf->list);
	icc_local_unlock(&iccdrv->buflock, &flags);
}

/*
 * The functions for a data receive from FIFO.
 */

static icc_fifocb_t *icc_get_active_fifo(void)
{
	icc_fifocb_t *fifo;
	int i;
	int current_cpu = icc_cpu_self();
	int from_cpu;

	for (i = 0; i < ICC_CPU_NUM; i++) {
#if defined(ICC_RECEIVE_ROUND_ROBIN)
		from_cpu = current_iccdrv()->last_fifo_idx + 1;
		if (from_cpu >= ICC_CPU_NUM) {
			from_cpu = 0;
		}
		current_iccdrv()->last_fifo_idx = from_cpu;
#else
		from_cpu = i;
#endif
		if (from_cpu == current_cpu) {
			continue;
		}
		fifo = icc_get_receivefifo(from_cpu);
		if (fifo && (icc_fifo_raddr(fifo) != icc_fifo_waddr(fifo))) {
			return fifo;
		}
	}
	return NULL;
}

int iccdrv_receive_data(int arg)
{
	struct iccdrv_info_struct *iccdrv = current_iccdrv();
	uint16_t dstid;
	uint16_t length;
	icc_fifocb_t *fifo;
	void *p;
	icc_local_data_t *buf;
	int err;
	struct linked_list *list, *next;

	while ((fifo = icc_get_active_fifo()) != NULL) {
		p = icc_fifo_raddr(fifo);
		if ((p = icc_read_fifo(fifo, &dstid, p, sizeof(dstid))) == NULL) {
#ifdef ICC_DEBUG
			ICC_ABORT();
#endif
			return ICC_E_FAULT;
		}
		if ((p = icc_read_fifo(fifo, &length, p, sizeof(length))) == NULL) {
#ifdef ICC_DEBUG
			ICC_ABORT();
#endif
			return ICC_E_FAULT;
		}
		buf = icc_localbuf_alloc();
		if (buf == NULL) {
			/*
			   When it can not allocate a local buffer,
			   the receive data left in FIFO.
			 */
			break;
		}
		buf->packet.dstid = dstid;
		buf->packet.length = length;
		if ((p = icc_read_fifo(fifo, ICC_PACKET_MESSAGE(&buf->packet), p, length)) == NULL) {
#ifdef ICC_DEBUG
			ICC_ABORT();
#endif
			icc_localbuf_free(buf);
			return ICC_E_FAULT;
		}
		icc_localbuf_add_list(buf);
		fifo->roffset = (uint16_t)icc_fifo_offset(fifo, p);

		/*
		  Clear the interrupt request with each cycle.
		  Because when this sequence receive a number of data,
		  may be left the interrupt requests.
		*/
		icc_clear_irq();
	}

	for_each_list(&iccdrv->read_list, list, next) {
		buf = (icc_local_data_t *)list;
		err = iccdrv_do_receive_data(buf);
		if (err) {
			/*
			   When the device is not open or the callback is not set
			   in not pending mode, free the receive data.
			*/
			icc_localbuf_remove_list(buf);
			icc_localbuf_free(buf);
		}
	}
	return 0;
}

icc_local_data_t *icc_get_local_data(unsigned short rcvid)
{
	struct iccdrv_info_struct *iccdrv = current_iccdrv();
	struct linked_list *list, *next;
	icc_local_data_t *tmp;

	for_each_list(&iccdrv->read_list, list, next) {
		tmp = (icc_local_data_t *)list;
		if (rcvid == tmp->packet.dstid) {
			return tmp;
		}
	}
	return NULL;
}

#if defined(ICC_LINUX)
int icc_core_read(unsigned short rcvid, void *buf,
		  int len, int timeout)
{
	icc_local_data_t *data = NULL;
	int err;
	int readlen;

	if (timeout < 0 && timeout != ICC_FOREVER) {
		err = ICC_E_PAR;
		goto out;
	}

	data = icc_get_local_data(rcvid);
	if (data == NULL) {
		if (timeout == 0) {
			err = ICC_E_NODATA;
			goto out;
		}
		if (timeout == ICC_FOREVER) {
			err = iccdrv_read_wait(rcvid, (icc_time_t)ICC_FOREVER);
		} else {
			err = iccdrv_read_wait(rcvid, msecs_to_icc_time(timeout));
		}
		if (err) {
			goto out;
		}
		data = icc_get_local_data(rcvid);
		if (data == NULL) {
#ifdef ICC_DEBUG
			ICC_ABORT();
#endif
			err = ICC_E_FAULT;
			goto out;
		}
	}
	icc_localbuf_remove_list(data);

	/*
	 * If a 'len' is shorter than a receive data length,
	 * the data is copied to the user buffer at a len size.
	 */
	readlen = (len < data->packet.length) ? len : data->packet.length;
	/*
	 * If a read length is not 4 byte multiple, round down.
	 */
	if (readlen & 0x3) {
		readlen &= ~0x3;
	}
	if (readlen > 0) {
		err = copy_local_to_user(buf, ICC_PACKET_MESSAGE(&data->packet), readlen);
	}
	icc_localbuf_free(data);

	if (err) {
		err = ICC_E_PAR;
		goto out;
	}
	return readlen;
out:
	return err;
}
#endif /* ICC_LINUX */
