
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <linux/mm.h>

#include <asm/io.h>

#include "icc_if.h"
#include "iccdrv.h"

#define	ICC_MAJOR	123
#define DRIVER_NAME	"icc"

struct iccdrv_info_struct iccdrv_info[ICC_DRIVER_NUM];
struct iccdrv_dev_struct iccdrv_device[ICC_MINOR_NUM];

int icc_major = ICC_MAJOR;
module_param(icc_major, int, S_IRUGO);

struct iccdrv_iomap_struct iccdrv_iomap[] = ICCDRV_IOMAP_REGION;

const int iccdrv_recv_irq[] = ICCDRV_RECV_IRQ;

#ifdef CONFIG_PROC_FS
extern int icc_init_proc(void);
extern int icc_exit_proc(void);
#endif

/*
 * interrupt handler for message receive
 */
static irqreturn_t icc_interrupt(int irq, void *dev_id)
{
	iccdrv_receive_data(irq);

	return IRQ_HANDLED;
}

/*
 * The file operation functions
 */
static int iccdrv_open(struct inode *inode, struct file *filp)
{
	struct iccdrv_dev_struct *device;
	int ret = 0;
	int minor;

	device = container_of(inode->i_cdev, struct iccdrv_dev_struct, cdev);
	minor = MINOR(device->cdev.dev);

	if (minor != 0) { /* minor 0 is device for system. */
		if (device->status & ICCDRV_STAT_OPEN) {
			/* already open this device */
			return -EBUSY;
		}
		device->status |= ICCDRV_STAT_OPEN;
		iccdrv_set_idtable(ICC_ID(icc_cpu_self(), minor));
	}
	filp->private_data = device;
	return ret;
}

static int iccdrv_release(struct inode *inode, struct file *filp)
{
	struct iccdrv_dev_struct *device;
	int minor;

	device = filp->private_data;

	if (device) {
		minor = MINOR(device->cdev.dev);
		if (minor != 0) {
			iccdrv_clr_idtable(ICC_ID(icc_cpu_self(), minor));
			device->status = 0;
		}
		filp->private_data = 0;
	}
	return 0;
}

static ssize_t iccdrv_read(struct file * filp, char __user * buf,
			    size_t count, loff_t *ppos)
{
	struct iccdrv_dev_struct *device = filp->private_data;
	int minor;
	unsigned short rcvid;
	ssize_t retval;

	minor = MINOR(device->cdev.dev);
	rcvid = ICC_ID(icc_cpu_self(), minor);
#ifdef ICC_DEBUG
	if (!icc_is_self_id(rcvid)) {
		return -EINVAL;
	}
	if (icc_check_id_range(rcvid)) {
		return -EINVAL;
	}
#endif
	if (!access_ok(VERIFY_WRITE, buf, count)) {
		return -EFAULT;
	}

	if (!mutex_trylock(&device->read_mutex)) {
		return -EBUSY;
	}
	retval = icc_core_read(rcvid, buf, count, device->read_tmo);
	if (retval < 0) {
		switch (retval) {
		case ICC_E_PAR:
			retval = -EINVAL;
			break;
		case ICC_E_NODATA:
			retval = -ENODATA;
			break;
		case ICC_E_NOLINK:
			retval = -ENOLINK;
			break;
		default: /* ICC_E_FAULT */
			retval = -EIO;
			break;
		}
	}
	mutex_unlock(&device->read_mutex);
	return retval;
}

static ssize_t iccdrv_write(struct file * filp, const char __user * buf,
			     size_t count, loff_t *ppos)
{
	struct iccdrv_dev_struct *device = filp->private_data;
	ssize_t retval;
	icc_packetcb_t packet;

	if (!access_ok(VERIFY_READ, buf, count)) {
		return -EFAULT;
	}
	if (copy_from_user(&packet, (int __user *)buf, sizeof(packet))) {
		return -EFAULT;
	}
	if (icc_is_self_id(packet.dstid)) {
		return -EINVAL;
	}
	retval = icc_core_write((icc_packetcb_t *)buf, device->write_tmo);
	if (retval < 0) {
		switch (retval) {
		case ICC_E_PAR:
			retval = -EINVAL;
			break;
		case ICC_E_NOMEM:
			retval = -ENOMEM;
			break;
		case ICC_E_NOLINK:
			retval = -ENOLINK;
			break;
		default: /* ICC_E_FAULT */
			retval = -EIO;
			break;
		}
	}
	return retval;
}

// add 2014.08.04 AVCMMS --->
/*
 *
 */
static int iccdrv_linux_sleep( unsigned int flagadr )
{
	unsigned int offset  = flagadr & ~PAGE_MASK;
	unsigned char* pbuff_base;
	unsigned int* pbuff;
			
	/* }bsO */
	pbuff_base = (unsigned*)ioremap_nocache( flagadr - offset, offset + 4 );
	if ( !pbuff_base ) {
		return -EINVAL;
	}
	pbuff = pbuff_base + offset;

	*pbuff = 0x11111111;

	/* A}bv */
	iounmap(pbuff_base);
		
	/* 荞݋֎~ */
	local_irq_disable();
	
	/* LinuxX[v(AȂ) */
	while(1) {
		__asm__ __volatile__(
		"nop	\n"
		"wfe	\n"
		);
	}
}
// <--- add 2014.08.04 AVCMMS

static int iccdrv_ioctl(struct inode *inode, struct file *filp,
			unsigned int cmd, unsigned long a)
{
	struct iccdrv_dev_struct *device = filp->private_data;
	int arg;
	icc_ctrl_t ctrl;

	switch (cmd) {
	case ICC_IOC_SET_READ_TMOUT:
		if (get_user(arg, (int *)a)) {
			return -EFAULT;
		}
		device->read_tmo = arg;
		break;
	case ICC_IOC_SET_WRITE_TMOUT:
		if (get_user(arg, (int *)a)) {
			return -EFAULT;
		}
		device->write_tmo = arg;
		break;
	case ICC_IOC_IS_SELF_ID:
		if (copy_from_user(&ctrl, (int __user *)a, sizeof(ctrl))) {
			return -EFAULT;
		}
		ctrl.flags = (unsigned int)icc_is_self_id(ctrl.id);
		if (copy_to_user((int __user *)a, &ctrl, sizeof(ctrl))) {
			return -EFAULT;
		}
		break;
// add 2014.08.04 AVCMMS --->
	case ICC_IOC_LINUX_SLEEP:
		{
			unsigned int flagadr = 0;
			if (get_user(flagadr, (int *)a)) {
				return -EFAULT;
			}
			iccdrv_linux_sleep( flagadr );
		}
		break;
// <--- add 2014.08.04 AVCMMS
	default:
		return -ENOTTY;
	}

	return 0;
}

static unsigned int iccdrv_poll(struct file *filp, struct poll_table_struct *wait)
{
	struct iccdrv_dev_struct *device = filp->private_data;
	int minor;
	unsigned short rcvid;
	int status = 0;

	minor = MINOR(device->cdev.dev);
	rcvid = ICC_ID(icc_cpu_self(), minor);

	poll_wait(filp, &device->readq, wait);

	if (icc_get_local_data(rcvid)) {
		status |= POLLIN;
	}

	return status;
}

const struct file_operations iccdrv_fops =
{
	.open		= iccdrv_open,
//	.flush		= iccdrv_flush,
	.release	= iccdrv_release,
	.read		= iccdrv_read,
	.write		= iccdrv_write,
	.ioctl		= iccdrv_ioctl,
	.poll		= iccdrv_poll,
};

/*
 * The driver init and exit functions
 */
static int __init icc_init_driver(void)
{
	struct iccdrv_dev_struct *device;
	dev_t dev = 0;
	int alloc_ret = -1;
	int cdev_ret = 0;
	int irq_ret = -1;
	int cnum = 0;
	int inum = 0;
	int i;
	int retval = 0;
#ifdef CONFIG_PROC_FS
	int proc_ret = -1;
#endif

	if (icc_major) {
		dev = MKDEV(icc_major, 0);
		alloc_ret = register_chrdev_region(dev, ICC_MINOR_NUM, DRIVER_NAME);
	} else {
		alloc_ret = alloc_chrdev_region(&dev, 0, ICC_MINOR_NUM, DRIVER_NAME);
		icc_major = MAJOR(dev);
	}
	if (alloc_ret) {
		printk(KERN_WARNING \
		       "iccdrv: register_chrdev_region failed %d\n", alloc_ret);
		retval = alloc_ret;
		goto init_error;
	}

	/* allocate device */
	for (cnum = 0; cnum < ICC_MINOR_NUM; cnum++) {
		device = &iccdrv_device[cnum];
		device->status = 0;
		device->read_tmo = 0;
		device->write_tmo = 0;

		init_waitqueue_head(&device->readq);
		mutex_init(&device->read_mutex);

		cdev_init(&device->cdev, &iccdrv_fops);
		device->cdev.owner = THIS_MODULE;
		device->cdev.ops = &iccdrv_fops;
		cdev_ret = cdev_add(&device->cdev, MKDEV(icc_major, cnum), 1);
		if (cdev_ret) {
			printk("icc_init_driver: cdev_add failed %d\n", cdev_ret);
			retval = cdev_ret;
			goto init_error;
		}
	}

	/* mapping ssc and irq controller register region */
	for (i = 0; iccdrv_iomap[i].size > 0; i++) {
		iccdrv_iomap[i].vaddr = \
			(unsigned long)ioremap_nocache(iccdrv_iomap[i].paddr, iccdrv_iomap[i].size);
		if (!iccdrv_iomap[i].vaddr) {
			printk("icc_init_driver: ioremap failed.\n");
			retval = -ENOMEM;
			goto init_error;
		}
	}

	iccdrv_set_state(icc_cpu_self(), ICC_STATE_START);

	/* setup iccdrv_info */
	init_list(&current_iccdrv()->read_list);

	spin_lock_init(&current_iccdrv()->buflock);
	spin_lock_init(&current_iccdrv()->fifolock);

	/* setup interrupt */
	for (inum = 0; iccdrv_recv_irq[inum] >= 0; inum++) {
		irq_ret = request_irq(iccdrv_recv_irq[inum], icc_interrupt,
				      IRQF_DISABLED, "iccdrv interrupt", NULL);
		if (irq_ret) {
			printk("icc_init_driver: request_irq failed %d\n", irq_ret);
			retval = irq_ret;
			goto init_error;
		}
	}

	iccdrv_set_state(icc_cpu_self(), ICC_STATE_RUN);

#ifdef CONFIG_PROC_FS
	proc_ret = icc_init_proc();
	if (proc_ret) {
		printk("icc_init_driver: icc_init_proc failed %d\n", proc_ret);
		goto init_error;
	}
#endif
	printk(KERN_ALERT "%s driver version %d.%d.%d (major %d)\n",
	       DRIVER_NAME, ICC_VER_MAJOR, ICC_VER_MINOR, ICC_VER_RELEASE, icc_major);

	return retval;

init_error:
	for (i = 0; i < inum; i++) {
		free_irq(iccdrv_recv_irq[i], NULL);
	}
	for (i = 0; iccdrv_iomap[i].size > 0; i++) {
		if (iccdrv_iomap[i].vaddr) {
			iounmap((void*)iccdrv_iomap[i].vaddr);
		}
	}
	for (i = 0; i < cnum; i++) {
		cdev_del(&iccdrv_device[i].cdev);
	}
	if (!alloc_ret) {
		unregister_chrdev_region(dev, ICC_MINOR_NUM);
	}
	return retval;
}

static void __exit icc_exit_driver(void)
{
	int i;

	for (i = 0; iccdrv_recv_irq[i] >= 0; i++) {
		free_irq(iccdrv_recv_irq[i], NULL);
	}

	iccdrv_set_state(icc_cpu_self(), ICC_STATE_STOP);

	for (i = 0; i < ICC_MINOR_NUM; i++) {
		cdev_del(&iccdrv_device[i].cdev);
	}

	unregister_chrdev_region(MKDEV(icc_major, 0), ICC_MINOR_NUM);

#ifdef CONFIG_PROC_FS
	icc_exit_proc();
#endif
	for (i = 0; iccdrv_iomap[i].size > 0; i++) {
		if (iccdrv_iomap[i].vaddr) {
			iounmap((void*)iccdrv_iomap[i].vaddr);
		}
	}

	printk(KERN_NOTICE "icc_exit_driver\n");

	return;
}

MODULE_AUTHOR("Panasonic");
MODULE_DESCRIPTION("Inter CPU Communication driver");
MODULE_LICENSE("Proprietary");

module_init(icc_init_driver);
module_exit(icc_exit_driver);


/*
 * The following functions are called from iccdrv_core.c
 */
int iccdrv_do_receive_data(icc_local_data_t *buf)
{
	struct iccdrv_dev_struct *device;
	unsigned short dstid;

	dstid = buf->packet.dstid;
	device = &iccdrv_device[dstid & ~ICC_IDBASE_MASK];

#if ICC_MESSAGE_PENDING_MODE == 0
	if (!(device->status & ICCDRV_STAT_OPEN)) {
		return ICC_E_NOLINK;
	}
#endif
	wake_up_interruptible(&device->readq);

	return 0;
}

int iccdrv_read_wait(unsigned short rcvid, icc_time_t timeout)
{
	struct iccdrv_dev_struct *device;
	int err;
	device = &iccdrv_device[rcvid & ~ICC_IDBASE_MASK];

	if (timeout == (icc_time_t)ICC_FOREVER) {
		err = wait_event_interruptible(
			device->readq, (icc_get_local_data(rcvid) != NULL));
	} else {
		err = wait_event_interruptible_timeout(
			device->readq, (icc_get_local_data(rcvid) != NULL), timeout);
		if (err == 0) {
			/* The data had not arrived until timeout. */
			return ICC_E_NODATA;
		}
	}
	if (err < 0) { /* signal */
		return ICC_E_FAULT;
	}
	return 0;
}
