
#ifndef _ICCDRV_LINUX_H_
#define _ICCDRV_LINUX_H_

#include <linux/cdev.h>
#include <linux/string.h>
#include <linux/jiffies.h>
#include <linux/sched.h>
#include <linux/mutex.h>

#include <asm/uaccess.h>

#include "icc_cfg.h"

struct iccdrv_info_struct {
	struct linked_list read_list;	/* list head for read message */

	/* Local buffers for stroe receive data. */
	uint8_t  localbuf[ICC_LOCALBUF_NUM][ICC_MESSAGE_MAX_LEN + sizeof(icc_local_data_t)];
	uint32_t bufmap[(ICC_LOCALBUF_NUM - 1) / 32 + 1];
	unsigned int bufflag;
#define ICCDRV_BUF_OVERFLOW	0x00000001

#ifdef ICC_RECEIVE_ROUND_ROBIN
	int last_fifo_idx;		/* last receive fifo index */
#endif
	/* For Linux */
	spinlock_t buflock;			 /* spinlock for local buffer */
	spinlock_t fifolock;			 /* spinlock for write to fifo  */
};
extern struct iccdrv_info_struct iccdrv_info[];

struct iccdrv_iomap_struct {
	unsigned long vaddr;	/* virtual (mapping) address */
	unsigned long paddr;	/* physical address  */
	int size;		/* size */
};

struct iccdrv_dev_struct {
	unsigned int status;		/* driver status */
	int read_tmo;			/* read timeout (msec) */
	int write_tmo;			/* write timeout (msec) */

	/* For Linux */
	struct cdev cdev;		/* character device struct */
	wait_queue_head_t readq;	/* wait queue for read */
	struct mutex read_mutex;	/* mutex for read */
};

/* driver status */
#define ICCDRV_STAT_OPEN	0x00000001

static inline int copy_to_iccfifo(void *dst, const void *src, int len)
{
	return copy_from_user(dst, src, len);
}

static inline int copy_from_iccfifo(void *dst, const void *src, int len)
{
	memcpy(dst, src, len);
	return 0;
}

static inline int copy_local_to_user(void *dst, const void *src, int len)
{
	return copy_to_user(dst, src, len);
}

/* timeout */
typedef unsigned long icc_time_t;

static inline icc_time_t msecs_to_icc_time(unsigned int msec)
{
	return msecs_to_jiffies(msec);
}

static inline unsigned int icc_time_to_msecs(icc_time_t time)
{
	return jiffies_to_msecs(time);
}

static inline int iccdrv_sleep(icc_time_t time)
{
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(time);
	return 0;
}

/* The mutual exclution among interrupts and tasks */
static inline void icc_local_lock(spinlock_t *lock, unsigned long *flags) {
	unsigned long tmp;
	spin_lock_irqsave(lock, tmp);
	*flags = tmp;
}

static inline void icc_local_unlock(spinlock_t *lock, unsigned long *flags) {
	unsigned long tmp = *flags;
	spin_unlock_irqrestore(lock, tmp);
}

static inline int icc_is_self_id(unsigned short id)
{
	if (icc_id_owner(id) == icc_cpu_self()) {
		return 1;
	}
	return 0;
}

extern struct iccdrv_iomap_struct iccdrv_iomap[];
static inline void __iomem *__iccdrv_io(unsigned long addr)
{
	int i;

	for (i = 0; iccdrv_iomap[i].size > 0; i++) {
		if (iccdrv_iomap[i].paddr <= addr &&
		    addr < iccdrv_iomap[i].paddr + iccdrv_iomap[i].size) {
			return (void __iomem*)(addr - iccdrv_iomap[i].paddr \
					       + iccdrv_iomap[i].vaddr);
		}
	}

	printk(KERN_ERR "__iccdrv_io: Unmapped address(%08lx)\n", addr);
	BUG();
}

extern int icc_core_write(icc_packetcb_t *packet, int timeout);
extern int icc_core_read(unsigned short rcvid, void *buf,
			 int len, int timeout);

#endif /* _ICCDRV_LINUX_H_ */
