
/*****************************************************************************
 * Inter CPU Shared Memory Driver
 *****************************************************************************/
#include "icsm_cfg.h"
#include "icsm_def.h"
#include "icsmdrv.h"

static inline int icsm_get_blkcb_mapnum(int blknum)
{
	ICSM_ASSERT(blknum > 0);
	return ((blknum - 1) / 32) + 1;
}

static icsm_blkcb_t *icsm_get_blkcb(unsigned short memid)
{
	icsm_blkcb_t *blkcb;
	unsigned short i;

	blkcb = (icsm_blkcb_t *)__icsmdrv_io(ICSM_CB_TOP_ADDR + sizeof(icsm_basecb_t));
	for (i = 0; i < memid; i++) {
		blkcb = (icsm_blkcb_t *)((uint32_t)blkcb +
					 sizeof(icsm_blkcb_t) +
					 icsm_get_blkcb_mapnum(blkcb->blknum) * sizeof(uint32_t));
	}
	return blkcb;
}

static int icsm_get_unused_block_number(icsm_blkcb_t *blkcb)
{
	int i, j;
	for (i = 0; i < icsm_get_blkcb_mapnum(blkcb->blknum); i++) {
		if (ICSM_BLKCB_BITMAP(blkcb, i) != 0xffffffff) {
			for (j = 0; j < 32; j++) {
				if (!(ICSM_BLKCB_BITMAP(blkcb, i) & (1 << j))) {
					return (i * 32) + j;
				}
			}
		}
	}
	return ICSM_E_NOMEM;
}

static void icsm_set_blkmap(icsm_blkcb_t *blkcb, int blkno)
{
	int bit = 1 << (blkno % 32);
	ICSM_BLKCB_BITMAP(blkcb, blkno / 32) |= bit;
}

static void icsm_clr_blkmap(icsm_blkcb_t *blkcb, int blkno)
{
	int bit = 1 << (blkno % 32);
	ICSM_BLKCB_BITMAP(blkcb, blkno / 32) &= ~bit;
}

static int icsm_is_set_blkmap(icsm_blkcb_t *blkcb, int blkno)
{
	int bit = 1 << (blkno % 32);
	if (ICSM_BLKCB_BITMAP(blkcb, blkno / 32) & bit) {
		return 1;
	}
	return 0;
}

int icsm_core_get_block(unsigned short memid, int timeout, icsm_block_t *sblock)
{
	int ret;
	icsm_blkcb_t *blkcb;
	int blkno;
	int lefttime;
	unsigned long lockaddr;
	unsigned long flags;

	if (memid >= ICSM_MEMBLK_MAX) {
		ret = ICSM_E_PAR;
		goto out;
	}
	if ((timeout < 0) && (timeout != ICSM_FOREVER)) { /* pgr1223 */
		ret = ICSM_E_PAR;
		goto out;
	}
	blkcb = icsm_get_blkcb(memid);

	lockaddr = ICSM_LOCK_ADDR;
	lefttime = timeout;
	for (;;) {
		icc_lock_irqsave(lockaddr, &flags);
		/* get unused block */
		blkno = icsm_get_unused_block_number(blkcb);
		if (blkno >= 0) {
			icsm_set_blkmap(blkcb, blkno);
			sblock->memid  = memid;
			sblock->maddr  = (void*)((unsigned char*)blkcb->topaddr + (blkcb->blklen * blkno));
			sblock->length = blkcb->blklen;

			icc_unlock_irqrestore(lockaddr, &flags);

			ret = 0;
			goto out;
		}
		icc_unlock_irqrestore(lockaddr, &flags);
		/* get block failed. */
		if (timeout == 0) {
			break;
		} else if (timeout != ICSM_FOREVER) {
			if (lefttime <= 0) {
				break;
			}
			lefttime -= icsm_time_to_msecs(ICSM_SLEEP_TIME);
		}
		icsm_sleep(ICSM_SLEEP_TIME);
		/* retry get block until succeed, or timeout */
	}
	/* when loop break. */
	ret = ICSM_E_NOMEM;
out:
	return ret;
}

int icsm_core_free_block(icsm_block_t *sblock)
{
	int ret = 0;
	icsm_blkcb_t *blkcb;
	int blkno;
	unsigned long lockaddr;
	unsigned long flags;

	if (sblock->memid >= ICSM_MEMBLK_MAX) {
		ret = ICSM_E_PAR;
		goto out;
	}
	blkcb = icsm_get_blkcb(sblock->memid);

	lockaddr = ICSM_LOCK_ADDR;
	icc_lock_irqsave(lockaddr, &flags);

	blkno = ((unsigned int)sblock->maddr - blkcb->topaddr) / sblock->length;
	if (!icsm_is_set_blkmap(blkcb, blkno)) {
		ret = ICSM_E_BUSY;
		goto out_unlock;
	}
	/* When use am32, need to memory buffer purge */
	icsm_memory_sync((unsigned long)sblock->maddr, sblock->length);

	icsm_clr_blkmap(blkcb, blkno);

out_unlock:
	icc_unlock_irqrestore(lockaddr, &flags);
out:
	return ret;
}
