
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "icsm_if.h"
#include "icsmdrv.h"

#define	ICSM_MAJOR	124
#define ICSM_MINOR_NUM	1
#define DRIVER_NAME	"icsm"

static struct cdev  icsm_cdev;

static int icsm_major = ICSM_MAJOR;
module_param(icsm_major, int, S_IRUGO);

icsmdrv_info_t icsmdrv_info;

struct icsmdrv_iomap_struct icsmdrv_iomap[] = ICSMDRV_IOMAP_REGION;

static int icsmdrv_open(struct inode *inode, struct file *file);
static int icsmdrv_release(struct inode *inode, struct file *filp);
static int icsmdrv_ioctl(struct inode *inode, struct file *filp,unsigned int cmd, unsigned long arg);

const struct file_operations icsm_fops =
{
	.open	= icsmdrv_open,
	.release= icsmdrv_release,
	.ioctl	= icsmdrv_ioctl,
};

extern int icsmdrv_get_block(icsm_block_t* sblock, unsigned short memid);
extern int icsmdrv_free_block(icsm_block_t* sblock);

static int icsmdrv_open(struct inode *inode, struct file *filp)
{
	return 0;
}

static int icsmdrv_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static int icsmdrv_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
			 unsigned long arg)
{
	int ret = 0;
	icsm_block_t sblock;
	unsigned long paddr;
	icsm_ctrl_t ctrl;
	int memid;

	switch(cmd) {
	case ICSM_IOC_GET_BLOCK:
		if (copy_from_user(&sblock, (int __user *)arg, sizeof(sblock))) {
			ret = -EFAULT;
			break;
		}
		memid = sblock.memid;
		if (memid >= ICSM_MEMBLK_MAX) {
			ret = -EINVAL;
			break;
		}
		ret = icsm_core_get_block(memid, icsmdrv_info.timeout[memid], &sblock);
		if (ret == ICSM_E_NOMEM) {
			ret = -ENOMEM;
			break;
		} else if (ret) {
			ret = -EINVAL;
			break;
		}
		if (copy_to_user((int __user *)arg, &sblock, sizeof(sblock))) {
			ret = -EFAULT;
		}
		break;

	case ICSM_IOC_FREE_BLOCK:
		if (copy_from_user(&sblock, (int __user *)arg, sizeof(sblock))) {
			ret = -EFAULT;
			break;
		}
		ret = icsm_core_free_block(&sblock);
		if (ret == ICSM_E_BUSY) {
			ret = -EBUSY;
			break;
		} else if (ret) {
			ret = -EINVAL;
		}
		break;

	case ICSM_IOC_GET_PADDR:
		if (copy_from_user(&sblock, (int __user *)arg, sizeof(sblock))) {
			ret = -EFAULT;
			break;
		}
		paddr = icsm_maddr_to_paddr((unsigned long)sblock.maddr);
		if (paddr == (unsigned long)(-1)) {
			ret = -EINVAL;
			break;
		}
		ret = (int)paddr;
		break;
    
	case ICSM_IOC_SET_TIMEOUT:
		if (copy_from_user(&ctrl, (int __user *)arg, sizeof(ctrl))) {
			ret = -EFAULT;
			break;
		}
		memid = ctrl.memid;
		if (memid >= ICSM_MEMBLK_MAX) {
			ret = -EINVAL;
			break;
		}
		icsmdrv_info.timeout[memid] = ctrl.timeout;
		break;
	default:
		ret = -ENOTTY;
		break;
	}

	return ret;
}

static int __init icsm_init_driver(void)
{
	dev_t dev = 0;
	int alloc_ret = -1;
	int cdev_ret = -1;
	int i;
	icsm_basecb_t *basecb;

	if (icsm_major) {
		dev = MKDEV(icsm_major, 0);
		alloc_ret = register_chrdev_region(dev, ICSM_MINOR_NUM, DRIVER_NAME);
	} else {
		alloc_ret = alloc_chrdev_region(&dev, 0, ICSM_MINOR_NUM, DRIVER_NAME);
		icsm_major = MAJOR(dev);
	}
	if (alloc_ret) {
		printk(KERN_WARNING \
		       "icsmdrv: Allocate device failed %d\n", alloc_ret);
		goto init_error;
	}

	cdev_init(&icsm_cdev, &icsm_fops);
	icsm_cdev.owner = THIS_MODULE;
	cdev_ret = cdev_add(&icsm_cdev, dev, ICSM_MINOR_NUM);
	if (cdev_ret) {
		goto init_error;
	}

	/* mapping ssc and register region */
	for (i = 0; icsmdrv_iomap[i].size > 0; i++) {
		unsigned long paddr = icsmdrv_iomap[i].paddr;
		int size = icsmdrv_iomap[i].size;

		icsmdrv_iomap[i].vaddr = (unsigned long)ioremap_nocache(paddr, size);
		if (!icsmdrv_iomap[i].vaddr) {
			printk("icsm_init_driver: ioremap failed.\n");
			goto init_error;
		}
		/* check icsm blcok control block or remap */
		if (paddr <= ICSM_CB_TOP_ADDR && ICSM_CB_TOP_ADDR < (paddr + size)) {
			basecb = (icsm_basecb_t*)__icsmdrv_io(ICSM_CB_TOP_ADDR);
			if (basecb->length & 0x3) {
				printk("icsm_init_driver: basecb->length (%d) is incorrectly.\n",
				       basecb->length);
				goto init_error;
			}
			if ((paddr + size) < ICSM_CB_TOP_ADDR + basecb->length) {
				printk("icsm_init_driver: basecb->length (%d) over the mapping size.\n",
				       basecb->length);
				iounmap((void*)icsmdrv_iomap[i].vaddr);
				icsmdrv_iomap[i].size = PAGE_ALIGN(ICSM_CB_TOP_ADDR + basecb->length - paddr);
				i--; /* try remapping */
			}
		}
	}

	printk(KERN_ALERT "%s driver version %d.%d.%d (major %d)\n",
	       DRIVER_NAME, ICSM_VER_MAJOR, ICSM_VER_MINOR, ICSM_VER_RELEASE, icsm_major);

	return 0;

init_error:
	for (i = 0; icsmdrv_iomap[i].size > 0; i++) {
		if (icsmdrv_iomap[i].vaddr) {
			iounmap((void*)icsmdrv_iomap[i].vaddr);
		}
	}
  	if (!cdev_ret) {
		cdev_del(&icsm_cdev);
	}
	if (!alloc_ret) {
		unregister_chrdev_region(dev, ICSM_MINOR_NUM);
	}
	return -1;
}


static void __exit icsm_exit_driver(void)
{
	dev_t dev = MKDEV(icsm_major, 0);
	int i;

	cdev_del(&icsm_cdev);
	unregister_chrdev_region(dev, ICSM_MINOR_NUM);

	for (i = 0; icsmdrv_iomap[i].size > 0; i++) {
		if (icsmdrv_iomap[i].vaddr) {
			iounmap((void*)icsmdrv_iomap[i].vaddr);
		}
	}
	printk(KERN_ALERT "%s driver removed.\n", DRIVER_NAME);
}

MODULE_AUTHOR("Panasonic");
MODULE_DESCRIPTION("Inter CPU Shared Memory driver");
MODULE_LICENSE("Proprietary");

module_init(icsm_init_driver);
module_exit(icsm_exit_driver);
