/*
 *  sveth_ipcc_kapi.c
 *
 *	The virtual ethernet driver for the SPiCa movie system.
 *
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 */

#if defined(CONFIG_SVETH_DEBUG)
static const char* devname_p = "sveth_ipcc";
#endif	/* CONFIG_SVETH_DEBUG */

#include "net_event.h"
#include "ppcdev.h"

#include "sveth_config.h"
#include "sveth_debug.h"

#include "sveth_util.h"
#include "sveth_ipcc_local.h"


/**
 *	private data
 */
static sveth_ipcc_local_t g_local[SVETH_IPCC_LOCAL_NUM];


/**
 *	for private function
 */

/**
 *	@brief		_netev_nemsg2pkt
 *	
 *	@param[in]	*nemsg_p			NWA netevent message
 *	
 *	@return		sveth_ipcc_pkt_t*	non-0 (sveth ipcc packet) on success
 *									0 on failure
 *	
 *	@details						get sveth ipcc packet from NWA netevent message
 */
static inline sveth_ipcc_pkt_t* _netev_nemsg2pkt(const NET_EVT_MSG* nemsg_p) {
	if(unlikely(nemsg_p == 0)) {
		return 0;
	}

	return (sveth_ipcc_pkt_t*)&nemsg_p->Data.SvethIpccPkt;
}


/**
 *	@brief		_netev_nemsg_len
 *	
 *	@param[in]	*nemsg_p	NWA netevent message
 *	
 *	@return		size_t		sveth ipcc packet length
 *	
 *	@details		return sveth ipcc packet length
 */
static inline size_t _netev_nemsg_len(const NET_EVT_MSG* nemsg_p) {
	sveth_ipcc_pkt_t* pkt_p = _netev_nemsg2pkt(nemsg_p);
	return offsetof(NET_EVT_MSG, Data) + offsetof(sveth_ipcc_pkt_t, cmd) + pkt_p->head.len;
}


/**
 *	@brief		_netev_initheader
 *	
 *	@param[in,out]	*nemsg_p	NWA netevent message
 *	
 *	@return		none
 *	
 *	@details		initialize the NWA netevent message header
 */
static inline void _netev_initheader(NET_EVT_MSG* nemsg_p) {
	/* init net event message  */
	nemsg_p->Evt = NET_EVT_SVETH_IPCC;
	nemsg_p->Type = NET_EVT_TYPE_REQUEST;
	nemsg_p->ReqId = 0xFFFFFFFF;				/* SVETHでは未使用 */
	nemsg_p->Result = NET_E_OK;
}



/**
 *	@brief		_sveth_ipcc_ppcdev_handler.
 *	
 *	@param[in]	ppcdev_fd	ipcc descriptor
 *	@param[in]	id			id
 *	@param[in]	*buff_p		NWA netevent message
 *	@param[in]	len			length of buff_p (unused parameter)
 *	
 *	@return		irqreturn_t	IRQ_HANDLED on success
 *							IRQ_NONE on failure or unknown irq
 *	@details		ppcdev interrupt event handler
 */
static irqreturn_t _sveth_ipcc_ppcdev_handler(int ppcdev_fd, int id, const void* buff_p, size_t len) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_cmd_t* cmd_p;
	unsigned char cmd_id = -1;
	int ret;

	// PDEBUG("ipcc handler called: ppcdev_fd=%d, id=%d, buff_p=%p, len=%zd\n", ppcdev_fd, id, buff_p, len);
	/* get fd */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_sveth_ipcc_foreach(p, g_local) {
		if(unlikely(_sveth_ipcc_get_id(p) == id)) {
			goto _findout;
		}
	}

	_sveth_unlock_irq(p, flags);	/* -------- unlock(return) -------- */
	PWARNING("ppcdev_fd=%d unregisterd\n", ppcdev_fd);
	/* invalid id */
	return IRQ_NONE;

_findout:

	nemsg_p = (NET_EVT_MSG*)buff_p;
	cmd_p = sveth_ipcc_get_cmd(_netev_nemsg2pkt(nemsg_p), &cmd_id);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	if(unlikely(cmd_p == 0)) {
		/* invalid pointer */
		PWARNING("invalid parameter: cmd_p\n");
		return IRQ_NONE;
	}
	
	switch(cmd_id) {
	case SVETH_IPCC_INIT_ACK:
		DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_INIT_ACK);
		if(p->event.pf_init_ack != 0) {
			ret = p->event.pf_init_ack(p->fd, p->param_p, cmd_p->init_ack.errcode, cmd_p->init_ack.vreg_p, cmd_p->init_ack.vreg_size);
			PERRMSG((ret == 0));
		}
		break;

	case SVETH_IPCC_QUIT_ACK:
		DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_QUIT_ACK);
		if(p->event.pf_quit_ack != 0) {
			ret = p->event.pf_quit_ack(p->fd, p->param_p, cmd_p->quit_ack.errcode);
			PERRMSG((ret == 0));
		}
		break;

	case SVETH_IPCC_RECV:
		DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_RECV);
		if(p->event.pf_recv_pkt != 0) {
			ret = p->event.pf_recv_pkt(p->fd, p->param_p, cmd_p->recv.rtd_p);
			PERRMSG((ret == 0));
		}
		break;

	case SVETH_IPCC_STATIND:
		DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_STATIND_RECV);
		if(p->event.pf_stat_ind != 0) {
			ret = p->event.pf_stat_ind(p->fd, p->param_p, cmd_p->statind.prm);
			PERRMSG((ret == 0));
		}
		break;

	case SVETH_IPCC_GETPRM_ACK:
		switch(cmd_p->getprm_ack.code) {
		case SVETH_IPCC_GETPRM_SUB_INTR:
			DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_SUB_INTR_ACK);
			if(p->event.pf_get_int_stat != 0) {
				ret = p->event.pf_get_int_stat(
							p->fd,
							p->param_p,
							cmd_p->getprm_ack.errcode,
							cmd_p->getprm_ack.req_id,
							cmd_p->getprm_ack.prm.intr.flg
				);
				PERRMSG((ret == 0));
			}
			break;

		case SVETH_IPCC_GETPRM_SUB_MAC:
			DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_SUB_MAC_ACK);
			if(p->event.pf_get_macaddr != 0) {
				ret = p->event.pf_get_macaddr(
							p->fd,
							p->param_p,
							cmd_p->getprm_ack.errcode,
							cmd_p->getprm_ack.req_id,
							cmd_p->getprm_ack.prm.mac.val
				);
				PERRMSG((ret == 0));
			}
			break;

		case SVETH_IPCC_GETPRM_SUB_LINK:
			DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_SUB_LINK_ACK);
			if(p->event.pf_get_link_stat != 0) {
				ret = p->event.pf_get_link_stat(
							p->fd,
							p->param_p,
							cmd_p->getprm_ack.errcode,
							cmd_p->getprm_ack.req_id,
							cmd_p->getprm_ack.prm.link.speed,
							cmd_p->getprm_ack.prm.link.duplex
				);
				PERRMSG((ret == 0));
			}
			break;

		default:
			/* invalid parameter */
			PWARNING("invalid parameter: cmd_p->getprm_ack.code=0x%04X\n", cmd_p->getprm_ack.code);
			DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_INVALID_ACK);
			return IRQ_NONE;
		}
		break;

	default:
		/* invalid parameter */
		PWARNING("invalid parameter: cmd_id=0x%02X\n", cmd_p->getprm_ack.code);
		DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_INVALID_PARAM);
		return IRQ_NONE;
	}

	return IRQ_HANDLED;
}

/**
 *	IPC API [debug]
 */

void sveth_ipcc_dump_priv_info(int fd) {
	sveth_ipcc_local_t* p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return;
	}
}

/**
 *	for internal API
 */


/* ipcc[dd]: init */
int __init sveth_ipcc_dd_init(void) {
	sveth_ipcc_local_t* p;
	unsigned long flags;

	/* init */
	_sveth_ipcc_foreach(p, g_local) {
		/* init local */
		_sveth_lock_init(p);

		_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
		_sveth_ipcc_init_local(p);
		_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	}
	
	return 0;
}


/* ipcc[dd]: exit */
void __exit sveth_ipcc_dd_exit(void) {
	sveth_ipcc_local_t* p;
	unsigned long flags;

	/* exit */
	_sveth_ipcc_foreach(p, g_local) {
		/* exit local */
		_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
		_sveth_ipcc_exit_local(p);
		_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	}
}


/* ipcc[dd]: open */
int sveth_ipcc_dd_open(int id, const sveth_ipcc_event_t* event_p, const void* param_p) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int i;
	int ppcdev_fd;

	/* check param */
	if(id < 0 || event_p == 0) {
		PWARNING("invalid parameter. id=%d, event_p=%p\n", id, event_p);
		/* out of range */
		return -EINVAL;
	}

	/* check unused */
	i = 0;
	_sveth_ipcc_foreach(p, g_local) {

		_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

		/* unused */
		if(1 == _sveth_ipcc_is_unused(p)) {
			/* open ppcdev */
			ppcdev_fd = ppcdev_kopen(id);
			if(ppcdev_fd < 0) {
				/* error */
				_sveth_unlock_irq(p, flags);	/* -------- unlock(return) -------- */
				return -ENODEV;
			}

			/* set handler */
			ppcdev_kreqirq(ppcdev_fd, _sveth_ipcc_ppcdev_handler);
			
			/* set parameter */
			_sveth_ipcc_set_local(p, i, id, ppcdev_fd, event_p, (void*)param_p);

			_sveth_unlock_irq(p, flags);	/* -------- unlock(return) -------- */
			return i;	/* as fd */
		}
		/* already used */
		else if(1 == _sveth_ipcc_id_is_registerd(p, id)) {
			/* already registerd */
			_sveth_unlock_irq(p, flags);	/* -------- unlock(return) -------- */
			return -EEXIST;
		}
		else {
			i++;
			_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
		}
	}

	/* unexpected error */
	return -ENODEV;
}


/* ipcc[dd]: close */
int sveth_ipcc_dd_close(int fd) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ppcdev_fd;
	int ret;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	/* check unused */
	if(1 == _sveth_ipcc_is_unused(p)) {
		/* unused */
		_sveth_unlock_irq(p, flags);	/* -------- unlock(return) -------- */
		return -ENXIO;
	}

	/* clear parameter */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	_sveth_ipcc_init_local(p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	/* close ppcdev */
	ret = ppcdev_kclose(ppcdev_fd);
	if(ret < 0) {
		/* fail */
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_INIT */
int sveth_ipcc_kapi_init(int fd) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.init.prm = 0x00;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_INIT;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(pkt_p, sizeof(pkt_p->cmd.init));

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_INIT);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_QUIT */
int sveth_ipcc_kapi_quit(int fd) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	/* get memblock */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.quit.prm = 0x00;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_QUIT;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(pkt_p, sizeof(pkt_p->cmd.quit));

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_QUIT);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_SEND */
int sveth_ipcc_kapi_send_packet(int fd, const sveth_ipcc_rtd_t* rtd_p) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* check param */
	if(rtd_p == 0) {
		PWARNING("invalid parameter. rtd_p is null.\n");
		/* out of range */
		return -EINVAL;
	}

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** TX request transfer descriptor */
	pkt_p->cmd.send.rtd_p = (sveth_ipcc_rtd_t*)rtd_p;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_SEND;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(pkt_p, sizeof(pkt_p->cmd.send));

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_SEND);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}

/* ipcc[kapi]: SVETH_IPCC_SETPRM(SVETH_IPCC_SETPRM_SUB_INTR) */
int sveth_ipcc_kapi_set_interrupt_mode(int fd, unsigned char flg) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.setprm.code = SVETH_IPCC_SETPRM_SUB_INTR;
	pkt_p->cmd.setprm.prm.intr.flg = flg;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_SETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			offsetof(sveth_ipcc_setprm_t, prm) + sizeof(pkt_p->cmd.setprm.prm.intr)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_SETPRM_SUB_INTR);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_SETPRM(SVETH_IPCC_SETPRM_SUB_MAC) */
int sveth_ipcc_kapi_set_mac_address(int fd, const unsigned char* macaddr_p) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;
	int i;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.setprm.code = SVETH_IPCC_SETPRM_SUB_MAC;
	for(i = 0; i < ETH_ALEN; i++) {
		pkt_p->cmd.setprm.prm.mac.val[i] = macaddr_p[i];
	}

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_SETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			offsetof(sveth_ipcc_setprm_t, prm) + sizeof(pkt_p->cmd.setprm.prm.mac)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_SETPRM_SUB_MAC);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_SETPRM(SVETH_IPCC_SETPRM_SUB_POWER) */
int sveth_ipcc_kapi_set_power_mode(int fd, unsigned char mode) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.setprm.code = SVETH_IPCC_SETPRM_SUB_POWER;
	pkt_p->cmd.setprm.prm.pwctl.mode = mode;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_SETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			offsetof(sveth_ipcc_setprm_t, prm) + sizeof(pkt_p->cmd.setprm.prm.pwctl)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_SETPRM_SUB_POWER);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_SETPRM(SVETH_IPCC_SETPRM_SUB_MCAST) */
int sveth_ipcc_kapi_set_muticast_mode(int fd, unsigned char flg) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.setprm.code = SVETH_IPCC_SETPRM_SUB_MCAST;
	pkt_p->cmd.setprm.prm.mcast.flg = flg;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_SETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			offsetof(sveth_ipcc_setprm_t, prm) + sizeof(pkt_p->cmd.setprm.prm.mcast)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_SETPRM_SUB_MCAST);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_SETPRM(SVETH_IPCC_SETPRM_SUB_WUP) */
int sveth_ipcc_kapi_set_wol_mode(int fd, unsigned char flg) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.setprm.code = SVETH_IPCC_SETPRM_SUB_WUP;
	pkt_p->cmd.setprm.prm.wup.flg = flg;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_SETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			offsetof(sveth_ipcc_setprm_t, prm) + sizeof(pkt_p->cmd.setprm.prm.wup)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_SETPRM_SUB_WUP);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_GETPRM(SVETH_IPCC_GETPRM_SUB_INTR) */
int sveth_ipcc_kapi_request_interrupt_state(int fd, unsigned short req_id) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.getprm.code = SVETH_IPCC_GETPRM_SUB_INTR;
	pkt_p->cmd.getprm.req_id = req_id;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_GETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			sizeof(pkt_p->cmd.getprm)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_SUB_INTR);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_GETPRM(SVETH_IPCC_GETPRM_SUB_MAC) */
int sveth_ipcc_kapi_request_mac_address(int fd, unsigned short req_id) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.getprm.code = SVETH_IPCC_GETPRM_SUB_MAC;
	pkt_p->cmd.getprm.req_id = req_id;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_GETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			sizeof(pkt_p->cmd.getprm)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_SUB_MAC);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_GETPRM(SVETH_IPCC_GETPRM_SUB_LINK) */
int sveth_ipcc_kapi_request_link_state(int fd, unsigned short req_id) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.getprm.code = SVETH_IPCC_GETPRM_SUB_LINK;
	pkt_p->cmd.getprm.req_id = req_id;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_GETPRM;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			sizeof(pkt_p->cmd.getprm)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_GETPRM_SUB_LINK);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}


/* ipcc[kapi]: SVETH_IPCC_STATIND */
int sveth_ipcc_kapi_notify_status(int fd, unsigned char prm) {
	sveth_ipcc_local_t* p;
	unsigned long flags;
	int ret;
	int id;
	int ppcdev_fd;
	size_t size;

	NET_EVT_MSG* nemsg_p;
	sveth_ipcc_pkt_t* pkt_p;

	/* get current */
	p = _sveth_ipcc_current_local(fd, g_local);
	if(p == 0) {
		PWARNING("invalid parameter. fd=%d\n", fd);
		/* out of range */
		return -EINVAL;
	}

	/* get memblock */
	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */
	nemsg_p = ppcdev_kmalloc(sizeof(*nemsg_p));
	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */
	if(nemsg_p == 0) {
		PWARNING("ppcdev_kmalloc() fail\n");
		/* not enough memory */
		return -ENOMEM;
	}

	_sveth_lock_irq(p, flags);	/* ++++++++ lock ++++++++ */

	_netev_initheader(nemsg_p);
	pkt_p = _netev_nemsg2pkt(nemsg_p);
	sveth_assert(pkt_p != 0);
	
	/** payload */
	pkt_p->cmd.statind.prm = prm;

	/** header */
	pkt_p->head.start_id = SVETH_IPCC_START_ID;
	pkt_p->head.cmd_id = SVETH_IPCC_STATIND;
	pkt_p->head.len = (UB)_sveth_ipcc_pkt_size(
			pkt_p,
			sizeof(pkt_p->cmd.statind)
	);

	/** size of net event message */
	ppcdev_fd = _sveth_ipcc_get_ppcdev_fd(p);
	id = _sveth_ipcc_get_id(p);
	size = _netev_nemsg_len(nemsg_p);

	_sveth_unlock_irq(p, flags);	/* -------- unlock -------- */

	DUMP_NEMSG(nemsg_p, CONFIG_SVETH_DUMP_MSG_IPCC_STATIND_SEND);

	/* request ppcdev */
	ret = ppcdev_ksend(ppcdev_fd, id, nemsg_p, size);
	if(ret < 0) {
		PWARNING("ppcdev_ksend() fail. ret=%d\n", ret);
		/* process failure */
		ppcdev_kfree(nemsg_p);
		return -ENXIO;
	}

	/* success */
	return 0;
}

