/*
 *  sveth_main.c
 *
 *	The virtual ethernet driver for the SPiCa movie system.
 *
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/moduleparam.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>

#include <linux/list.h>

#include <linux/in.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>

#include <linux/in6.h>
#include <linux/crc32.h>
#include <asm/checksum.h>

#include <asm/hardware/arm_timer.h>
#include <asm/mach/time.h>


MODULE_LICENSE("Proprietary");
MODULE_AUTHOR("Panasonic");
MODULE_DESCRIPTION("SPiCa virtual ethernet driver");


/* devices */
#if defined(CONFIG_SVETH_DEBUG)
static const char* devname_p  = "sveth_main";
#endif	/* CONFIG_SVETH_DEBUG */

#include "sveth_config.h"
#include "sveth_debug.h"

#include "sveth_ipcc_kapi.h"

#include "sveth_util.h"
#include "sveth.h"


#include "sveth_mpool.h"
#include "sveth_que.h"


struct sveth_txpkt_t;
struct sveth_rxpkt_t;
struct sveth_priv_t;


struct net_device* g_netdev_table[SVETH_DEV_NUM];

static int opencount;

/* timeout value */
static int timeout = SVETH_TIMEOUT;
module_param(timeout, int, 0);


/* for TX packet */
struct sveth_txpkt_t {
	struct sveth_priv_t*		priv_p;
	
	sveth_ipcc_ptd_t*			ptd_p;
};


/* for RX packet */
struct sveth_rxpkt_t {
	struct sveth_priv_t*		priv_p;
	
	sveth_ipcc_rtd_t*			rtd_p;
};


/* sveth private data */
struct sveth_priv_t {

	/* net info */
	struct net_device_stats		stats;

	/* TX/RX mpool */
	struct sveth_mpool_que_t	tx_mpool;
	struct sveth_mpool_que_t	rx_mpool;

	/* TX/RX pkt queue */
	struct sveth_que_t			tx_pktque;
	struct sveth_que_t			rx_pktque;

	struct net_device*			netdev_p;

	/* TX/RX Event thread */
	struct task_struct*			tx_event_thread_p;
	struct task_struct*			rx_event_thread_p;

	/* TX/RX Event queue head */
	wait_queue_head_t			tx_event_qh;
	wait_queue_head_t			rx_event_qh;

#if 0
	/* TX-RTD */
	int							tx_rtd_current;

	/* TX-PTD */
	int							tx_ptd_current;
#endif

	/* for shmem */
	struct resource *			vreg_resource_p;
	volatile sveth_ipcc_vreg_t*	vreg_vaddr_p;
	volatile sveth_ipcc_vreg_t*	vreg_p;
	unsigned long				vreg_size;

	int							opened;
	spinlock_t					sveth_lock;
	int							fd;
	unsigned long				ipcc_stat_flg;
	wait_queue_head_t			ipcc_kapi_qh;
	sveth_ipcc_event_t*			ipcc_event_p;


#if defined(CONFIG_SVETH_DEBUG)
	unsigned long				rx_ptd_free_cnt;
	unsigned long				rx_ptd_recv_cnt;

	unsigned long				ipcc_ev_recv_pkt_cnt;

#endif	/* CONFIG_SVETH_DEBUG */
};


/* prototype */
static inline int _open_process(struct net_device* netdev_p);
static inline int _release_process(struct net_device* netdev_p);

/* net device operation */
static int sveth_ndo_open(struct net_device* netdev_p);
static int sveth_ndo_release(struct net_device* netdev_p);
static netdev_tx_t sveth_ndo_start_xmit(struct sk_buff* skb_p, struct net_device* netdev_p);
static int sveth_ndo_ioctl(struct net_device* netdev_p, struct ifreq* rq, int cmd);
static struct net_device_stats* sveth_ndo_get_stats(struct net_device* netdev_p);
static int sveth_ndo_change_mtu(struct net_device* netdev_p, int new_mtu);
static void sveth_ndo_tx_timeout(struct net_device* netdev_p);


/* the other function */
static int sveth_check_send_available(struct net_device* netdev_p);




static inline void* sveth_shmem_virt_to_phys(struct sveth_priv_t* priv_p, void* vaddr_p) {
	if(!vaddr_p || !priv_p->vreg_vaddr_p ||
			(uintptr_t)vaddr_p < (uintptr_t)priv_p->vreg_vaddr_p ||
			(uintptr_t)vaddr_p >= (uintptr_t)priv_p->vreg_vaddr_p + priv_p->vreg_size) {
		/* invalid address */
		return 0;
	}
	
	/* success */
	return (void*)(((uintptr_t)vaddr_p - (uintptr_t)priv_p->vreg_vaddr_p) + (uintptr_t)priv_p->vreg_p);
}


static inline void* sveth_shmem_phys_to_virt(struct sveth_priv_t* priv_p, void* paddr_p) {
	if(!priv_p->vreg_vaddr_p || !paddr_p ||
			(uintptr_t)paddr_p < (uintptr_t)priv_p->vreg_p ||
			(uintptr_t)paddr_p >= (uintptr_t)priv_p->vreg_p + priv_p->vreg_size) {
		/* invalid address */
		return 0;
	}
	
	/* success */
	return (void*)(((uintptr_t)paddr_p - (uintptr_t)priv_p->vreg_p) + (uintptr_t)priv_p->vreg_vaddr_p);
}


static inline int _sveth_shmem_check_event_is_expired(unsigned long timeout) {
	unsigned long expired = jiffies + timeout;

	return (likely(timeout > 0)) ?
		time_before(jiffies, expired) :		/* check timed */
		1;									/* inifinity */
}

static inline int sveth_shmem_check_event(struct sveth_priv_t* priv_p, unsigned long event, unsigned long timeout) {
	unsigned long flags;
	int ret;
	unsigned long timed;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);
	/* check Tx req */
	{
		timed = jiffies + SVETH_CHECK_EVENT_SCHED;
		do {
			/* check event */
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			ret = _sveth_issetflg(priv_p->vreg_vaddr_p->event, event);
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
			if(unlikely(!ret)) {
				goto _out;
			}

			/* delay */
			if(unlikely(time_after(jiffies, timed))) {
				/* update expired and call scheduler */
				timed = jiffies + SVETH_CHECK_EVENT_SCHED;
				schedule();
			}
			else {
				udelay(SVETH_CHECK_EVENT_POLLING);
			}
		} while(_sveth_shmem_check_event_is_expired(timeout));
	}

	return -EFAULT;		/* expired */

_out:
	return 0;
}


static inline void sveth_shmem_set_event(struct sveth_priv_t* priv_p, unsigned long event) {
	unsigned long flags;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);
	/* set Tx req */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_setflg(priv_p->vreg_vaddr_p->event, event);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
}


static inline void sveth_shmem_clear_event(struct sveth_priv_t* priv_p, unsigned long event) {
	unsigned long flags;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);
	/* set Tx req */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_clrflg(priv_p->vreg_vaddr_p->event, event);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
}


static inline void* sveth_shmem_ptd_alloc_tx(struct sveth_priv_t* priv_p, void* data_p, unsigned long len) {
	unsigned long flags;
	void* ptd_p;
	volatile sveth_ipcc_ptd_t* vptd_p;
	void* vaddr_p;
	int i;
	unsigned long copy_len;

	sveth_assert(priv_p != 0);
	sveth_assert(data_p != 0);
	sveth_assert(len > 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	/* init tx_current */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);

#if 0
	for(	i = priv_p->tx_ptd_current;
			i != _sveth_index_prevof(priv_p->tx_ptd_current, priv_p->vreg_vaddr_p->tx_ptd_num);
			i = _sveth_index_nextof(i, priv_p->vreg_vaddr_p->tx_ptd_num)
	) {
		/* check flag */
		ptd_p = (void*)(priv_p->vreg_vaddr_p->tx_ptd_top + priv_p->vreg_vaddr_p->tx_ptd_size * i);
		sveth_assert(ptd_p != 0);

		vptd_p = (volatile sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p);
		sveth_assert(vptd_p != 0);

		if(likely(!_sveth_issetval(vptd_p->status, SVETH_IPCC_PTD_STATUS_REQ))) {
			/* unused block found */
			priv_p->tx_ptd_current = _sveth_index_nextof(i, priv_p->vreg_vaddr_p->tx_ptd_num);
			goto _found;
		}
	}
#else
	for(i = 0; i < priv_p->vreg_vaddr_p->tx_ptd_num; i++) {
		/* check flag */
		ptd_p = (void*)(priv_p->vreg_vaddr_p->tx_ptd_top + priv_p->vreg_vaddr_p->tx_ptd_size * i);
		sveth_assert(ptd_p != 0);

		vptd_p = (volatile sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p);
		sveth_assert(vptd_p != 0);

		if(likely(!_sveth_issetval(vptd_p->status, SVETH_IPCC_PTD_STATUS_REQ))) {
			/* unused block found */
			goto _found;
		}
	}
#endif

	/* not found */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return 0;


_found:
	/* copy data */
	sveth_assert(vptd_p->data_p != 0);
	vaddr_p = (void*)sveth_shmem_phys_to_virt(priv_p, vptd_p->data_p);
	sveth_assert(vaddr_p != 0);
	memcpy(vaddr_p, data_p, len);
	memset((void*)((uintptr_t)vaddr_p + len), 0x0, ETH_FCS_LEN);		/* for CRC32 */
	/* with FCS */
	copy_len = len;							/* without CRC32 */

	/* set TX-PTD param */
	vptd_p->next_p = 0;
	vptd_p->len = copy_len;
	_sveth_setval(vptd_p->status, SVETH_IPCC_PTD_STATUS_REQ);

	/* success */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return ptd_p;
}


static inline void sveth_shmem_ptd_free(struct sveth_priv_t* priv_p, sveth_ipcc_ptd_t* ptd_p) {
	unsigned long flags;
	volatile sveth_ipcc_ptd_t* vptd_p;

	sveth_assert(priv_p != 0);
	if(ptd_p == 0) {
		/* invalid parameter */
		PWARNING("invalid param. ptd_p=%p\n", ptd_p);
		return;
	}

	/* set status to IDLE */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	vptd_p = (volatile sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p);
	sveth_assert(vptd_p != 0);
	if(vptd_p == 0) {
		PWARNING("invalid param. vptd_p=%p\n", vptd_p);
		goto _out;
	}
	_sveth_setval(vptd_p->status, SVETH_IPCC_PTD_STATUS_IDLE);

_out:
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
}


static inline int sveth_shmem_ptd_is_empty_tx(struct sveth_priv_t* priv_p) {
	unsigned long flags;
	void* ptd_p;
	volatile sveth_ipcc_ptd_t* vptd_p;
	int i;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	/* init tx_current */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);

#if 0
	for(	i = priv_p->tx_ptd_current;
			i != _sveth_index_prevof(priv_p->tx_ptd_current, priv_p->vreg_vaddr_p->tx_ptd_num);
			i = _sveth_index_nextof(i, priv_p->vreg_vaddr_p->tx_ptd_num)
	) {
		/* check flag */
		ptd_p = (void*)(priv_p->vreg_vaddr_p->tx_ptd_top + priv_p->vreg_vaddr_p->tx_ptd_size * i);
		sveth_assert(ptd_p != 0);

		vptd_p = (volatile sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p);
		sveth_assert(vptd_p != 0);

		if(likely(!_sveth_issetval(vptd_p->status, SVETH_IPCC_PTD_STATUS_REQ))) {
			/* unused TX-PTD found */
			goto _found;
		}
	}
#else
	for(i = 0; i < priv_p->vreg_vaddr_p->tx_ptd_num; i++) {
		/* check flag */
		ptd_p = (void*)(priv_p->vreg_vaddr_p->tx_ptd_top + priv_p->vreg_vaddr_p->tx_ptd_size * i);
		sveth_assert(ptd_p != 0);

		vptd_p = (volatile sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p);
		sveth_assert(vptd_p != 0);

		if(likely(!_sveth_issetval(vptd_p->status, SVETH_IPCC_PTD_STATUS_REQ))) {
			/* unused TX-PTD found */
			goto _found;
		}
	}
#endif

	/* not found */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return 1;


_found:
	/* success: unused TX-PTD found */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return 0;
}



static inline void sveth_shmem_ptd_check_tx(struct sveth_priv_t* priv_p) {
	unsigned long flags;
	void* ptd_p;
	volatile sveth_ipcc_ptd_t* vptd_p;
	int i;
	unsigned long free_cnt = 0;
	unsigned long used_cnt = 0;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	spin_lock_irqsave(&priv_p->sveth_lock, flags);

	for(i = 0; i < priv_p->vreg_vaddr_p->tx_ptd_num; i++) {
		/* check flag */
		ptd_p = (void*)(priv_p->vreg_vaddr_p->tx_ptd_top + priv_p->vreg_vaddr_p->tx_ptd_size * i);
		sveth_assert(ptd_p != 0);

		vptd_p = (volatile sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p);
		sveth_assert(vptd_p != 0);

		if(likely(!_sveth_issetval(vptd_p->status, SVETH_IPCC_PTD_STATUS_REQ))) {
			/* unused TX-PTD found */
			free_cnt++;
		}
		else {
			/* used TX-PTD found */
			used_cnt++;
		}
	}

	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	PNOTICE("free=%lu, used=%lu\n", free_cnt, used_cnt);
}


static inline void sveth_shmem_ptd_check_rx(struct sveth_priv_t* priv_p) {
	unsigned long flags;
	void* ptd_p;
	volatile sveth_ipcc_ptd_t* vptd_p;
	int i;
	unsigned long free_cnt = 0;
	unsigned long used_cnt = 0;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	spin_lock_irqsave(&priv_p->sveth_lock, flags);

	for(i = 0; i < priv_p->vreg_vaddr_p->rx_ptd_num; i++) {
		/* check flag */
		ptd_p = (void*)(priv_p->vreg_vaddr_p->rx_ptd_top + priv_p->vreg_vaddr_p->rx_ptd_size * i);
		sveth_assert(ptd_p != 0);

		vptd_p = (volatile sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p);
		sveth_assert(vptd_p != 0);

		if(likely(!_sveth_issetval(vptd_p->status, SVETH_IPCC_PTD_STATUS_REQ))) {
			/* unused RX-PTD found */
			free_cnt++;
		}
		else {
			/* used TX-PTD found */
			used_cnt++;
		}
	}

	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	PNOTICE("free=%lu, used=%lu\n", free_cnt, used_cnt);
}



static inline void sveth_shmem_rtd_check_tx(struct sveth_priv_t* priv_p) {
	unsigned long flags;
	void* rtd_p;
	volatile sveth_ipcc_rtd_t* vrtd_p;
	int i;
	unsigned long free_cnt = 0;
	unsigned long used_cnt = 0;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	spin_lock_irqsave(&priv_p->sveth_lock, flags);

	for(i = 0; i < priv_p->vreg_vaddr_p->tx_rtd_num; i++) {
		/* check flag */
		rtd_p = (void*)(priv_p->vreg_vaddr_p->tx_rtd_top + priv_p->vreg_vaddr_p->tx_rtd_size * i);
		sveth_assert(rtd_p != 0);

		vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
		sveth_assert(vrtd_p != 0);

		if(likely(!_sveth_issetval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_REQ))) {
			/* unused TX-RTD found */
			free_cnt++;
		}
		else {
			/* used TX-RTD found */
			used_cnt++;
		}
	}

	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	PNOTICE("free=%lu, used=%lu\n", free_cnt, used_cnt);
}


static inline void sveth_shmem_rtd_check_rx(struct sveth_priv_t* priv_p) {
	unsigned long flags;
	void* rtd_p;
	volatile sveth_ipcc_rtd_t* vrtd_p;
	int i;
	unsigned long free_cnt = 0;
	unsigned long used_cnt = 0;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	spin_lock_irqsave(&priv_p->sveth_lock, flags);

	for(i = 0; i < priv_p->vreg_vaddr_p->rx_rtd_num; i++) {
		/* check flag */
		rtd_p = (void*)(priv_p->vreg_vaddr_p->rx_rtd_top + priv_p->vreg_vaddr_p->rx_rtd_size * i);
		sveth_assert(rtd_p != 0);

		vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
		sveth_assert(vrtd_p != 0);

		if(likely(!_sveth_issetval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_REQ))) {
			/* unused RX-RTD found */
			free_cnt++;
		}
		else {
			/* used TX-RTD found */
			used_cnt++;
		}
	}

	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	PNOTICE("free=%lu, used=%lu\n", free_cnt, used_cnt);
}



static inline void* sveth_shmem_rtd_alloc_tx(struct sveth_priv_t* priv_p) {
	unsigned long flags;
	void* rtd_p;
	volatile sveth_ipcc_rtd_t* vrtd_p;
	int i;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	/* init tx_current */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);

#if 0
	for(	i = priv_p->tx_rtd_current;
			i != _sveth_index_prevof(priv_p->tx_rtd_current, priv_p->vreg_vaddr_p->tx_rtd_num);
			i = _sveth_index_nextof(i, priv_p->vreg_vaddr_p->tx_rtd_num)
	) {
		/* check flag */
		rtd_p = (void*)(priv_p->vreg_vaddr_p->tx_rtd_top + priv_p->vreg_vaddr_p->tx_rtd_size * i);
		sveth_assert(rtd_p != 0);

		vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
		sveth_assert(vrtd_p != 0);

		if(likely(!_sveth_issetval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_REQ))) {
			/* unused block found */
			priv_p->tx_rtd_current = _sveth_index_nextof(i, priv_p->vreg_vaddr_p->tx_rtd_num);
			goto _found;
		}
	}
#else
	for(i = 0; i < priv_p->vreg_vaddr_p->tx_rtd_num; i++) {
		/* check flag */
		rtd_p = (void*)(priv_p->vreg_vaddr_p->tx_rtd_top + priv_p->vreg_vaddr_p->tx_rtd_size * i);
		sveth_assert(rtd_p != 0);

		vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
		sveth_assert(vrtd_p != 0);

		if(likely(!_sveth_issetval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_REQ))) {
			/* unused block found */
			goto _found;
		}
	}
#endif

	/* not found */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return 0;


_found:
	//PINFO("   rtd_p=%p\n", rtd_p);
	/* set TX param */
	vrtd_p->head_p	= 0;
	vrtd_p->tail_p	= 0;
	vrtd_p->num		= 0;
	_sveth_setval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_REQ);

	/* success */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return rtd_p;
}


static inline void sveth_shmem_rtd_free(struct sveth_priv_t* priv_p, sveth_ipcc_rtd_t* rtd_p) {
	unsigned long flags;
	volatile sveth_ipcc_rtd_t* vrtd_p;

	sveth_assert(priv_p != 0);
	if(rtd_p == 0) {
		/* invalid parameter */
		PWARNING("invalid param. rtd_p=%p\n", rtd_p);
		return;
	}

	/* set status to IDLE */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
	sveth_assert(vrtd_p != 0);
	if(vrtd_p == 0) {
		PWARNING("invalid param. vrtd_p=%p\n", vrtd_p);
		goto _out;
	}
	_sveth_setval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_IDLE);

_out:
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
}



static inline int sveth_shmem_rtd_is_empty_tx(struct sveth_priv_t* priv_p) {
	unsigned long flags;
	void* rtd_p;
	volatile sveth_ipcc_rtd_t* vrtd_p;
	int i;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	/* init tx_current */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);

#if 0
	for(	i = priv_p->tx_rtd_current;
			i != _sveth_index_prevof(priv_p->tx_rtd_current, priv_p->vreg_vaddr_p->tx_rtd_num);
			i = _sveth_index_nextof(i, priv_p->vreg_vaddr_p->tx_rtd_num)
	) {
		/* check flag */
		rtd_p = (void*)(priv_p->vreg_vaddr_p->tx_rtd_top + priv_p->vreg_vaddr_p->tx_rtd_size * i);
		sveth_assert(rtd_p != 0);

		vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
		sveth_assert(vrtd_p != 0);

		if(likely(!_sveth_issetval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_REQ))) {
			/* unused TX-RTD found */
			goto _found;
		}
	}
#else
	for(i = 0; i < priv_p->vreg_vaddr_p->tx_rtd_num; i++) {
		/* check flag */
		rtd_p = (void*)(priv_p->vreg_vaddr_p->tx_rtd_top + priv_p->vreg_vaddr_p->tx_rtd_size * i);
		sveth_assert(rtd_p != 0);

		vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
		sveth_assert(vrtd_p != 0);

		if(likely(!_sveth_issetval(vrtd_p->status, SVETH_IPCC_RTD_STATUS_REQ))) {
			/* unused TX-RTD found */
			goto _found;
		}
	}
#endif

	/* not found */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return 1;


_found:
	/* success: unused TX-RTD found */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return 0;
}


static inline int sveth_shmem_rtd_get_num(struct sveth_priv_t* priv_p, sveth_ipcc_rtd_t* rtd_p) {
	unsigned long flags;
	volatile sveth_ipcc_rtd_t* vrtd_p;
	int num;

	sveth_assert(priv_p != 0);
	sveth_assert(priv_p->vreg_vaddr_p != 0);

	/* init tx_current */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);

	vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
	sveth_assert(vrtd_p != 0);

	num = vrtd_p->num;

	/* success */
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return num;
}




static inline void sveth_shmem_rtd_set_ptd(struct sveth_priv_t* priv_p, sveth_ipcc_rtd_t* rtd_p, sveth_ipcc_ptd_t* ptd_p) {
	unsigned long flags;
	volatile sveth_ipcc_rtd_t* vrtd_p;

	sveth_ipcc_ptd_t* vtail_p;

	sveth_assert(rtd_p != 0);
	sveth_assert(ptd_p != 0);

	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
	sveth_assert(vrtd_p != 0);

	if(unlikely(vrtd_p->num == 0) || unlikely(vrtd_p->head_p == 0) || unlikely(vrtd_p->tail_p == 0)) {
		/* set PTD as head */
		vrtd_p->head_p	= ptd_p;
		vrtd_p->tail_p	= ptd_p;
		vrtd_p->num		= 1;
	}
	else {
		/* append PTD to tail */
		vtail_p = sveth_shmem_phys_to_virt(priv_p, vrtd_p->tail_p);
		vtail_p->next_p = ptd_p;
		vrtd_p->tail_p = ptd_p;
		vrtd_p->num++;
	}
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
}


static inline void* sveth_shmem_rtd_get_ptd(struct sveth_priv_t* priv_p, sveth_ipcc_rtd_t* rtd_p) {
	unsigned long flags;
	volatile sveth_ipcc_rtd_t* vrtd_p;

	sveth_ipcc_ptd_t* ptd_p;
	sveth_ipcc_ptd_t* next_p;

	sveth_assert(priv_p != 0);
	sveth_assert(rtd_p != 0);

	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	vrtd_p = (volatile sveth_ipcc_rtd_t*)sveth_shmem_phys_to_virt(priv_p, rtd_p);
	sveth_assert(vrtd_p != 0);

	if(unlikely(vrtd_p->num == 0) || unlikely(vrtd_p->head_p == 0) || unlikely(vrtd_p->tail_p == 0)) {
		/* RTD is empty */
		ptd_p = 0;
	}
	else if(unlikely(vrtd_p->num == 1) || unlikely(vrtd_p->head_p == vrtd_p->tail_p)) {
		/* pop PTD from head and clear RTD */
		ptd_p = vrtd_p->head_p;

		vrtd_p->head_p	= 0;
		vrtd_p->tail_p	= 0;
		vrtd_p->num		= 0;

		((sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p))->next_p = 0;
	}
	else {
		/* pop PTD from head */
		ptd_p = vrtd_p->head_p;
		next_p = ((sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p))->next_p;
		vrtd_p->head_p = next_p;
		vrtd_p->num--;

		((sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p))->next_p = 0;
	}
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	
	return ptd_p;
}


/* tx_mpool expiration handler */
static void _tx_expired(void* data_p) {
	unsigned long flags;
	struct sveth_priv_t* priv_p;
	struct sveth_txpkt_t* txpkt_p = data_p;
	sveth_ipcc_ptd_t* ptd_p;

	PWARNING("    TX[X]\n");

	priv_p	= txpkt_p->priv_p;
	ptd_p	= txpkt_p->ptd_p;
	sveth_ipcc_dump_priv_info(priv_p->fd);

	/* update netif status */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	priv_p->stats.tx_dropped++;
	priv_p->stats.tx_errors++;
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	/* release TX-PTD */
	sveth_shmem_ptd_free(priv_p, ptd_p);

	/* resume xfer */
	if(netif_queue_stopped(priv_p->netdev_p)) {
		netif_start_queue(priv_p->netdev_p);
	}
}


/* rx_mpool expiration handler */
static void _rx_expired(void* data_p) {
	unsigned long flags;
	struct sveth_priv_t* priv_p;
	struct sveth_rxpkt_t* rxpkt_p = data_p;
	sveth_ipcc_rtd_t* rtd_p;
	sveth_ipcc_ptd_t* ptd_p;

	PWARNING("    RX[X]\n");

	priv_p	= rxpkt_p->priv_p;
	rtd_p	= rxpkt_p->rtd_p;

	/* release all of RX-PTD */
	while(sveth_shmem_rtd_get_num(priv_p, rtd_p) > 0) {
		/* get RX-PTD */
		ptd_p = sveth_shmem_rtd_get_ptd(priv_p, rtd_p);
		if(ptd_p != 0) {

			/* update netif status */
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			priv_p->stats.rx_dropped++;
			priv_p->stats.rx_errors++;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

			/* release RX-PTD */
			sveth_shmem_ptd_free(priv_p, ptd_p);

#if defined(CONFIG_SVETH_DEBUG)
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			priv_p->rx_ptd_free_cnt++;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
#endif	/* CONFIG_SVETH_DEBUG */

		}
	}
}


/* tx_mpool cleanup handler */
static void _tx_cleanup(void* data_p) {
	struct sveth_priv_t* priv_p;
	struct sveth_txpkt_t* txpkt_p = data_p;
	sveth_ipcc_ptd_t* ptd_p;

	PWARNING("    TX[X]\n");

	priv_p	= txpkt_p->priv_p;
	ptd_p	= txpkt_p->ptd_p;
	sveth_ipcc_dump_priv_info(priv_p->fd);

	/* release TX-PTD */
	sveth_shmem_ptd_free(priv_p, ptd_p);
}


/* rx_mpool cleanup handler */
static void _rx_cleanup(void* data_p) {
	struct sveth_priv_t* priv_p;
	struct sveth_rxpkt_t* rxpkt_p = data_p;
	sveth_ipcc_rtd_t* rtd_p;
	sveth_ipcc_ptd_t* ptd_p;

	PWARNING("    RX[X]\n");

	priv_p	= rxpkt_p->priv_p;
	rtd_p	= rxpkt_p->rtd_p;

	/* release all of RX-PTD */
	while(sveth_shmem_rtd_get_num(priv_p, rtd_p) > 0) {
		/* get RX-PTD */
		ptd_p = sveth_shmem_rtd_get_ptd(priv_p, rtd_p);
		if(ptd_p != 0) {
			/* release RX-PTD */
			sveth_shmem_ptd_free(priv_p, ptd_p);
		}
	}
}


/* tx_pktque req expiration handler */
static void _tx_req_expired(void* data_p) {
	unsigned long flags;
	struct sveth_priv_t* priv_p;
	sveth_mpool_handle_t handle = (sveth_mpool_handle_t)data_p;
	struct sveth_txpkt_t* txpkt_p = _sveth_mpool_handle_to_ptr(handle);
	sveth_ipcc_ptd_t* ptd_p;

	PINFO("    TX_PKTQUE[X] **** req expired, handle=0x%p\n", (void*)handle);

	priv_p	= txpkt_p->priv_p;
	ptd_p	= txpkt_p->ptd_p;
	sveth_ipcc_dump_priv_info(priv_p->fd);

	/* release TX mpool */
	sveth_mpool_free(&priv_p->tx_mpool, handle);

	/* update netif status */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	priv_p->stats.tx_dropped++;
	priv_p->stats.tx_errors++;
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	/* release TX-PTD */
	sveth_shmem_ptd_free(priv_p, ptd_p);

	/* resume xfer */
	if(netif_queue_stopped(priv_p->netdev_p)) {
		netif_start_queue(priv_p->netdev_p);
	}
}

/* rx_pktque req expiration handler */
static void _rx_req_expired(void* data_p) {
	unsigned long flags;
	struct sveth_priv_t* priv_p;
	sveth_mpool_handle_t handle = (sveth_mpool_handle_t)data_p;
	struct sveth_rxpkt_t* rxpkt_p = _sveth_mpool_handle_to_ptr(handle);
	sveth_ipcc_rtd_t* rtd_p;
	sveth_ipcc_ptd_t* ptd_p;

	PINFO("    RX_PKTQUE[X] **** req expired, handle=0x%p\n", (void*)handle);

	priv_p	= rxpkt_p->priv_p;
	rtd_p	= rxpkt_p->rtd_p;

	/* release RX mpool */
	sveth_mpool_free(&priv_p->rx_mpool, handle);

	/* release all of RX-PTD */
	while(sveth_shmem_rtd_get_num(priv_p, rtd_p) > 0) {
		/* get RX-PTD */
		ptd_p = sveth_shmem_rtd_get_ptd(priv_p, rtd_p);
		if(ptd_p != 0) {

			/* update netif status */
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			priv_p->stats.rx_dropped++;
			priv_p->stats.rx_errors++;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

			/* release RX-PTD */
			sveth_shmem_ptd_free(priv_p, ptd_p);

#if defined(CONFIG_SVETH_DEBUG)
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			priv_p->rx_ptd_free_cnt++;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
#endif	/* CONFIG_SVETH_DEBUG */

		}
	}
}

/* tx_pktque req cleanup handler */
static void _tx_req_cleanup(void* data_p) {
	struct sveth_priv_t* priv_p;
	sveth_mpool_handle_t handle = (sveth_mpool_handle_t)data_p;
	struct sveth_txpkt_t* txpkt_p = _sveth_mpool_handle_to_ptr(handle);
	sveth_ipcc_ptd_t* ptd_p;

	PINFO("    TX_PKTQUE[X] **** req cleanup, handle=0x%p\n", (void*)handle);

	priv_p	= txpkt_p->priv_p;
	ptd_p	= txpkt_p->ptd_p;

	/* release TX mpool */
	sveth_mpool_free(&priv_p->tx_mpool, handle);

	/* release TX-PTD */
	sveth_shmem_ptd_free(priv_p, ptd_p);
}


/* rx_pktque req cleanup handler */
static void _rx_req_cleanup(void* data_p) {
	struct sveth_priv_t* priv_p;
	sveth_mpool_handle_t handle = (sveth_mpool_handle_t)data_p;
	struct sveth_rxpkt_t* rxpkt_p = _sveth_mpool_handle_to_ptr(handle);
	sveth_ipcc_rtd_t* rtd_p;
	sveth_ipcc_ptd_t* ptd_p;

	PINFO("    RX_PKTQUE[X] **** req cleanup, handle=0x%p\n", (void*)handle);

	priv_p	= rxpkt_p->priv_p;
	rtd_p	= rxpkt_p->rtd_p;

	/* release RX mpool */
	sveth_mpool_free(&priv_p->rx_mpool, handle);

	/* release all of RX-PTD */
	while(sveth_shmem_rtd_get_num(priv_p, rtd_p) > 0) {
		/* get RX-PTD */
		ptd_p = sveth_shmem_rtd_get_ptd(priv_p, rtd_p);
		if(ptd_p != 0) {
			/* release RX-PTD */
			sveth_shmem_ptd_free(priv_p, ptd_p);
		}
	}
}



/*
 * for Event thread
 */


static int inline _sveth_abs_ev_rx_to_upper(struct net_device* netdev_p, void* data_p, int len)
{
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	struct sk_buff* skb_p = 0;
	void* skbdata_p;

	/* Build RX skb and indicate it to upper */
	skb_p = dev_alloc_skb(roundup(len + 2, 4));
	if(!skb_p) {
		PWARNING("    skb_p alloc fail.\n");
		goto _out;
	}

	/* sk_buff alloc success */
	skb_reserve(skb_p, NET_IP_ALIGN);		/* align IP on 16bit boundary */
	skbdata_p = skb_put(skb_p, len);
	sveth_assert(skbdata_p != 0);

	{
		void* vaddr_p = 0;
		vaddr_p = sveth_shmem_phys_to_virt(priv_p, data_p);

		sveth_assert(vaddr_p != 0);
		memcpy(skbdata_p, vaddr_p, len);
	}

	skb_p->protocol = eth_type_trans(skb_p, netdev_p);

	/* indicate RX skb to upper */
	if(unlikely(netif_rx_ni(skb_p)) == NET_RX_DROP) {
		/* drop packet */
		goto _out;
	}

	netdev_p->last_rx = jiffies;
	priv_p->stats.rx_packets++;
	priv_p->stats.rx_bytes += len;

	/* success */
	return 0;

_out:
	priv_p->stats.rx_dropped++;

	return -ENOMEM;
}


/* Receive a packet: retrieve, encapsulate and pass over to upper levels */
static inline void _sveth_abs_ev_rx(struct net_device* netdev_p)
{
	unsigned long flags;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;
	sveth_mpool_handle_t handle;
	struct sveth_rxpkt_t* rxpkt_p;
	sveth_ipcc_rtd_t* rtd_p;

	/* dequeue from rx_queue */
	for( /* nop */; likely(!sveth_que_is_empty(&priv_p->rx_pktque)); /* nop */ ) {

		ret = sveth_que_dequeue(&priv_p->rx_pktque, (void**)&handle);
		if(unlikely(ret != 0)) {
			PWARNING("    rx_req: no pool, exit loop.\n");
			/* rx_queue is empty */
			goto _out;
		}
	
		/* get RX-RTD from RX mpool */
		rxpkt_p = _sveth_mpool_handle_to_ptr(handle);
		sveth_assert(rxpkt_p != 0);
		rtd_p = rxpkt_p->rtd_p;
		//PINFO("    rtd_p=%p\n", rtd_p);
	
		/* release RX mpool */
		sveth_mpool_free(&priv_p->rx_mpool, handle);
	
		/* RX sequense */
		while(likely(sveth_shmem_rtd_get_num(priv_p, rtd_p) > 0)) {
			sveth_ipcc_ptd_t* ptd_p;
			void* data_p;
			int len;
			
			/* get RX-PTD */
			ptd_p = sveth_shmem_rtd_get_ptd(priv_p, rtd_p);
			if(unlikely(ptd_p == 0)) {
				/* no RX-PTD */
				PWARNING("    ptd_p is null\n");
				goto _done;
			}
	
			/* get RX len & phys-addr of data */
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			data_p = ((sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p))->data_p;
			len = ((sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p))->len;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	
			/* indicate to upper */
			ret = _sveth_abs_ev_rx_to_upper(netdev_p, data_p, len);
			if(ret < 0) {
				PWARNING("    _sveth_abs_ev_rx_to_upper fail. ret=%d\n", ret);
			}
	
			/* release RX-PTD */
			sveth_shmem_ptd_free(priv_p, ptd_p);
	
#if defined(CONFIG_SVETH_DEBUG)
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			priv_p->rx_ptd_free_cnt++;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
#endif	/* CONFIG_SVETH_DEBUG */

		}
_done:
		/* release RX-RTD */
		sveth_shmem_rtd_free(priv_p, rtd_p);
	}

	/**
	 *	for TX resume operation.
	 */
	/* check TX available (for the next request) */
	if(likely(sveth_check_send_available(priv_p->netdev_p))) {
		if(netif_queue_stopped(netdev_p)) {
			/* save the timestamp */
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			netdev_p->trans_start = jiffies;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

			PINFO("    *** resume netif queue.\n");
			netif_wake_queue(netdev_p);			/* wakeup */
		}
	}

_out:
	return;
}


/* Send a packet */
static inline int _sveth_abs_ev_tx_to_lower(struct net_device* netdev_p) {
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	unsigned long flags;
	int ret;
	sveth_mpool_handle_t handle;
	struct sveth_txpkt_t* txpkt_p;
	sveth_ipcc_rtd_t* rtd_p;
	sveth_ipcc_ptd_t* ptd_p;
	int total_len;
	int num;

	/* dequeue from tx_queue */
	total_len = 0;
	rtd_p = sveth_shmem_rtd_alloc_tx(priv_p);
	if(unlikely(rtd_p == 0)) {
		PWARNING("    no TX-RTD.\n");
		goto _out0;
	}
	
	for( /* nop */; likely(!sveth_que_is_empty(&priv_p->tx_pktque)); /* nop */ ) {

		ret = sveth_que_dequeue(&priv_p->tx_pktque, (void**)&handle);
		if(unlikely(ret != 0)) {
			PWARNING("    tx_req: no pool, exit loop.\n");
			/* no pool: exit loop */
			goto _try_to_send;
		}

		/* get TX-PTD from TX mpool */
		txpkt_p = _sveth_mpool_handle_to_ptr(handle);
		sveth_assert(txpkt_p != 0);
		ptd_p = txpkt_p->ptd_p;

		/* release TX mpool */
		sveth_mpool_free(&priv_p->tx_mpool, handle);

		/* get length */
		spin_lock_irqsave(&priv_p->sveth_lock, flags);
		total_len += ((sveth_ipcc_ptd_t*)sveth_shmem_phys_to_virt(priv_p, ptd_p))->len;
		spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

		/* set TX-PTD */
		sveth_shmem_rtd_set_ptd(priv_p, rtd_p, ptd_p);
	}

_try_to_send:
	/* try to send */
	num = sveth_shmem_rtd_get_num(priv_p, rtd_p);
	if(num > 0) {
		/* set TX handshake event and send TX packet */
		sveth_shmem_set_event(priv_p, SVETH_IPCC_VREG_EVENT_TX_REQ);
		ret = sveth_ipcc_kapi_send_packet(priv_p->fd, rtd_p);
		if(ret < 0) {
			PWARNING("    send tx packet send fail: ret=%d\n", ret);
			goto _out2;
		}

		/* check TX-RTD resource */
		if(unlikely(sveth_shmem_rtd_is_empty_tx(priv_p))) {
			PINFO("    TX handshake: start.\n");
			/* check TX handshake */
			if(unlikely(sveth_shmem_check_event(priv_p, SVETH_IPCC_VREG_EVENT_TX_REQ, SVETH_TX_DONE_CHECK_EXPIRED) != 0)) {
				/* TX handshake expired */
				/* note: release resoures not in this sequence, but in the _tx_expired */
				PWARNING("    Tx handshake: expired.\n");
				goto _out1;
			}
		}

		/* update TX length */
		spin_lock_irqsave(&priv_p->sveth_lock, flags);
		priv_p->stats.tx_packets += num;
		priv_p->stats.tx_bytes += total_len;
		spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	}
	
	/**
	 *	for TX resume operation.
	 */
	/* check TX available (for the next request) */
	if(likely(sveth_check_send_available(netdev_p))) {
		if(netif_queue_stopped(netdev_p)) {
			/* save the timestamp */
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			netdev_p->trans_start = jiffies;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

			PINFO("    *** resume netif queue.\n");
			netif_wake_queue(netdev_p);			/* wakeup */
		}
	}

	/* success */
	return 0;


_out2:
	while(likely(sveth_shmem_rtd_get_num(priv_p, rtd_p) > 0)) {
		/* update netif status */
		spin_lock_irqsave(&priv_p->sveth_lock, flags);
		priv_p->stats.tx_dropped++;
		priv_p->stats.tx_errors++;
		spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

		/* release TX shmem area */
		ptd_p = sveth_shmem_rtd_get_ptd(priv_p, rtd_p);
		if(ptd_p != 0) {
			sveth_shmem_ptd_free(priv_p, ptd_p);
		}
	}

	/* release TX-RTD */
	sveth_shmem_rtd_free(priv_p, rtd_p);

_out1:
	/* clear TX handshake */
	sveth_shmem_clear_event(priv_p, SVETH_IPCC_VREG_EVENT_TX_REQ);

_out0:
	return -EFAULT;
}


/* Send a packet */
static inline int _sveth_abs_ev_tx(struct net_device* netdev_p) {
	int ret;

	/* TX sequense */
	ret = _sveth_abs_ev_tx_to_lower(netdev_p);
	if(unlikely(ret != 0)) {
		goto _err0;
	}

	/* success */
	return 0;

_err0:
	PWARNING("    _sveth_abs_ev_tx_to_lower fail (num=1).\n");
	return -EFAULT;
}


static inline int _sveth_abs_rx_evevt_thread(void* prm_p) {
	struct net_device* netdev_p = prm_p;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;

	for(;;) {

		/* wait event */
		ret = wait_event_interruptible(
				priv_p->rx_event_qh,
				(
					kthread_should_stop() || 
					!sveth_que_is_empty(&priv_p->rx_pktque)
				)
		);
		if(ret != 0) {
			/* bad condition */
			PWARNING("    bad condition\n");
			return -ERESTART;
		}

		/* check thread condition */
		if(unlikely(kthread_should_stop())) {
			/* terminate myself */
			PNOTICE("    terminate myself.\n");
			return 0;
		}

		/* execute RX event */
		_sveth_abs_ev_rx(netdev_p);
	}

	return 0;
}

static inline int _sveth_abs_tx_evevt_thread(void* prm_p) {

	struct net_device* netdev_p = prm_p;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;

	for(;;) {

		/* wait event */
		ret = wait_event_interruptible(
				priv_p->tx_event_qh,
				(
					kthread_should_stop() || 
					!sveth_que_is_empty(&priv_p->tx_pktque)
				)
		);
		if(ret != 0) {
			/* bad condition */
			PWARNING("    bad condition\n");
			return -ERESTART;
		}

		/* check thread condition */
		if(unlikely(kthread_should_stop())) {
			/* terminate myself */
			PNOTICE("    terminate myself.\n");
			return 0;
		}

		/* execute TX event */
		_sveth_abs_ev_tx(netdev_p);
	}

	return 0;
}

/**
 *	Interrupt Control
 */
static int sveth_abs_set_interrupt_state(struct net_device* netdev_p, int enable)
{
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;

	/* ipcc[kapi]: set_interrupt_mode */
	ret = sveth_ipcc_kapi_set_interrupt_mode(priv_p->fd, enable);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_set_interrupt_mode(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* success */
	return 0;
}



/**
 *	get mac address
 */
static int _sveth_abs_get_mac(struct net_device* netdev_p) {
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	unsigned long flags;
	int ret;

	/* ipcc[kapi]: request_mac_address */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_setflg(priv_p->ipcc_stat_flg, SVETH_FLG_REQ_MAC_ADDR);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	ret = sveth_ipcc_kapi_request_mac_address(priv_p->fd, SVETH_ID_REQ_MAC_ADDR);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_request_mac_address(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* check result */
	ret = wait_event_interruptible_timeout(priv_p->ipcc_kapi_qh, !_sveth_issetflg(priv_p->ipcc_stat_flg, SVETH_FLG_REQ_MAC_ADDR), SVETH_REQ_MAC_ADDR_TO);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_request_mac_address(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* success */
	return 0;
}



/**
 *	set mac address
 */
static inline int _sveth_abs_set_mac(struct net_device* netdev_p) {
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;

	/* ipcc[kapi]: set macaddr */
	ret = sveth_ipcc_kapi_set_mac_address(priv_p->fd, netdev_p->dev_addr);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_set_mac_address(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* success */
	return 0;
}



/**
 *	notify status
 */
static inline int _sveth_abs_notify_status(struct net_device* netdev_p) {
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;

	/* ipcc[kapi]: notify status */
	ret = sveth_ipcc_kapi_notify_status(priv_p->fd, 0xFF);		/* status error */
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_notify_status(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* success */
	return 0;
}



static inline int _sveth_ipcc_ev_init_ack(int fd, const void* param_p, unsigned short errcode, volatile sveth_ipcc_vreg_t* vreg_p, unsigned long vreg_size) {
	struct net_device* netdev_p;
	struct sveth_priv_t* priv_p;
	unsigned long flags;

	if(param_p == 0) {
		return -EINVAL;
	}
	netdev_p = (struct net_device*)param_p;
	priv_p = netdev_priv(netdev_p);
	if(priv_p == 0) {
		return -EINVAL;
	}

	/* ipcc[kapi]: init */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	priv_p->vreg_p = vreg_p;
	priv_p->vreg_size = vreg_size;

	_sveth_clrflg(priv_p->ipcc_stat_flg, SVETH_FLG_INIT);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	wake_up_interruptible(&priv_p->ipcc_kapi_qh);

	return 0;
}

static inline int _sveth_ipcc_ev_quit_ack(int fd, const void* param_p, unsigned short errcode) {
	struct net_device* netdev_p;
	struct sveth_priv_t* priv_p;
	unsigned long flags;

	if(param_p == 0) {
		return -EINVAL;
	}
	netdev_p = (struct net_device*)param_p;
	priv_p = netdev_priv(netdev_p);
	if(priv_p == 0) {
		return -EINVAL;
	}

	/* ipcc[kapi]: quit */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_clrflg(priv_p->ipcc_stat_flg, SVETH_FLG_QUIT);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	wake_up_interruptible(&priv_p->ipcc_kapi_qh);

	return 0;
}


static inline int _set_rx_req_queue(struct sveth_priv_t* priv_p, sveth_ipcc_rtd_t* rtd_p) {
	unsigned long flags;
	sveth_mpool_handle_t handle;
	struct sveth_rxpkt_t* rxpkt_p;
	int ret;
	int wakeup;

	handle = sveth_mpool_alloc(&priv_p->rx_mpool);
	if(unlikely(handle == 0)) {
		/* alloc fail */
		PWARNING("    rx_mpool: alloc fail.\n");
		goto _out0;
	}

	rxpkt_p = _sveth_mpool_handle_to_ptr(handle);
	sveth_assert(rxpkt_p != 0);

	/* set rxpkt param */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	rxpkt_p->priv_p		= priv_p;
	rxpkt_p->rtd_p		= rtd_p;
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	/* check request queue */
	wakeup = sveth_que_is_empty(&priv_p->rx_pktque);
	
	ret = sveth_que_enqueue(&priv_p->rx_pktque, (void*)handle, _rx_req_expired, SVETH_RX_REQ_EXPIRED);
	if(unlikely(ret < 0)) {
		PWARNING("    rx_pktque: no pool\n");
		goto _out1;
	}

	/* success */
	sveth_mpool_set_expiration(&priv_p->rx_mpool, handle, _rx_expired, SVETH_RX_EXPIRED);

	if(unlikely(wakeup)) {
		/* activate rx_event thread */
		//PINFO("    wakeup.\n");
		wake_up_interruptible(&priv_p->rx_event_qh);
	}

	return 0;

_out1:	/* release RX mpool */
	sveth_mpool_free(&priv_p->rx_mpool, handle);

_out0:	/* error return */
	return -EFAULT;
}



static inline int _sveth_ipcc_ev_recv_pkt(int fd, const void* param_p, sveth_ipcc_rtd_t* rtd_p) {
	unsigned long flags;
	struct net_device* netdev_p;
	struct sveth_priv_t* priv_p;
	int ret;
	sveth_ipcc_ptd_t* ptd_p;

	if(param_p == 0) {
		/* invalid parameter */
		/* RX-RTD cannot release in this case */
		PWARNING("    param_p is null.\n");
		goto _err0;
	}
	netdev_p = (struct net_device*)param_p;
	priv_p = netdev_priv(netdev_p);
	
	if(priv_p == 0) {
		/* invalid priv_p */
		/* RX-RTD cannot release in this case */
		PWARNING("    priv_p is null.\n");
		goto _err0;
	}

#if defined(CONFIG_SVETH_DEBUG)
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	priv_p->ipcc_ev_recv_pkt_cnt++;
	priv_p->rx_ptd_recv_cnt += sveth_shmem_rtd_get_num(priv_p, rtd_p);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
#endif	/* CONFIG_SVETH_DEBUG */	/* recv sequence */

	ret = _set_rx_req_queue(priv_p, rtd_p);
	if(ret < 0) {
		/* no pool */
		PWARNING("    no rx_pool.\n");
		goto _err1;
	}

	/* success */
	return 0;


_err1:
	/* drop RX packets */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	priv_p->stats.rx_dropped += sveth_shmem_rtd_get_num(priv_p, rtd_p);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	
	/* release all of RX-PTD */
	while(likely(sveth_shmem_rtd_get_num(priv_p, rtd_p) > 0)) {
		/* get RX-PTD */
		ptd_p = sveth_shmem_rtd_get_ptd(priv_p, rtd_p);
		if(ptd_p != 0) {
			/* release RX-PTD */
			sveth_shmem_ptd_free(priv_p, ptd_p);

#if defined(CONFIG_SVETH_DEBUG)
			spin_lock_irqsave(&priv_p->sveth_lock, flags);
			priv_p->rx_ptd_free_cnt++;
			spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
#endif	/* CONFIG_SVETH_DEBUG */

		}
	}

	/* release RX-RTD */
	sveth_shmem_rtd_free(priv_p, rtd_p);

_err0:
	return -EINVAL;
}


static inline int _sveth_ipcc_ev_stat_ind(int fd, const void* param_p, unsigned char prm) {
	unsigned long flags;
	struct net_device* netdev_p;
	struct sveth_priv_t* priv_p;

	if(param_p == 0) {
		return -EINVAL;
	}
	netdev_p = (struct net_device*)param_p;
	priv_p = netdev_priv(netdev_p);
	if(priv_p == 0) {
		return -EINVAL;
	}

	/* ipcc[kapi]: stat_ind */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_clrflg(priv_p->ipcc_stat_flg, SVETH_FLG_NOTIFY_STAT);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	wake_up_interruptible(&priv_p->ipcc_kapi_qh);

	return 0;
}


static inline int _sveth_ipcc_ev_get_int_stat(int fd, const void* param_p, unsigned short errcode, unsigned short req_id, unsigned char flag) {
	unsigned long flags;
	struct net_device* netdev_p;
	struct sveth_priv_t* priv_p;

	if(param_p == 0) {
		return -EINVAL;
	}
	netdev_p = (struct net_device*)param_p;
	priv_p = netdev_priv(netdev_p);
	if(priv_p == 0) {
		return -EINVAL;
	}

	if(req_id != SVETH_ID_REQ_INT_STAT) {
		PWARNING("    invalid req_id. fd=%d, req_id=0x%04x\n", fd, req_id);
		return -EINVAL;
	}

	/* ipcc[kapi]: int_stat */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_clrflg(priv_p->ipcc_stat_flg, SVETH_FLG_REQ_INT_STAT);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	wake_up_interruptible(&priv_p->ipcc_kapi_qh);

	return 0;
}


static inline int _sveth_ipcc_ev_get_macaddr(int fd, const void* param_p, unsigned short errcode, unsigned short req_id, unsigned char* macaddr_p) {
	unsigned long flags;
	struct net_device* netdev_p = (struct net_device*)param_p;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);

	if(param_p == 0) {
		return -EINVAL;
	}
	netdev_p = (struct net_device*)param_p;
	priv_p = netdev_priv(netdev_p);
	if(priv_p == 0) {
		return -EINVAL;
	}
	
	if(req_id != SVETH_ID_REQ_MAC_ADDR) {
		PWARNING("    invalid req_id. fd=%d, req_id=0x%04x\n", fd, req_id);
		return -EINVAL;
	}

	/* ipcc[kapi]: get_macaddr */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	/* store mac address */
	memcpy(netdev_p->dev_addr, macaddr_p, ETH_ALEN);
	_sveth_clrflg(priv_p->ipcc_stat_flg, SVETH_FLG_REQ_MAC_ADDR);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	wake_up_interruptible(&priv_p->ipcc_kapi_qh);

	return 0;
}


/* for sveth_ipcc event */
static const struct sveth_ipcc_event_t ipcc_event = {
	.pf_init_ack			= _sveth_ipcc_ev_init_ack,
	.pf_quit_ack			= _sveth_ipcc_ev_quit_ack,
	.pf_recv_pkt			= _sveth_ipcc_ev_recv_pkt,
	.pf_stat_ind			= _sveth_ipcc_ev_stat_ind,
	.pf_get_int_stat		= _sveth_ipcc_ev_get_int_stat,
	.pf_get_macaddr			= _sveth_ipcc_ev_get_macaddr,
	.pf_get_link_stat		= 0,
};



/**
 *	open
 */
static inline int _open_process(struct net_device* netdev_p) {
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;
	unsigned long flags;


	/**
	 *	init privates
	 */	

	/* initialize TX/RX Event thread queue head */
	priv_p->tx_event_thread_p = 0;
	priv_p->rx_event_thread_p = 0;
	init_waitqueue_head(&priv_p->tx_event_qh);
	init_waitqueue_head(&priv_p->rx_event_qh);

	/* init TX/RX mpool */
	sveth_mpool_init(&priv_p->tx_mpool);
	sveth_mpool_init(&priv_p->rx_mpool);

	/* init TX/RX packet queue */
	sveth_que_init(&priv_p->tx_pktque);
	sveth_que_init(&priv_p->rx_pktque);


	/* ipcc[dd]: open */
	ret = sveth_ipcc_dd_open(opencount | 0x10, priv_p->ipcc_event_p, netdev_p);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_dd_open fail. err=%d\n", ret);
		return -EFAULT;
	}
	/* success */
	priv_p->fd = ret;

	/* ipcc[kapi]: init */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_setflg(priv_p->ipcc_stat_flg, SVETH_FLG_INIT);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	ret = sveth_ipcc_kapi_init(priv_p->fd);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_init(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}
	ret = wait_event_interruptible_timeout(priv_p->ipcc_kapi_qh, !_sveth_issetflg(priv_p->ipcc_stat_flg, SVETH_FLG_INIT), SVETH_INIT_TO);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_init(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* get Tx / Rx shared memory */
	PNOTICE("sizeof(sveth_ipcc_pkt_t)=%zd\n", sizeof(sveth_ipcc_pkt_t));
	PNOTICE("******** VREG INFO ********\n");
	PNOTICE("    vreg_p=%p\n",			priv_p->vreg_p);
	PNOTICE("    vreg_size=%lu\n",		priv_p->vreg_size);

	if(priv_p->vreg_p == 0 || priv_p->vreg_size == 0) {
		/* fail */
		PWARNING("vreg get fail. vreg_p=%p, vreg_size=%lu\n", priv_p->vreg_p, priv_p->vreg_size);
		return -EFAULT;
	}

	priv_p->vreg_resource_p = request_mem_region((unsigned long)priv_p->vreg_p, priv_p->vreg_size, SVETH_DRV_DESC);
	if(!priv_p->vreg_resource_p) {
		/* fail */
		PWARNING("request_mem_region fail.\n");
		return -EFAULT;
	}

	priv_p->vreg_vaddr_p = ioremap_nocache((unsigned long)priv_p->vreg_p, priv_p->vreg_size);
	if(unlikely(priv_p->vreg_vaddr_p == 0)) {
		/* fail */
		PWARNING("ioremap_nocache failed.\n");
		return -EFAULT;
	}

	PNOTICE("    version=%lu\n",			priv_p->vreg_vaddr_p->version);
	PNOTICE("    event=%lu\n",				priv_p->vreg_vaddr_p->event);

	PNOTICE("    TX-RTD area top=0x%lx\n",	priv_p->vreg_vaddr_p->tx_rtd_top);
	PNOTICE("    TX-RTD size=%lu\n",		priv_p->vreg_vaddr_p->tx_rtd_size);
	PNOTICE("    TX-RTD num=%lu\n",			priv_p->vreg_vaddr_p->tx_rtd_num);

	PNOTICE("    RX-RTD area top=0x%lx\n",	priv_p->vreg_vaddr_p->rx_rtd_top);
	PNOTICE("    RX-RTD size=%lu\n",		priv_p->vreg_vaddr_p->rx_rtd_size);
	PNOTICE("    RX-RTD num=%lu\n",			priv_p->vreg_vaddr_p->rx_rtd_num);

	PNOTICE("    TX-PTD area top=0x%lx\n",	priv_p->vreg_vaddr_p->tx_ptd_top);
	PNOTICE("    TX-PTD size=%lu\n",		priv_p->vreg_vaddr_p->tx_ptd_size);
	PNOTICE("    TX-PTD num=%lu\n",			priv_p->vreg_vaddr_p->tx_ptd_num);

	PNOTICE("    RX-PTD area top=0x%lx\n",	priv_p->vreg_vaddr_p->rx_ptd_top);
	PNOTICE("    RX-PTD size=%lu\n",		priv_p->vreg_vaddr_p->rx_ptd_size);
	PNOTICE("    RX-PTD num=%lu\n",			priv_p->vreg_vaddr_p->rx_ptd_num);



	/* setup TX/RX mpool */
	sveth_mpool_setup(&priv_p->tx_mpool, priv_p->vreg_vaddr_p->tx_ptd_num, sizeof(struct sveth_txpkt_t));
	sveth_mpool_setup(&priv_p->rx_mpool, priv_p->vreg_vaddr_p->rx_ptd_num, sizeof(struct sveth_rxpkt_t));

	/* setup TX/RX packet queue */
	sveth_que_setup(&priv_p->tx_pktque, priv_p->vreg_vaddr_p->tx_ptd_num);
	sveth_que_setup(&priv_p->rx_pktque, priv_p->vreg_vaddr_p->rx_ptd_num);

	/* initialize RX Event thread */
	priv_p->rx_event_thread_p = kthread_run(_sveth_abs_rx_evevt_thread, netdev_p, "%s", "_sveth_abs_rx_evevt_thread");
	if(IS_ERR(priv_p->rx_event_thread_p)) {
		PWARNING("    Error: creating the Rx Event thread\n");
		return -EAGAIN;
	}

	/* initialize TX Event thread */
	priv_p->tx_event_thread_p = kthread_run(_sveth_abs_tx_evevt_thread, netdev_p, "%s", "_sveth_abs_tx_evevt_thread");
	if(IS_ERR(priv_p->tx_event_thread_p)) {
		PWARNING("    Error: creating the Tx Event thread\n");
		return -EAGAIN;
	}

	/* get macaddr */
	_sveth_abs_get_mac(netdev_p);

	/* success */
	return 0;
}

static int sveth_ndo_open(struct net_device* netdev_p)
{
	int ret;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);

	/* check opened */
	if(priv_p->opened) {
		/* already opend */
		PWARNING("    already opend.\n");
		_release_process(netdev_p);			/* release */
	}
	
	/* open process */
	ret = _open_process(netdev_p);
	if(ret < 0) {
		/* fail */
		PWARNING("    open fail. err=%d\n", ret);
		_release_process(netdev_p);			/* release */
		return -EFAULT;
	}

	/* enable Interrupt */
	sveth_abs_set_interrupt_state(netdev_p, 1);

	/* start queue */
	netif_start_queue(netdev_p);

	priv_p->opened = 1;

	return 0;
}

/**
 *	close
 */
static inline int _release_process(struct net_device* netdev_p) {
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	int ret;
	unsigned long flags;

	/* ipcc[kapi]: quit */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	_sveth_setflg(priv_p->ipcc_stat_flg, SVETH_FLG_QUIT);
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	ret = sveth_ipcc_kapi_quit(priv_p->fd);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_quit(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}
	ret = wait_event_interruptible_timeout(priv_p->ipcc_kapi_qh, !_sveth_issetflg(priv_p->ipcc_stat_flg, SVETH_FLG_QUIT), SVETH_QUIT_TO);
	if(ret < 0) {
		/* fail */
		PWARNING("    sveth_ipcc_kapi_quit(%d) fail. err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* ipcc[dd]: close */
	ret = sveth_ipcc_dd_close(priv_p->fd);
	if(ret < 0) {
		PWARNING("    sveth_ipcc_dd_close fail. fd=%d, err=%d\n", priv_p->fd, ret);
		return -EFAULT;
	}

	/* Stop the Rx thread before we exit */
	if(priv_p->rx_event_thread_p != 0) {
		kthread_stop(priv_p->rx_event_thread_p);
	}

	/* Stop the Tx thread before we exit */
	if(priv_p->tx_event_thread_p != 0) {
		kthread_stop(priv_p->tx_event_thread_p);
	}

	/* cleanup TX/RX packet queue */
	sveth_que_cleanup(&priv_p->tx_pktque, _tx_req_cleanup);
	sveth_que_cleanup(&priv_p->rx_pktque, _rx_req_cleanup);

	/* cleanup TX/RX mpool */
	sveth_mpool_cleanup(&priv_p->tx_mpool, _tx_cleanup);
	sveth_mpool_cleanup(&priv_p->rx_mpool, _rx_cleanup);

	if(priv_p->vreg_resource_p) {
		release_resource(priv_p->vreg_resource_p);
	}
	if(priv_p->vreg_vaddr_p != 0) {
		iounmap(priv_p->vreg_vaddr_p);
	}

	/* success */
	return 0;
}


static int sveth_ndo_release(struct net_device* netdev_p)
{
	int ret;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	/* release ports, irq and such -- like fops->close */

#if defined(CONFIG_SVETH_DEBUG)
	/* dump */
	{
		if(netif_queue_stopped(netdev_p)) {
			PINFO("    netif queue is in stopped\n");
		}
		else {
			PINFO("    netif queue is in active\n");
		}

	
		PINFO("    rx_ptd_recv_cnt=%lu\n", priv_p->rx_ptd_recv_cnt);
		PINFO("    rx_ptd_free_cnt=%lu\n", priv_p->rx_ptd_free_cnt);

		PINFO("    ipcc_ev_recv_pkt_cnt=%lu\n", priv_p->ipcc_ev_recv_pkt_cnt);
	
		sveth_shmem_rtd_check_tx(priv_p);
		sveth_shmem_rtd_check_rx(priv_p);

		sveth_shmem_ptd_check_tx(priv_p);
		sveth_shmem_ptd_check_rx(priv_p);
	
		PINFO("    dump: ipcc kapi TX/RX num.\n");
		sveth_ipcc_dump_priv_info(priv_p->fd);

		PINFO("    dump: tx_pktque\n");
		sveth_que_dump_priv_info(&priv_p->tx_pktque);
		PINFO("    dump: rx_pktque\n");
		sveth_que_dump_priv_info(&priv_p->rx_pktque);

		PINFO("    dump: tx_mpool\n");
		sveth_mpool_dump_priv_info(&priv_p->tx_mpool);
		PINFO("    dump: rx_mpool\n");
		sveth_mpool_dump_priv_info(&priv_p->rx_mpool);
	}
#endif	/* CONFIG_SVETH_DEBUG */


	PWARNING("    REL[1] **** netif_stop_queue !!!\n");
	netif_stop_queue(netdev_p);		/* can't transmit any more */

	/* disable Interrupts */
	sveth_abs_set_interrupt_state(netdev_p, 0);

	/* release process */
	ret = _release_process(netdev_p);
	if(ret < 0) {
		/* fail */
		PWARNING("    release fail. err=%d\n", ret);
		return -EFAULT;
	}

	priv_p->opened = 0;

	return 0;
}


static int sveth_check_send_available(struct net_device* netdev_p) {
	int ret;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);

	/* check mpool & queue */
	ret =	unlikely(sveth_mpool_is_empty(&priv_p->tx_mpool)) || 
			unlikely(sveth_que_is_full(&priv_p->tx_pktque)) || 
			unlikely(sveth_shmem_ptd_is_empty_tx(priv_p));

	return (ret == 0);
}

/*
 * request TX
 */
static int sveth_send_req(struct net_device* netdev_p, void* addr_p, unsigned int len)
{
	unsigned long flags;
	int ret;
	struct sveth_txpkt_t* txpkt_p;
	sveth_mpool_handle_t handle;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	sveth_ipcc_ptd_t* ptd_p;
	int wakeup;

	sveth_assert(addr_p != 0);
	if(len > ETH_FRAME_LEN) {
		/* Jumbo-frames are not supported. */
		PWARNING("    Jumbo-frames are not supported.\n");
		goto _out0;
	}

	/* get unused TX-PTD */
	ptd_p = sveth_shmem_ptd_alloc_tx(priv_p, addr_p, len);
	if(ptd_p == 0) {
		/* alloc fail */
		PWARNING("    TX shmem alloc fail.\n");
		goto _out0;
	}

	/* allocate from mpool */
	handle = sveth_mpool_alloc(&priv_p->tx_mpool);
	if(unlikely(handle == 0)) {
		/* alloc fail */
		PWARNING("    tx_mpool alloc fail.\n");
		goto _out1;
	}

	/* set tx_pkt param */
	txpkt_p = _sveth_mpool_handle_to_ptr(handle);
	sveth_assert(txpkt_p != 0);

	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	txpkt_p->priv_p		= priv_p;
	txpkt_p->ptd_p		= ptd_p;
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	/* check request queue */
	wakeup = sveth_que_is_empty(&priv_p->tx_pktque);
	
	/* set request queue */
	ret = sveth_que_enqueue(&priv_p->tx_pktque, (void*)handle, _tx_req_expired, SVETH_TX_REQ_EXPIRED);
	if(unlikely(ret != 0)) {
		/* send fail */
		PWARNING("    send fail.\n");
		goto _out2;
	}

	/* success */
	sveth_mpool_set_expiration(&priv_p->tx_mpool, handle, _tx_expired, SVETH_TX_EXPIRED);

	/* check wakeup tx_event_qh or not */
	if(unlikely(wakeup)) {
		/* activate tx_event thread */
		//PINFO("    wakeup.\n");
		wake_up_interruptible(&priv_p->tx_event_qh);
	}

	return 0;


_out2:	/* release TX mpool */
	sveth_mpool_dump_priv_info(&priv_p->tx_mpool);
	sveth_mpool_free(&priv_p->tx_mpool, handle);

_out1:
	/* release unused TX-PTD */
	PWARNING("    TX[1] **** release TX-PTD=0x%p\n", ptd_p);
	sveth_shmem_ptd_free(priv_p, ptd_p);

_out0:
	return -EFAULT;
}


/*
 * Transmit a packet (called by the kernel)
 */
static netdev_tx_t sveth_ndo_start_xmit(struct sk_buff* skb_p, struct net_device* netdev_p)
{
	unsigned long flags;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);


	/**
	 *	for TX suspend operation.
	 */
	/* check TX available (to next request) */
	if(unlikely(!sveth_check_send_available(netdev_p))) {
		/* error: no resource */
		PWARNING("    *** no resource.\n");
		goto _out0;
	}

	/* save the timestamp */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	netdev_p->trans_start = jiffies;
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);

	/* transmit packet (request to event thread) */
	if(sveth_send_req(netdev_p, skb_p->data, skb_p->len) < 0) {
		/* Tx fail */
		PWARNING("    *** Tx fail.\n");
		goto _out0;
	}

	/* success */

	/* release TX skb */
	dev_kfree_skb(skb_p);
	
	/**
	 *	for TX suspend operation.
	 */
	/* check available (for the next request) */
	if(unlikely(!sveth_check_send_available(netdev_p))) {
		/* suspend netif queue because of not available to next request */
		PINFO("    *** suspend netif queue.\n");
		netif_stop_queue(netdev_p);
	}

	return NETDEV_TX_OK;


_out0:
	/* suspend netif queue */
	netif_stop_queue(netdev_p);

#if defined(CONFIG_SVETH_TX_BUSY_RETRY)
	return NETDEV_TX_BUSY;

#else	/* CONFIG_SVETH_TX_BUSY_RETRY */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	priv_p->stats.tx_dropped++;
	priv_p->stats.tx_errors++;
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	dev_kfree_skb(skb_p);

	return NETDEV_TX_OK;
#endif	/* CONFIG_SVETH_TX_BUSY_RETRY */
}


/*
 * Deal with a tx timeout.
 */
static void sveth_ndo_tx_timeout(struct net_device* netdev_p)
{
	int ret;
	
	PWARNING("    Transmit timeout at %ld, latency %ld\n", jiffies, jiffies - netdev_p->trans_start);

	/* save the timestamp */
	netdev_p->trans_start = jiffies;
	netif_wake_queue(netdev_p);
}





/*
 * Ioctl commands 
 */
static int sveth_ndo_ioctl(struct net_device* netdev_p, struct ifreq* rq, int cmd)
{
	PDEBUG("ioctl\n");
	return 0;
}


static int sveth_ndo_set_mac_address(struct net_device *netdev_p, void *addr_p)
{
	struct sockaddr *sockaddr_p = addr_p;

	memcpy(netdev_p->dev_addr, sockaddr_p->sa_data, netdev_p->addr_len);
	_sveth_abs_set_mac(netdev_p);

	return 0;
}


/*
 * Return statistics to the caller
 */
static struct net_device_stats* sveth_ndo_get_stats(struct net_device* netdev_p)
{
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);
	return &priv_p->stats;
}

/*
 * The "change_mtu" method is usually not needed.
 * If you need it, it must be like this.
 */
static int sveth_ndo_change_mtu(struct net_device* netdev_p, int new_mtu)
{
	unsigned long flags;
	struct sveth_priv_t* priv_p = netdev_priv(netdev_p);

	/* check ranges */
	if((new_mtu < SVETH_MIN_PKTSIZE) || (new_mtu > SVETH_MAX_PKTSIZE))
		return -EINVAL;
	/*
	 * Do anything you need, and the accept the value
	 */
	spin_lock_irqsave(&priv_p->sveth_lock, flags);
	netdev_p->mtu = new_mtu;
	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
	return 0; /* success */
}

static const struct net_device_ops sveth_netdev_ops = {
	.ndo_open				= sveth_ndo_open,
	.ndo_stop				= sveth_ndo_release,
	.ndo_do_ioctl			= sveth_ndo_ioctl,
	.ndo_start_xmit			= sveth_ndo_start_xmit,
	.ndo_set_mac_address	= sveth_ndo_set_mac_address,
	.ndo_get_stats			= sveth_ndo_get_stats,
	.ndo_change_mtu			= sveth_ndo_change_mtu,
	.ndo_tx_timeout			= sveth_ndo_tx_timeout,
};

/*
 * The init function (sometimes called probe).
 * It is invoked by register_netdev()
 */
void sveth_init(struct net_device* netdev_p)
{
	struct sveth_priv_t* priv_p;
	unsigned long flags;
	unsigned char macaddr[] = SVETH_DEFAULT_MAC_ADDR;

	/* 
	 * Then, assign other fields in netdev_p, using ether_setup() and some
	 * hand assignments
	 */
	ether_setup(netdev_p); /* assign some of the fields */

	netdev_p->netdev_ops		= &sveth_netdev_ops;
	netdev_p->watchdog_timeo	= timeout;

	/*
	 * Then, initialize the priv_p field. This encloses the statistics
	 * and a few private fields.
	 */
	priv_p = netdev_priv(netdev_p);
	memset(priv_p, 0, sizeof(struct sveth_priv_t));

	/* init ipcc private */
	spin_lock_init(&priv_p->sveth_lock);
	spin_lock_irqsave(&priv_p->sveth_lock, flags);

	init_waitqueue_head(&priv_p->ipcc_kapi_qh);

	priv_p->netdev_p = netdev_p;
	priv_p->ipcc_event_p = (sveth_ipcc_event_t*)&ipcc_event;

	/* set default MAC address */
	memcpy(netdev_p->dev_addr, macaddr, ETH_ALEN);

	spin_unlock_irqrestore(&priv_p->sveth_lock, flags);
}


/*
 * Finally, the module stuff
 */
void __exit sveth_cleanup_module(void)
{
	struct net_device** netdev_pp;

	_sveth_foreach(netdev_pp, g_netdev_table) {
		if(netdev_pp != 0 && *netdev_pp != 0) {
			unregister_netdev(*netdev_pp);
			free_netdev(*netdev_pp);
		}
	}

	/* exit ipcc */
	sveth_ipcc_dd_exit();
}

int __init sveth_init_module(void)
{
	int ret;
	struct net_device** netdev_pp;

	/* init ipcc */
	sveth_ipcc_dd_init();

	/* Init devices */
	_sveth_foreach(netdev_pp, g_netdev_table) {
		*netdev_pp = NULL;
	}
	
	/* allocate devices */
	_sveth_foreach(netdev_pp, g_netdev_table) {
		*netdev_pp = alloc_netdev(sizeof(struct sveth_priv_t), SVETH_NDEV_NAME, sveth_init);
		if(*netdev_pp == NULL) {
			ret = -ENOMEM;
			goto exit_fail;
		}
	}

	/* register devices to the kernel */
	_sveth_foreach(netdev_pp, g_netdev_table) {
		ret = register_netdev(*netdev_pp);
		if(ret < 0) {
			PWARNING("    error %i registering device \"%s\"\n", ret, (*netdev_pp)->name);
			ret = -ENODEV;
			goto exit_fail;
		}
	}
	
	/* success */
	return 0;


/* fail */
exit_fail:
	/* cleanup devices */
	sveth_cleanup_module();
	return ret;
}


module_init(sveth_init_module);
module_exit(sveth_cleanup_module);

