/*
 *  sveth_mpool.c
 *
 *	The virtual ethernet driver for the SPiCa movie system.
 *
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 */

#if defined(CONFIG_SVETH_DEBUG)
static const char* devname_p = "sveth_mpool";
#endif	/* CONFIG_SVETH_DEBUG */

#include "sveth_config.h"
#include "sveth_debug.h"
#include "sveth_mpool.h"



/**
 *	@brief		_sveth_mpool_init_private
 *	
 *	@param[in,out]	*que_p:		mempool queue
 *	
 *	@return			none
 *	
 *	@details		initialize the mempool
 */
static inline void _sveth_mpool_init_private(struct sveth_mpool_que_t* que_p) {
	/* init private */
	que_p->node_p = 0;
	que_p->data_p = 0;

	que_p->free.num = 0;
	que_p->free.cnt = 0;

	que_p->used.num = 0;
	que_p->used.cnt = 0;

	INIT_LIST_HEAD(&que_p->free.head);
	INIT_LIST_HEAD(&que_p->used.head);
}


/**
 *	@brief		_timer_handler
 *	
 *	@param[in]	data:	extra data of timer handler
 *	
 *	@return		none
 *	
 *	@details	timer handler
 */
static void _timer_handler(unsigned long data) {
	unsigned long flags;
	struct sveth_mpool_qn_t* node_p = (struct sveth_mpool_qn_t*)data;
	struct sveth_mpool_que_t* que_p = node_p->que_p;
	pf_sveth_mpool_tm_handler tm_handler;
	void* data_p;
	
	/* store param & clear param */
	write_lock_irqsave(&que_p->lock, flags);
	tm_handler	= node_p->tm_handler;
	data_p		= node_p->data_p;

	node_p->tm_handler	= 0;
	write_unlock_irqrestore(&que_p->lock, flags);
	
	/* free this */
	write_lock_irqsave(&que_p->lock, flags);
	list_move_tail(&node_p->this, &que_p->free.head);
	que_p->free.num++;
	que_p->free.cnt++;
	
	que_p->used.num--;
	write_unlock_irqrestore(&que_p->lock, flags);

	/* call user-handler */
	tm_handler(data_p);
}


/**
 *	@brief		sveth_mpool_dump_priv_info
 *	
 *	@param[in,out]	*que_p:		mempool queue
 *	
 *	@return			none
 *	
 *	@details		print the internal value (for debuging)
 */
void sveth_mpool_dump_priv_info(struct sveth_mpool_que_t* que_p) {
	unsigned long flags;
	read_lock_irqsave(&que_p->lock, flags);
	PNOTICE("  que_p=0x%p\n", que_p);
	PNOTICE("    free: num=%zd, cnt=%zd\n", que_p->free.num, que_p->free.cnt);
	PNOTICE("    used: num=%zd, cnt=%zd\n", que_p->used.num, que_p->used.cnt);
	read_unlock_irqrestore(&que_p->lock, flags);
}



/**
 *	@brief		sveth_mpool_init
 *	
 *	@param[in,out]	*que_p:		mempool queue
 *	
 *	@return		int				0 on success
 *								<0 on failure
 *	
 *	@details		setting up the mempool
 */
void sveth_mpool_init(struct sveth_mpool_que_t* que_p) {
	unsigned long flags;

	rwlock_init(&que_p->lock);

	/* init private */
	write_lock_irqsave(&que_p->lock, flags);
	_sveth_mpool_init_private(que_p);
	write_unlock_irqrestore(&que_p->lock, flags);
}

/**
 *	@brief		sveth_mpool_setup
 *	
 *	@param[in,out]	*que_p:		mempool queue
 *	@param[in]		num:		num of nodes
 *	@param[in]		spn:		size per node
 *	
 *	@return		int				0 on success
 *								<0 on failure
 *	
 *	@details		setting up the mempool
 */
int sveth_mpool_setup(struct sveth_mpool_que_t* que_p, size_t num, size_t spn) {
	unsigned long flags;
	int ret;
	size_t i;
	struct sveth_mpool_qn_t* node_p;
	void* data_p;
	size_t aligned_spn;

	rwlock_init(&que_p->lock);

	/* init private */
	write_lock_irqsave(&que_p->lock, flags);
	_sveth_mpool_init_private(que_p);
	write_unlock_irqrestore(&que_p->lock, flags);

	/* allocate node */
	aligned_spn = _sveth_align(sizeof(*node_p), SVETH_MPOOL_ALIGN);
	node_p = kmalloc(aligned_spn * num, GFP_KERNEL);
	if(node_p == 0) {
		/* no memory */
		ret = -ENOMEM;
		goto _err0;
	}

	/* allocate user field */
	aligned_spn = _sveth_align(spn, SVETH_MPOOL_ALIGN);
	data_p = kmalloc(aligned_spn * num, GFP_KERNEL);
	if(data_p == 0) {
		/* no memory */
		ret = -ENOMEM;
		goto _err1;
	}
	
	/* init node_p and add to pool */
	write_lock_irqsave(&que_p->lock, flags);
	for(i = 0; i < num; i++) {

		INIT_LIST_HEAD(&node_p[i].this);
		
		node_p[i].size			= spn;
		node_p[i].data_p		= data_p + (aligned_spn * i);

		node_p[i].tm_handler	= 0;
		node_p[i].que_p			= que_p;

//		PNOTICE("node_p=%p, node_p->data_p=%p\n", node_p, node_p->data_p);

		/* add to tail */
		list_add_tail(&node_p[i].this, &que_p->free.head);
		que_p->free.num++;
	}
	write_unlock_irqrestore(&que_p->lock, flags);

	write_lock_irqsave(&que_p->lock, flags);
	que_p->node_p = node_p;
	que_p->data_p = data_p;
	write_unlock_irqrestore(&que_p->lock, flags);
	
	/* success */
	return 0;


_err1:	/* err1: release node_p */
	kfree(node_p);

_err0:	/* err0: fail */
	return ret;
}



/**
 *	@brief		_cleanup
 *	
 *	@param[in,out]	*qh_p	quehead of mempool
 *	
 *	@return		none
 *	
 *	@details		cleanup the mempool
 */
static inline void _cleanup(struct sveth_mpool_qh_t* qh_p, pf_sveth_mpool_cl_handler cl_handler) {
	struct sveth_mpool_qn_t* node_p;
	struct sveth_mpool_qn_t* tmp_p;
	
	list_for_each_entry_safe(node_p, tmp_p, &qh_p->head, this) {
		if(node_p != 0) {
//			PNOTICE("node_p=%p", node_p);

			/* check expiration */
			if(node_p->tm_handler) {
				/* cancel timer */
				del_timer_sync(&node_p->timer);
				node_p->tm_handler	= 0;
			}
			
			/* execute cleanup handler */
			if(cl_handler != 0) {
				cl_handler(node_p->data_p);
			}
			
			node_p->que_p	= 0;
			node_p->data_p	= 0;
			node_p->size	= 0;
			
			list_del(&node_p->this);
			qh_p->num--;
		}
	}
}


/**
 *	@brief		sveth_mpool_cleanup
 *	
 *	@param[in,out]	*que_p	mempool queue
 *	
 *	@return			none
 *	
 *	@details		cleanup mempool
 */
void sveth_mpool_cleanup(struct sveth_mpool_que_t* que_p, pf_sveth_mpool_cl_handler cl_handler) {
	unsigned long flags;

	struct sveth_mpool_qn_t* node_p;
	void* data_p;

	/* free: used */
	write_lock_irqsave(&que_p->lock, flags);
	_cleanup(&que_p->used, cl_handler);
	write_unlock_irqrestore(&que_p->lock, flags);

	/* free: free */
	write_lock_irqsave(&que_p->lock, flags);
	_cleanup(&que_p->free, 0);
	write_unlock_irqrestore(&que_p->lock, flags);

	/* free: user field and node_p */
	write_lock_irqsave(&que_p->lock, flags);
	data_p = que_p->data_p;
	node_p = que_p->node_p;
	write_unlock_irqrestore(&que_p->lock, flags);
	kfree(data_p);
	kfree(node_p);

	/* init private */
	write_lock_irqsave(&que_p->lock, flags);
	_sveth_mpool_init_private(que_p);
	write_unlock_irqrestore(&que_p->lock, flags);

}



/**
 *	@brief		sveth_mpool_alloc
 *	
 *	@param[in,out]	*que_p		mempool queue
 *	
 *	@return		sveth_mpool_handle_t	handle of mempool
 *	
 *	@details					allocate a new mempool
 */
sveth_mpool_handle_t sveth_mpool_alloc(struct sveth_mpool_que_t* que_p) {
	unsigned long flags;
	struct sveth_mpool_qn_t* node_p;
	int empty;

	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->free.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	if(unlikely(empty)) {
		PDEBUG("NG: unavailable from free queue. (que_p=0x%p)\n", (void*)que_p);
		/* unavailable from free queue */
		return 0;
	}

	/* get empty from free */
	read_lock_irqsave(&que_p->lock, flags);
	node_p = list_first_entry(&que_p->free.head, struct sveth_mpool_qn_t, this);
	read_unlock_irqrestore(&que_p->lock, flags);
	sveth_assert(node_p != 0);
	sveth_assert(node_p->data_p != 0);


	/* init local param */
	write_lock_irqsave(&que_p->lock, flags);
	node_p->tm_handler	= 0;
	write_unlock_irqrestore(&que_p->lock, flags);


	/* move to used queue */
	write_lock_irqsave(&que_p->lock, flags);
	list_move_tail(&node_p->this, &que_p->used.head);
	que_p->free.num--;
	que_p->used.num++;
	que_p->used.cnt++;
	write_unlock_irqrestore(&que_p->lock, flags);

	// PDEBUG("OK: handle=0x%p\n", (void*)&node_p->data_p);
	return &node_p->data_p;
}


/**
 *	@brief		sveth_mpool_free
 *	
 *	@param[in,out]	*que_p		mempool queue
 *	@param[in]		handle		handle of mempool
 *	
 *	@return		int				0 on success
 *								<0 on failure
 *	@details		release allocated mempool
 */
int sveth_mpool_free(struct sveth_mpool_que_t* que_p, sveth_mpool_handle_t handle) {
	unsigned long flags;
	struct sveth_mpool_qn_t* node_p;
	struct sveth_mpool_qn_t* tmp_p;
	int empty;

	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->used.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	if(unlikely(empty)) {
		/* no used */
		PDEBUG("NG: used queue is empty. (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);
		return -ENOMEM;
	}

	write_lock_irqsave(&que_p->lock, flags);
	list_for_each_entry_safe(node_p, tmp_p, &que_p->used.head, this) {
		if(likely(node_p != 0) && unlikely(&node_p->data_p == (void**)handle)) {
			/* check expiration */
			if(node_p->tm_handler) {
				/* cancel timer */
				del_timer_sync(&node_p->timer);
				node_p->tm_handler	= 0;
			}
		
			/* find: free from used queue */
			list_move_tail(&node_p->this, &que_p->free.head);
			que_p->free.num++;
			que_p->free.cnt++;
			que_p->used.num--;
			/* success */
			goto _done;
		}
	}
	/* not found */
	write_unlock_irqrestore(&que_p->lock, flags);
	PDEBUG("NG: not found. (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);
	return -ENOMEM;

_done:
	write_unlock_irqrestore(&que_p->lock, flags);
	// PDEBUG("OK: (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);

	return 0;
}



/**
 *	@brief		sveth_mpool_set_expiration
 *	
 *	@param[in,out]	*que_p			mempool queue
 *	@param[in]		handle			handle of mempool
 *	@param[in]		tm_handler		timer handler
 *	@param[in]		expires			expiration time
 *	
 *	@return		int					0 on success
 *									<0 on failure
 *	@details		set expiration on allocated mempool
 */
int sveth_mpool_set_expiration(struct sveth_mpool_que_t* que_p, sveth_mpool_handle_t handle, pf_sveth_mpool_tm_handler tm_handler, unsigned long expires) {
	unsigned long flags;
	struct sveth_mpool_qn_t* node_p;
	struct sveth_mpool_qn_t* tmp_p;
	int empty;

	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->used.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	if(unlikely(empty)) {
		/* no used */
		PDEBUG("NG: used queue is empty. (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);
		return -ENOMEM;
	}

	write_lock_irqsave(&que_p->lock, flags);
	list_for_each_entry_safe(node_p, tmp_p, &que_p->used.head, this) {
		if(likely(node_p != 0) && unlikely(&node_p->data_p == (void**)handle)) {
			/* set timer */
			if(tm_handler && expires) {
				node_p->tm_handler	= tm_handler;
				
				init_timer(&node_p->timer);
				node_p->timer.expires	= jiffies + expires;
				node_p->timer.data		= (unsigned long)node_p;
				node_p->timer.function	= _timer_handler;
				
				add_timer(&node_p->timer);
			}
			/* success */
			goto _done;
		}
	}
	/* not found */
	write_unlock_irqrestore(&que_p->lock, flags);
	PDEBUG("NG: not found. (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);
	return -ENOMEM;

_done:
	write_unlock_irqrestore(&que_p->lock, flags);
	// PDEBUG("OK: (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);

	return 0;
}



/**
 *	@brief		sveth_mpool_is_valid
 *	
 *	@param[in,out]	*que_p		mempool queue
 *	@param[in]		handle		handle of mempool
 *	
 *	@return		int				0 on valid handle
 *								<0 on invalid handle
 *	@details		check handle
 */
int sveth_mpool_is_valid(struct sveth_mpool_que_t* que_p, sveth_mpool_handle_t handle) {
	unsigned long flags;
	struct sveth_mpool_qn_t* node_p;
	struct sveth_mpool_qn_t* tmp_p;
	int empty;

	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->used.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	if(unlikely(empty)) {
		/* no used */
		PDEBUG("NG: used queue is empty. (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);
		return -ENOMEM;
	}

	write_lock_irqsave(&que_p->lock, flags);
	list_for_each_entry_safe(node_p, tmp_p, &que_p->used.head, this) {
		if(likely(node_p != 0) && unlikely(&node_p->data_p == (void**)handle)) {
			/* valid handle */
			goto _valid;
		}
	}
	/* not found: invalid handle */
	write_unlock_irqrestore(&que_p->lock, flags);
	PDEBUG("NG: not found. (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);
	return -ENOMEM;

_valid:
	write_unlock_irqrestore(&que_p->lock, flags);
	// PDEBUG("OK: (que_p=0x%p, handle=0x%p)\n", (void*)que_p, (void*)handle);

	return 0;
}


/* check empty */
int sveth_mpool_is_empty(struct sveth_mpool_que_t* que_p) {
	unsigned long flags;
	int empty;
	
	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->free.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	
	return empty;
}



