/*
 *  sveth_que.c
 *
 *	The virtual ethernet driver for the SPiCa movie system.
 *
 *  Copyright (C) 2012 Panasonic Corporation. All Rights Reserved.
 *
 */

#if defined(CONFIG_SVETH_DEBUG)
static const char* devname_p = "sveth_que";
#endif	/* CONFIG_SVETH_DEBUG */

#include "sveth_config.h"
#include "sveth_debug.h"
#include "sveth_que.h"


/**
 *	prototype for local
 */

static void _timer_handler(unsigned long data);


/**
 *	inline
 */

/* init member */
static inline void _sveth_que_init_private(struct sveth_que_t* que_p) {
	/* init private */
	que_p->node_p		= 0;

	que_p->pool.num		= 0;
	que_p->req.num		= 0;
	
	INIT_LIST_HEAD(&que_p->pool.head);
	INIT_LIST_HEAD(&que_p->req.head);
}

/* set to qh_p */
static inline int _sveth_que_set_data(struct sveth_que_t* que_p, const void* data_p, pf_sveth_que_tm_handler tm_handler, unsigned long expires, struct sveth_que_qh_t* qh_p) {
	unsigned long flags;
	struct sveth_que_qn_t* node_p;
	int empty;
	
	/* check available */
	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->pool.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	if(unlikely(empty)) {
		/* no pool */
		return -ENOMEM;
	}

	/* get empty from pool */
	read_lock_irqsave(&que_p->lock, flags);
	node_p = list_first_entry(&que_p->pool.head, struct sveth_que_qn_t, this);
	read_unlock_irqrestore(&que_p->lock, flags);
	sveth_assert(node_p != 0);

	/* set user data */
	write_lock_irqsave(&que_p->lock, flags);
	/* set timer */
	if(tm_handler && expires) {
		node_p->tm_handler	= tm_handler;
		
		init_timer(&node_p->timer);
		node_p->timer.expires	= jiffies + expires;
		node_p->timer.data		= (unsigned long)node_p;
		node_p->timer.function	= _timer_handler;
		
		add_timer(&node_p->timer);
	}

	/* set param */
	node_p->qh_p	= qh_p;
	node_p->data_p	= (void*)data_p;
	write_unlock_irqrestore(&que_p->lock, flags);
	
	/* move to qh_p */
	write_lock_irqsave(&que_p->lock, flags);
	list_move_tail(&node_p->this, &qh_p->head);
	que_p->pool.num--;
	qh_p->num++;
	write_unlock_irqrestore(&que_p->lock, flags);

	/* success */
	return 0;
}


/* get from qh_p */
static inline int _sveth_que_get_data(struct sveth_que_t* que_p, void** data_pp, struct sveth_que_qh_t* qh_p) {
	unsigned long flags;
	struct sveth_que_qn_t* node_p;
	int empty;

	/* check available */
	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&qh_p->head));
	read_unlock_irqrestore(&que_p->lock, flags);
	if(unlikely(empty)) {
		/* no pool */
		return -ENOMEM;
	}

	/* get qh_p from req queue */
	read_lock_irqsave(&que_p->lock, flags);
	node_p = list_first_entry(&qh_p->head, struct sveth_que_qn_t, this);
	read_unlock_irqrestore(&que_p->lock, flags);
	sveth_assert(node_p != 0);

	/* get user data */
	read_lock_irqsave(&que_p->lock, flags);
	/* check expiration */
	if(node_p->tm_handler) {
		/* cancel timer */
		del_timer_sync(&node_p->timer);
		node_p->tm_handler	= 0;
	}

	*data_pp = node_p->data_p;
	read_unlock_irqrestore(&que_p->lock, flags);
	
	/* move to pool */
	write_lock_irqsave(&que_p->lock, flags);
	list_move_tail(&node_p->this, &que_p->pool.head);
	que_p->pool.num++;
	qh_p->num--;
	write_unlock_irqrestore(&que_p->lock, flags);

	return 0;
}


/**
 *	local
 */

static void _timer_handler(unsigned long data) {
	unsigned long flags;
	struct sveth_que_qn_t* node_p = (struct sveth_que_qn_t*)data;
	struct sveth_que_t* que_p = node_p->que_p;
	pf_sveth_que_tm_handler tm_handler;
	void* data_p;
	
	/* store param & clear param */
	write_lock_irqsave(&que_p->lock, flags);
	tm_handler	= node_p->tm_handler;
	data_p		= node_p->data_p;

	node_p->tm_handler	= 0;
	write_unlock_irqrestore(&que_p->lock, flags);
	
	/* free this to pool queue */
	write_lock_irqsave(&que_p->lock, flags);
	list_move_tail(&node_p->this, &que_p->pool.head);
	que_p->pool.num++;
	node_p->qh_p->num--;
	write_unlock_irqrestore(&que_p->lock, flags);

	/* call user-handler */
	tm_handler(data_p);
}



/* private info (for debugging) */
void sveth_que_dump_priv_info(struct sveth_que_t* que_p) {
	unsigned long flags;
	read_lock_irqsave(&que_p->lock, flags);
	PNOTICE("  que_p=0x%p\n", que_p);
	PNOTICE("    pool: num=%zd\n", que_p->pool.num);
	PNOTICE("    req:  num=%zd\n", que_p->req.num);
	read_unlock_irqrestore(&que_p->lock, flags);
}

/* init */
void sveth_que_init(struct sveth_que_t* que_p) {
	unsigned long flags;

	rwlock_init(&que_p->lock);

	/* init private */
	write_lock_irqsave(&que_p->lock, flags);
	_sveth_que_init_private(que_p);
	write_unlock_irqrestore(&que_p->lock, flags);
}


/* setup */
int sveth_que_setup(struct sveth_que_t* que_p, size_t num) {
	unsigned long flags;
	int ret;
	size_t i;
	struct sveth_que_qn_t* node_p;

	rwlock_init(&que_p->lock);

	/* init private */
	write_lock_irqsave(&que_p->lock, flags);
	_sveth_que_init_private(que_p);
	write_unlock_irqrestore(&que_p->lock, flags);

	/* allocate node */
	node_p = kmalloc(sizeof(*node_p) * num, GFP_KERNEL);
	if(node_p == 0) {
		/* no memory */
		ret = -ENOMEM;
		goto _err0;
	}

	/* init node_p and add to pool */
	write_lock_irqsave(&que_p->lock, flags);
	for(i = 0; i < num; i++) {

		/* init param */
		INIT_LIST_HEAD(&node_p[i].this);
		node_p[i].data_p		= 0;
		node_p[i].tm_handler	= 0;
		node_p[i].que_p			= que_p;
		node_p[i].qh_p			= 0;

		//PNOTICE("pool: alloc=%p, p->data_p=%p\n", &node_p[i], node_p[i].data_p);

		/* add to tail */
		list_add_tail(&node_p[i].this, &que_p->pool.head);
		que_p->pool.num++;
	}
	write_unlock_irqrestore(&que_p->lock, flags);

	/* success */
	write_lock_irqsave(&que_p->lock, flags);
	que_p->node_p = node_p;
	write_unlock_irqrestore(&que_p->lock, flags);
	
	return 0;

_err0:	/* err0: fail */
	return ret;
}


/* cleanup */
static inline void _cleanup(struct sveth_que_qh_t* qh_p, pf_sveth_que_cl_handler cl_handler) {
	struct sveth_que_qn_t* node_p;
	struct sveth_que_qn_t* tmp_p;

	list_for_each_entry_safe(node_p, tmp_p, &qh_p->head, this) {
		if(node_p != 0) {
			//PNOTICE("free=%p", node_p);
			/* check expiration */
			if(node_p->tm_handler) {
				/* cancel timer */
				del_timer_sync(&node_p->timer);
				node_p->tm_handler	= 0;
			}
			
			/* execute cleanup handler */
			if(cl_handler != 0) {
				cl_handler(node_p->data_p);
			}
			
			/* clear param */
			node_p->que_p	= 0;
			node_p->qh_p	= 0;
			node_p->data_p	= 0;

			list_del(&node_p->this);
			qh_p->num--;
		}
	}
}

void sveth_que_cleanup(struct sveth_que_t* que_p, pf_sveth_que_cl_handler cl_handler) {
	unsigned long flags;
	struct sveth_que_qn_t* node_p;
	
	/* free: req */
	write_lock_irqsave(&que_p->lock, flags);
	_cleanup(&que_p->req, cl_handler);
	write_unlock_irqrestore(&que_p->lock, flags);

	/* free: pool */
	write_lock_irqsave(&que_p->lock, flags);
	_cleanup(&que_p->pool, 0);
	write_unlock_irqrestore(&que_p->lock, flags);

	/* free: node_p */
	write_lock_irqsave(&que_p->lock, flags);
	node_p = que_p->node_p;
	write_unlock_irqrestore(&que_p->lock, flags);
	kfree(node_p);

	/* init private */
	write_lock_irqsave(&que_p->lock, flags);
	_sveth_que_init_private(que_p);
	write_unlock_irqrestore(&que_p->lock, flags);
}


/* check empty */
int sveth_que_is_empty(struct sveth_que_t* que_p) {
	unsigned long flags;
	int empty;
	
	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->req.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	
	return empty;
}


/* check full */
int sveth_que_is_full(struct sveth_que_t* que_p) {
	unsigned long flags;
	int empty;
	
	read_lock_irqsave(&que_p->lock, flags);
	empty = unlikely(!! list_empty_careful(&que_p->pool.head));
	read_unlock_irqrestore(&que_p->lock, flags);
	
	return empty;
}


/* get enqueue num */
size_t sveth_que_get_enqueue_num(struct sveth_que_t* que_p) {
	unsigned long flags;
	size_t num;
	
	read_lock_irqsave(&que_p->lock, flags);
	num = que_p->req.num;
	read_unlock_irqrestore(&que_p->lock, flags);
	
	return num;
}


/* enqueue */
int sveth_que_enqueue(struct sveth_que_t* que_p, const void* data_p, pf_sveth_que_tm_handler tm_handler, unsigned long expires) {
	/* set to req queue */
	return _sveth_que_set_data(que_p, data_p, tm_handler, expires, &que_p->req);
}


/* dequeue */
int sveth_que_dequeue(struct sveth_que_t* que_p, void** data_pp) {
	/* get from req queue */
	return _sveth_que_get_data(que_p, data_pp, &que_p->req);
}


