/* MN2PS On-chip serial port UART driver
 *
 * Copyright (C) 2012 Panasonic Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version
 * 2 of the Licence, or (at your option) any later version.
 */

static const char serial_name[] = "MN2PS Serial driver";
static const char serial_version[] = "serial-mn2ps-1.0";
static const char serial_revdate[] = "2012-07-31";

#include <linux/tty.h>
#include <linux/console.h>
#include <linux/serial_core.h>

#include <unit/serial_mn2ps.h>

#define no_printk(FMT, ...) do {} while(0)

#define kenter(FMT, ...) \
	printk(KERN_DEBUG "-->%s(" FMT ")\n", __func__, ##__VA_ARGS__)
#define _enter(FMT, ...) \
	no_printk(KERN_DEBUG "-->%s(" FMT ")\n", __func__, ##__VA_ARGS__)
#define kdebug(FMT, ...) \
	printk(KERN_DEBUG "--- " FMT "\n", ##__VA_ARGS__)
#define _debug(FMT, ...) \
	no_printk(KERN_DEBUG "--- " FMT "\n", ##__VA_ARGS__)
#define kproto(FMT, ...) \
	printk(KERN_DEBUG "### MNSERIAL " FMT " ###\n", ##__VA_ARGS__)
#define _proto(FMT, ...) \
	no_printk(KERN_DEBUG "### MNSERIAL " FMT " ###\n", ##__VA_ARGS__)

#define tx_can_transfer_and_empty(status)	\
	(((status) & (SC01STR_TBSY | SC01STR_TBF)) == \
	 (SC01STR_TBSY_IDLE | SC01STR_TBF_EMPTY))

static void serial_mn2ps_console_write(struct console *co,
					   const char *s, unsigned count);
static int __init serial_mn2ps_console_setup(struct console *co,
						 char *options);

static struct uart_driver serial_mn2ps_driver;
static struct console serial_mn2ps_console = {
	.name		= "ttySM",
	.write		= serial_mn2ps_console_write,
	.device		= uart_console_device,
	.setup		= serial_mn2ps_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &serial_mn2ps_driver,
};

static struct uart_driver serial_mn2ps_driver = {
	.owner		= NULL,
	.driver_name	= "mn2ps-serial",
	.dev_name	= "ttySM",
	.major		= TTY_MAJOR,
	.minor		= 128,
	.nr		= NR_PORTS,
	.cons		= &serial_mn2ps_console,
};

/*
 * the first on-chip serial port: ttySM0
 */
static struct serial_mn2ps_port serial_mn2ps_port_sif0 = {
	.name		= "ttySM0",
	.rx_name	= "ttySM0/Rx",
	.tx_name	= "ttySM0/Tx",
	.type           = PORT_MN2PS,

	.uart.ops	= &serial_mn2ps_direct_ops,
	.uart.membase	= (void __iomem *) 0,
	.uart.mapbase	= (unsigned long) SC0_BASE,
	.uart.iotype	= UPIO_MEM,
	.uart.irq	= SC0RXIRQ,
	.uart.uartclk	= 0,
	.uart.fifosize	= 1,
	.uart.flags	= UPF_BOOT_AUTOCONF,
	.uart.line	= 0,
	.uart.type	= PORT_MN2PS,
	.uart.lock	=
	__SPIN_LOCK_UNLOCKED(serial_mn2ps_port_sif0.uart.lock),

	.exclksel_mapbase   = SC0_EXCLK_SEL_BASE,
	.delay_mapbase      = SC_TX_DELAY1_BASE,
	.exclksel_membase   = (void __iomem *)0,
	.delay_membase      = (void __iomem *)0,

	.tx_xchar	= 0,
	.rx_irq		= SC0RXIRQ,
	.tx_irq		= SC0TXIRQ,

        .ioclk		= MN_PERI_CLK,
};

/*
 * the second on-chip serial port: ttySM1
 */
static struct serial_mn2ps_port serial_mn2ps_port_sif1 = {
	.name		= "ttySM1",
	.rx_name	= "ttySM1/Rx",
	.tx_name	= "ttySM1/Tx",
	.type           = PORT_MN2PS,

	.uart.ops	= &serial_mn2ps_direct_ops,
	.uart.membase	= (void __iomem *) 0,
	.uart.mapbase	= (unsigned long) SC1_BASE,
	.uart.iotype	= UPIO_MEM,
	.uart.irq	= SC1RXIRQ,
	.uart.uartclk	= 0,
	.uart.fifosize	= 1,
	.uart.flags	= UPF_BOOT_AUTOCONF,
	.uart.line	= 1,
	.uart.type	= PORT_MN2PS,
	.uart.lock	=
	__SPIN_LOCK_UNLOCKED(serial_mn2ps_port_sif1.uart.lock),

	.exclksel_mapbase   = SC1_EXCLK_SEL_BASE,
	.delay_mapbase      = SC_TX_DELAY1_BASE,
	.exclksel_membase   = (void __iomem *)0,
	.delay_membase      = (void __iomem *)0,

	.tx_xchar	= 0,
	.rx_irq		= SC1RXIRQ,
	.tx_irq		= SC1TXIRQ,

        .ioclk		= MN_PERI_CLK,
};

/*
 * the third on-chip serial port: ttySM2
 */
static struct serial_mn2ps_port serial_mn2ps_port_sif2 = {
	.name		= "ttySM2",
	.rx_name	= "ttySM2/Rx",
	.tx_name	= "ttySM2/Tx",
	.type           = PORT_MN2PS_FIFO,

	.uart.ops	= &serial_mn2ps_fifo_ops,
	.uart.membase	= (void __iomem *) 0,
	.uart.mapbase	= (unsigned long) SC2_BASE,
	.uart.iotype	= UPIO_MEM,
	.uart.irq	= SC2RXIRQ,
	.uart.uartclk	= 0,
	.uart.fifosize	= SERIAL_FIFO_SIZE,
	.uart.flags	= UPF_BOOT_AUTOCONF,
	.uart.line	= 2,
	.uart.type	= PORT_MN2PS_FIFO,
	.uart.lock	=
	__SPIN_LOCK_UNLOCKED(serial_mn2ps_port_sif2.uart.lock),

	.tx_xchar	= 0,
	.rx_irq		= SC2RXIRQ,
	.tx_irq		= SC2TXIRQ,

	.ioclk		= MN_PERI_CLK,
};

/*
 * the forth on-chip serial port: ttySM3
 */
static struct serial_mn2ps_port serial_mn2ps_port_sif3 = {
	.name		= "ttySM3",
	.rx_name	= "ttySM3/Rx",
	.tx_name	= "ttySM3/Tx",
	.type           = PORT_MN2PS_FIFO,

	.uart.ops	= &serial_mn2ps_fifo_ops,
	.uart.membase	= (void __iomem *) 0,
	.uart.mapbase	= (unsigned long) SC3_BASE,
	.uart.iotype	= UPIO_MEM,
	.uart.irq	= SC3RXIRQ,
	.uart.uartclk	= 0,
	.uart.fifosize	= SERIAL_FIFO_SIZE,
	.uart.flags	= UPF_BOOT_AUTOCONF,
	.uart.line	= 3,
	.uart.type	= PORT_MN2PS_FIFO,
	.uart.lock	=
	__SPIN_LOCK_UNLOCKED(serial_mn2ps_port_sif3.uart.lock),

	.tx_xchar	= 0,
	.rx_irq		= SC3RXIRQ,
	.tx_irq		= SC3TXIRQ,

        .ioclk		= MN_PERI_CLK,
};

/*
 * list of available serial ports
 */
static struct serial_mn2ps_port *serial_mn2ps_ports[NR_PORTS + 1] = {
	[0]	= &serial_mn2ps_port_sif0,
	[1]	= &serial_mn2ps_port_sif1,
	[2]	= &serial_mn2ps_port_sif2,
	[3]	= &serial_mn2ps_port_sif3,
	[NR_PORTS] = NULL,
};

/*
 * offset to the base address of serial N register
 */
struct sc_reg_offset {
	unsigned long ofs_base;
	unsigned long ofs_exclksel;
	unsigned long ofs_delay;
};

static const struct sc_reg_offset sc_base_ofs[NR_PORTS] = {
	{SC0_BASE_OFFSET, SC0_EXCLK_SEL_BASE_OFFSET, SC_TX_DELAY1_BASE_OFFSET},
	{SC1_BASE_OFFSET, SC1_EXCLK_SEL_BASE_OFFSET, SC_TX_DELAY1_BASE_OFFSET},
	{SC2_BASE_OFFSET, 0, 0},
	{SC3_BASE_OFFSET, 0, 0},
};

/********************************************************************
 * DESCRIPTION	:
 *		initialise the MN2PS on-chip UARTs
 *
 * INPUT	:
 *		void
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int
 *
 *******************************************************************/
static int __init serial_mn2ps_init(void)
{
	struct serial_mn2ps_port *port;
	int ret, i;

	printk(KERN_INFO "%s version %s (%s)\n",
	       serial_name, serial_version, serial_revdate);

	ret = uart_register_driver(&serial_mn2ps_driver);
	if (!ret) {
		for (i = 0 ; i < NR_PORTS ; i++) {
			port = serial_mn2ps_ports[i];
			if (!port) {
				continue;
			}

			ret = uart_add_one_port(&serial_mn2ps_driver,
						&port->uart);

			if (ret < 0) {
				_debug("ERROR %d", -ret);
				break;
			}
		}

		if (ret) {
			uart_unregister_driver(&serial_mn2ps_driver);
		}
	}

	return ret;
}

__initcall(serial_mn2ps_init);

/********************************************************************
 * DESCRIPTION	:
 *		print a string to the serial port without disturbing the real user of the
 *		port too much
 *		- the console must be locked by the caller
 *
 * INPUT	:
 *		struct console *co	console
 *		const char *s		pointer of output buffer
 *		unsigned count		output count
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_console_write(struct console *co,
                                       const char *s, unsigned count)
{

	struct serial_mn2ps_port *port;

	port = serial_mn2ps_ports[co->index];

	if (port->type == PORT_MN2PS) {
		serial_mn2ps_direct_console_write(port, s, count);
	} else if (port->type == PORT_MN2PS_FIFO) {
		serial_mn2ps_fifo_console_write(port, s, count);
	} else {
		return;
	}
}

/********************************************************************
 * DESCRIPTION	:
 * 		set up a serial port as a console
 * 		- construct a cflag setting for the first rs_open()
 * 		- initialize the serial port
 * 		- return non-zero if we didn't find a serial port.
 *
 * INPUT	:
 *		struct console *co	console
 *		char *options		configuration information of a console
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int			    0: normal
 *					other: error
 *
 *******************************************************************/
static int __init serial_mn2ps_console_setup(struct console *co,
						 char *options)
{
	struct serial_mn2ps_port *port;
	int parity = 'n', baud = 9600, bits = 8, flow = 0;
	int ret;

	if ((co->index < 0) || (co->index > (NR_PORTS - 1))) {
		return -ENODEV;
	}

	port = serial_mn2ps_ports[co->index];
	if (!port->uart.membase) {
		/* need to ioremap  */
		ret = serial_mn2ps_ioremap(port);
		if (ret) {
			return ret;
		}
	}

	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);

	return uart_set_options(&port->uart, co, baud, parity, bits, flow);
}

/********************************************************************
 * DESCRIPTION	:
 * 		register console
 *
 * INPUT	:
 *		void
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int
 *
 *******************************************************************/
static int __init serial_mn2ps_console_init(void)
{
	register_console(&serial_mn2ps_console);
	return 0;
}

console_initcall(serial_mn2ps_console_init);

/********************************************************************
 * DESCRIPTION	:
 *		serial resources are remaped.
 *
 * INPUT	:
 *		struct serial_mn2ps_port* port	serial port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int	error code
 *			0	: normal
 *			-ENOMEM	: error
 *
 *******************************************************************/
int serial_mn2ps_ioremap(struct serial_mn2ps_port* port) {

	int ret = 0, i;
	void *map_adr;

	map_adr = ioremap_nocache(SC_MAP_BASE, SC_MAP_SIZE);
	if (!map_adr) {
		ret = -ENOMEM;
	}

	for (i = 0; i < NR_PORTS; i++) {
		serial_mn2ps_ports[i]->uart.membase     = (void*)((unsigned int)map_adr + sc_base_ofs[i].ofs_base);
		serial_mn2ps_ports[i]->exclksel_membase = (void*)((unsigned int)map_adr + sc_base_ofs[i].ofs_exclksel);
		serial_mn2ps_ports[i]->delay_membase    = (void*)((unsigned int)map_adr + sc_base_ofs[i].ofs_delay);
	}

	return ret;
}
