/* MN2PS On-chip serial port UART driver
 *
 * Copyright (C) 2012 Panasonic Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version
 * 2 of the Licence, or (at your option) any later version.
 */

#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/serial_core.h>
#include <linux/serial_mn2ps.h>

#include <asm/uaccess.h>
#include <unit/serial_mn2ps.h>

#include <mach/intctl-regs.h>

#define no_printk(FMT, ...) do {} while(0)

#define kenter(FMT, ...) \
	printk(KERN_DEBUG "-->%s(" FMT ")\n", __func__, ##__VA_ARGS__)
#define _enter(FMT, ...) \
	no_printk(KERN_DEBUG "-->%s(" FMT ")\n", __func__, ##__VA_ARGS__)
#define kdebug(FMT, ...) \
	printk(KERN_DEBUG "--- " FMT "\n", ##__VA_ARGS__)
#define _debug(FMT, ...) \
	no_printk(KERN_DEBUG "--- " FMT "\n", ##__VA_ARGS__)
#define kproto(FMT, ...) \
	printk(KERN_DEBUG "### MNSERIAL " FMT " ###\n", ##__VA_ARGS__)
#define _proto(FMT, ...) \
	no_printk(KERN_DEBUG "### MNSERIAL " FMT " ###\n", ##__VA_ARGS__)

#define tx_can_transfer_and_empty(status)	\
	((status & SC01STR_TBF) == SC01STR_TBF_EMPTY)

static unsigned int serial_mn2ps_direct_tx_empty(struct uart_port *);
static void serial_mn2ps_direct_set_mctrl(struct uart_port *, unsigned int);
static unsigned int serial_mn2ps_direct_get_mctrl(struct uart_port *);
static void serial_mn2ps_direct_stop_tx(struct uart_port *);
static void serial_mn2ps_direct_start_tx(struct uart_port *);
static void serial_mn2ps_direct_stop_rx(struct uart_port *);
static void serial_mn2ps_direct_enable_ms(struct uart_port *);
static void serial_mn2ps_direct_break_ctl(struct uart_port *, int);
static int serial_mn2ps_direct_startup(struct uart_port *);
static void serial_mn2ps_direct_shutdown(struct uart_port *);
static void serial_mn2ps_direct_set_termios(struct uart_port *,
					 struct ktermios *,
					 struct ktermios *);
static const char *serial_mn2ps_direct_type(struct uart_port *);
static void serial_mn2ps_direct_release_port(struct uart_port *);
static int serial_mn2ps_direct_request_port(struct uart_port *);
static void serial_mn2ps_direct_config_port(struct uart_port *, int);
static int serial_mn2ps_direct_verify_port(struct uart_port *,
					struct serial_struct *);
static int serial_mn2ps_direct_ioctl(struct uart_port *, unsigned int, unsigned long);

const struct uart_ops serial_mn2ps_direct_ops = {
	.tx_empty	= serial_mn2ps_direct_tx_empty,
	.set_mctrl	= serial_mn2ps_direct_set_mctrl,
	.get_mctrl	= serial_mn2ps_direct_get_mctrl,
	.stop_tx	= serial_mn2ps_direct_stop_tx,
	.start_tx	= serial_mn2ps_direct_start_tx,
	.stop_rx	= serial_mn2ps_direct_stop_rx,
	.enable_ms	= serial_mn2ps_direct_enable_ms,
	.break_ctl	= serial_mn2ps_direct_break_ctl,
	.startup	= serial_mn2ps_direct_startup,
	.shutdown	= serial_mn2ps_direct_shutdown,
	.set_termios	= serial_mn2ps_direct_set_termios,
	.type		= serial_mn2ps_direct_type,
	.release_port	= serial_mn2ps_direct_release_port,
	.request_port	= serial_mn2ps_direct_request_port,
	.config_port	= serial_mn2ps_direct_config_port,
	.verify_port	= serial_mn2ps_direct_verify_port,
	.ioctl		= serial_mn2ps_direct_ioctl,
};

static irqreturn_t serial_mn2ps_direct_interrupt(int irq, void *dev_id);
static void serial_mn2ps_direct_receive_interrupt(struct serial_mn2ps_port *);
static void serial_mn2ps_direct_transmit_interrupt(struct serial_mn2ps_port *);
static int serial_mn2ps_direct_wait_for_enable_transfer(struct serial_mn2ps_port *);
static int serial_mn2ps_direct_wait_for_tx_complete(struct serial_mn2ps_port *);
static void serial_mn2ps_direct_console_putchar(struct uart_port *, int);

/********************************************************************
 * DESCRIPTION	:
 *		receive chars from the ring buffer for this serial port
 *		- must do break detection here (not done in the UART)
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port		serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_receive_interrupt(struct serial_mn2ps_port *port)
{
	struct uart_icount *icount = &port->uart.icount;
	struct tty_struct *tty = port->uart.state->port.tty;
	u32 status;
	int max_count = 256;
	unsigned char ch;
	char flag;

	status = __raw_readl(port->uart.membase + SC0STR);
	while (((status & SC01STR_RXA) == SC01STR_RXA_AVAIL) &&
               (max_count-- > 0)) {
		ch = (unsigned char)__raw_readl(port->uart.membase + SC0RXB);
		flag = TTY_NORMAL;
		icount->rx++;

		if (unlikely(status & SC01STR_ERROR_BITS)) {
			if (status & SC01STR_FEF)
				icount->frame++;
			if (status & SC01STR_PEF)
				icount->parity++;
			if (status & SC01STR_OEF)
				icount->overrun++;

			if (status & SC01STR_PEF)
				flag = TTY_PARITY;
			else if (status & SC01STR_FEF)
				flag = TTY_FRAME;
		}

		if (uart_handle_sysrq_char(&port->uart, ch))
			goto ignore_char;

		uart_insert_char(&port->uart, status, SC01STR_OEF, ch, flag);

ignore_char:
		status = __raw_readl(port->uart.membase + SC0STR);
	}
	spin_unlock(&port->uart.lock);
	tty_flip_buffer_push(tty);
	spin_lock(&port->uart.lock);
}

/********************************************************************
 * DESCRIPTION	:
 *		Transmission interruption is processed.
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port		serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_transmit_interrupt(struct serial_mn2ps_port *port)
{
	struct circ_buf *xmit = &port->uart.state->xmit;
	u32 control, status;

	if (port->tx_xchar) {
		__raw_writel(port->tx_xchar, port->uart.membase + SC0TXB);
		port->uart.icount.tx++;
		port->tx_xchar = 0;
		return;
	}

	if (uart_circ_empty(xmit)) {
		serial_mn2ps_direct_stop_tx(&port->uart);
		return;
	}

	control = __raw_readl(port->uart.membase + SC0CTR);
	status  = __raw_readl(port->uart.membase + SC0STR);
	if (((control & SC01CTR_TXE) != SC01CTR_TXE) ||
	    ((status & SC01STR_TBF) == SC01STR_TBF_FULL)) {
		serial_mn2ps_direct_stop_tx(&port->uart);
		return;
	}

	__raw_writel(xmit->buf[xmit->tail], port->uart.membase + SC0TXB);
	xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
	port->uart.icount.tx++;

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS) {
		uart_write_wakeup(&port->uart);
	}
}

/********************************************************************
 * DESCRIPTION	:
 *		The handler of interruption
 *
 * INPUT	:
 *		int irq		interrupt number
 *              void *dev_id	device id
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		irqreturn_t
 *
 *******************************************************************/
static irqreturn_t serial_mn2ps_direct_interrupt(int irq, void *dev_id)
{
	struct serial_mn2ps_port *port = dev_id;
	unsigned long flags;
	u32 status;

	status = __raw_readl(port->uart.membase + SC0STR);
	spin_lock_irqsave(&port->uart.lock, flags);
	if ((status & SC01STR_RXA) == SC01STR_RXA_AVAIL)
		serial_mn2ps_direct_receive_interrupt(port);

	if (tx_can_transfer_and_empty(status))
		serial_mn2ps_direct_transmit_interrupt(port);

	spin_unlock_irqrestore(&port->uart.lock, flags);

	return IRQ_HANDLED;
}

/********************************************************************
 * DESCRIPTION	:
 *		return indication of whether the hardware transmit buffer is empty
 *
 * INPUT	:
 *		struct uart_port *uart	uart port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		unsigned int
 *              		0		not empty
 *              		TIOCSER_TEMT	transmit buffer is empty.
 *
 *******************************************************************/
static unsigned int serial_mn2ps_direct_tx_empty(struct uart_port *uart)
{
	unsigned int ret = 0;
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 status;

	_enter("%s", port->name);

	status = __raw_readl(uart->membase + SC0STR);
	if (tx_can_transfer_and_empty(status))
		ret = TIOCSER_TEMT;

	return ret;
}


/********************************************************************
 * DESCRIPTION	:
 *		set the modem control lines (we don't have any)
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *              unsigned int mctrl	modem control flag
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_set_mctrl(struct uart_port *uart,
				     unsigned int mctrl)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s,%x", port->name, mctrl);
}

/********************************************************************
 * DESCRIPTION	:
 *		get the modem control line status
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		unsigned int	modem control line status
 *
 *******************************************************************/
static unsigned int serial_mn2ps_direct_get_mctrl(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	return TIOCM_CAR | TIOCM_CTS | TIOCM_DSR;
}

/********************************************************************
 * DESCRIPTION	:
 *		stop transmitting characters
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_stop_tx(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s", port->name);

	control = __raw_readl(uart->membase + SC0CTR);
	control &= ~SC01CTR_TXE;
	__raw_writel(control, uart->membase + SC0CTR);
}

/********************************************************************
 * DESCRIPTION	:
 *		start transmitting characters
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_start_tx(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s", port->name);

	control = __raw_readl(uart->membase + SC0CTR);
	control |= (SC01CTR_TXE | SC01CTR_TI_EMPTY);
	__raw_writel(control, uart->membase + SC0CTR);

	mn_intc_set(port->tx_irq);
}

/********************************************************************
 * DESCRIPTION	:
 *		stop receiving characters
 *		- called whilst the port is being closed
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_stop_rx(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s,%02x", port->name, ch);

	control = __raw_readl(uart->membase + SC0CTR);
	control &= ~SC01CTR_RXE;
	__raw_writel(control, uart->membase + SC0CTR);
}

/********************************************************************
 * DESCRIPTION	:
 *		enable modem status interrupts
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_enable_ms(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);
}

/********************************************************************
 * DESCRIPTION	:
 *		transmit or cease transmitting a break signal
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		int ctl			break control flag
 *					  1: break signal on
 *					  0: break signal off
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_break_ctl(struct uart_port *uart, int ctl)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	unsigned long flags;
	u32 control;

	_enter("%s,%d", port->name, ctl);

	spin_lock_irqsave(&port->uart.lock, flags);
	control = __raw_readl(uart->membase + SC0CTR);
	if (ctl) {
		control |= SC01CTR_BKE;
	} else {
		control &= ~SC01CTR_BKE;
	}
	__raw_writel(control, uart->membase + SC0CTR);
	spin_unlock_irqrestore(&port->uart.lock, flags);
}

/********************************************************************
 * DESCRIPTION	:
 *		grab the interrupts and enable the port for reception
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static int serial_mn2ps_direct_startup(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s,%d", port->name, ctl);

	if (request_irq(port->rx_irq, serial_mn2ps_direct_interrupt,
			IRQF_DISABLED, port->rx_name, port) < 0)
		goto error1;
	if (request_irq(port->tx_irq, serial_mn2ps_direct_interrupt,
			IRQF_DISABLED, port->tx_name, port) < 0)
		goto error2;

	control = __raw_readl(uart->membase + SC0CTR);
	control |= (SC01CTR_TXE | SC01CTR_RXE | SC01CTR_TI_EMPTY);
	__raw_writel(control, uart->membase + SC0CTR);

	return 0;

error2:
	free_irq(port->rx_irq, port);
error1:
	return -EBUSY;
}

/********************************************************************
 * DESCRIPTION	:
 *		 shutdown the port and release interrupts
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_shutdown(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s,%d", port->name, ctl);

	control = __raw_readl(uart->membase + SC0CTR);
	control &= ~(SC01CTR_TXE | SC01CTR_RXE | SC01CTR_BKE);
	__raw_writel(control, uart->membase + SC0CTR);

	free_irq(port->rx_irq, port);
	free_irq(port->tx_irq, port);
}

/********************************************************************
 * DESCRIPTION	:
 *		set the terminal I/O parameters
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		struct ktermios *new	termios to be set up from now on
 *		struct ktermios *old	termios before a setup
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_set_termios(struct uart_port *uart,
					 struct ktermios *new,
					 struct ktermios *old)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	unsigned cflag;
	int baud, bits, xdiv;
	u32 cycle, scxctr, tmp, exclk_sel;
	unsigned long flags;

	/* byte size and parity */
	cflag = new->c_cflag;
	switch (cflag & CSIZE) {
	case CS7: scxctr = SC01CTR_CLN_7BIT; bits = 9;  break;
	case CS8: scxctr = SC01CTR_CLN_8BIT; bits = 10; break;
	default:  scxctr = SC01CTR_CLN_8BIT; bits = 10; break;
	}

	if (cflag & CSTOPB) {
		scxctr |= SC01CTR_STB_2BIT;
		bits++;
	}

	if (cflag & PARENB) {
		bits++;
		if (cflag & PARODD)
			scxctr |= SC01CTR_PB_ODD;
		else if (cflag & CMSPAR)
			scxctr |= SC01CTR_PB_FIXED0;
		else
			scxctr |= SC01CTR_PB_EVEN;
	}

	/* change Transfer bit-order (LSB/MSB) */
	if (cflag & CODMSB) {
		scxctr |= SC01CTR_OD_MSBFIRST; /* MSB MODE */
	} else {
		scxctr &= ~SC01CTR_OD_MSBFIRST; /* LSB MODE */
	}

	/* make the changes */
	spin_lock_irqsave(&port->uart.lock, flags);

	tmp = __raw_readl(uart->membase + SC0CTR);
	if ((tmp & SC01CTR_MD) == SC01CTR_MD_STST_SYNC) {
		/* change protocol start-stop transmission */
		scxctr |= (SC01CTR_MD_STST_SYNC | SC01CTR_CK_IOCLK_16);
		port->uart.uartclk = port->ioclk / 16;

		/* Determine divisor based on baud rate */
		baud = uart_get_baud_rate(&port->uart, new, old, 0,
					  port->uart.uartclk);
		_debug("ALT [baud %d]", baud);

		if (!baud)
			baud = 9600;	/* B0 transition handled in rs_set_termios */

		xdiv = 1;
		if (baud == 134) {
			baud = 269;	/* 134 is really 134.5 */
			xdiv = 2;
		}

		cycle = port->uart.uartclk / (2 * baud / xdiv) - 1;
		_debug("UARTCLK: %u / %hu", port->uart.uartclk, cycle);

		uart_update_timeout(&port->uart, new->c_cflag, baud);

		/* set prescaler */
		__raw_writel(cycle, port->uart.membase + SIOTOP_CYCLE0);
		__raw_writel(SC0_EXCLK_SEL_TIMER, port->exclksel_membase);

	} else { /* SC01CTR_MD_CLOCK_SYNC */
		scxctr |= (SC01CTR_MD_CLOCK_SYNC | (tmp & SC01CTR_CK));
		if ((tmp & SC01CTR_CK) == SC01CTR_CK_IOCLK_2) {
			port->uart.uartclk = port->ioclk / 2;
		} else if ((tmp & SC01CTR_CK) == SC01CTR_CK_IOCLK_16) {
			port->uart.uartclk = port->ioclk / 16;
		} else {
			port->uart.uartclk = port->ioclk;
		}
		/* Calculate baud rate */
		exclk_sel = __raw_readl(port->exclksel_membase);
		if (exclk_sel == SC0_EXCLK_SEL_TIMER) {
			cycle = __raw_readl(uart->membase + SIOTOP_CYCLE0);
			baud = port->uart.uartclk / (2 * (cycle + 1));
		} else if (exclk_sel == SC0_EXCLK_SEL_SYSCLK) {
			cycle = 0;
			baud = port->uart.uartclk;
		} else {
			goto error;
		}
		_debug("UARTCLK: [baud %d  clk:%u cycle:%hu]",
		       baud, port->uart.uartclk, cycle);

		uart_update_timeout(&port->uart, new->c_cflag, baud);
	}

	/* set up parity check flag */
	port->uart.read_status_mask = (1 << TTY_NORMAL) | (1 << TTY_OVERRUN);
	if (new->c_iflag & INPCK)
		port->uart.read_status_mask |=
			(1 << TTY_PARITY) | (1 << TTY_FRAME);
	if (new->c_iflag & (BRKINT | PARMRK))
		port->uart.read_status_mask |= (1 << TTY_BREAK);

	/* characters to ignore */
	port->uart.ignore_status_mask = 0;
	if (new->c_iflag & IGNPAR)
		port->uart.ignore_status_mask |=
			(1 << TTY_PARITY) | (1 << TTY_FRAME);
	if (new->c_iflag & IGNBRK) {
		port->uart.ignore_status_mask |= (1 << TTY_BREAK);
		/*
		 * If we're ignoring parity and break indicators,
		 * ignore overruns to (for real raw support).
		 */
		if (new->c_iflag & IGNPAR)
			port->uart.ignore_status_mask |= (1 << TTY_OVERRUN);
	}

	/* Ignore all characters if CREAD is not set */
	if ((new->c_cflag & CREAD) == 0)
		port->uart.ignore_status_mask |= (1 << TTY_NORMAL);

	tmp = __raw_readl(port->uart.membase + SC0CTR);
	scxctr |= tmp & (SC01CTR_TXE | SC01CTR_RXE | SC01CTR_BKE);
	__raw_writel(scxctr, port->uart.membase + SC0CTR);

error:
	spin_unlock_irqrestore(&port->uart.lock, flags);
}

/********************************************************************
 * DESCRIPTION	:
 *		return description of port type
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		const char*	port description
 *
 *******************************************************************/
static const char *serial_mn2ps_direct_type(struct uart_port *uart)
{
	return "MN2PS SIF";
}

/********************************************************************
 * DESCRIPTION	:
 *		request I/O and memory regions for port
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int	error code
 *			0	: normal
 *			-EBUSY	: error
 *
 *******************************************************************/
static int serial_mn2ps_direct_request_port(struct uart_port *uart)
{
	int ret = 0;
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	/* request_mem_reqion */
	if (!request_mem_region(port->uart.mapbase,
                                SC_MAP_SIZE, port->name)) {
		ret = -EBUSY;
		goto request_out;
	}

	/* ioremap */
	if (!port->uart.membase) {
		ret = serial_mn2ps_ioremap(port);
	}

request_out:
	return ret;
}


/********************************************************************
 * DESCRIPTION	:
 *		configure the type and reserve the ports
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		int type
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_config_port(struct uart_port *uart, int type)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	port->uart.type = PORT_MN2PS;

	serial_mn2ps_direct_request_port(uart);
}


/********************************************************************
 * DESCRIPTION	:
 *		release I/O and memory regions in use by port
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_release_port(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	iounmap(port->uart.membase);

	release_mem_region(port->uart.mapbase, SC_MAP_SIZE);
}


/********************************************************************
 * DESCRIPTION	:
 *		verify serial parameters are suitable for this port type
 *
 * INPUT	:
 *		struct uart_port *uart		uart port
 *		struct serial_struct *ss	serial information
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int	0      : normal
 *			EINVAL : error
 *
 *******************************************************************/
static int serial_mn2ps_direct_verify_port(struct uart_port *uart,
					struct serial_struct *ss)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	void *mapbase = (void *) (unsigned long) port->uart.mapbase;

	_enter("%s", port->name);

	/* these things may not be changed */
	if (ss->irq		!= port->uart.irq	||
	    ss->port		!= port->uart.iobase	||
	    ss->io_type		!= port->uart.iotype	||
	    ss->iomem_base	!= mapbase ||
	    ss->iomem_reg_shift	!= port->uart.regshift	||
	    ss->hub6		!= port->uart.hub6	||
	    ss->xmit_fifo_size	!= port->uart.fifosize)
		return -EINVAL;

	if (ss->type != port->uart.type)
		return -EINVAL;

	if (ss->type != PORT_MN2PS)
		return -EINVAL;

	return 0;
}

/********************************************************************
 * DESCRIPTION	:
 *		serial control of MN2PS original
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		unsigned int cmd	ioctl command
 *		unsigned long arg	comand argument
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int
 *
 *******************************************************************/
static int serial_mn2ps_direct_ioctl(struct uart_port *uart, unsigned int cmd,
				unsigned long a)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	struct mn2ps_serial_sync_speed speed;
	unsigned int arg;
	u32 tmp;

	switch (cmd) {
	case TIOSSGET:
		tmp = __raw_readl(uart->membase + SC0CTR);
		if ((tmp & SC01CTR_MD) == SC01CTR_MD_STST_SYNC) {
			return -EIO;
		}
		speed.clk_src = tmp & SC01CTR_CK;
		tmp = __raw_readl(port->exclksel_membase);
		speed.exclk_sel = tmp;
		tmp = __raw_readl(uart->membase + SIOTOP_CYCLE0);
		speed.prescaler_cycle = tmp;
		if (copy_to_user((struct mn2ps_serial_sync_speed *)a, &speed,
				 sizeof(struct mn2ps_serial_sync_speed))) {
			return -EFAULT;
		}
		break;
	case TIOSSSET:
		if (copy_from_user(&speed, (struct mn2ps_serial_sync_speed *)a,
				   sizeof(struct mn2ps_serial_sync_speed))) {
			return -EFAULT;
		}
		tmp = __raw_readl(uart->membase + SC0CTR);
		if ((tmp & SC01CTR_MD) == SC01CTR_MD_STST_SYNC) {
			return -EIO;
		}
		if ((speed.clk_src != CLK_SRC_SBT) &&
		    (speed.clk_src != CLK_SRC_2)   &&
		    (speed.clk_src != CLK_SRC_16)) {
			return -EINVAL;
		}
		if ((speed.exclk_sel != EXCLK_PERI) &&
		    (speed.exclk_sel != EXCLK_PRESCALER)) {
			return -EINVAL;
		}
		if (speed.prescaler_cycle > 0xffff) {
			return -EINVAL;
		}
		switch (speed.clk_src) {
		case CLK_SRC_SBT:
			tmp = (tmp & ~SC01CTR_CK);
			break;
		case CLK_SRC_2:
			tmp = (tmp & ~SC01CTR_CK) | SC01CTR_CK_IOCLK_2;
			break;
		case CLK_SRC_16:
			tmp = (tmp & ~SC01CTR_CK) | SC01CTR_CK_IOCLK_16;
			break;
		default:
			return -EINVAL;
		}
		__raw_writel(tmp, uart->membase + SC0CTR);

		switch(speed.exclk_sel) {
		case EXCLK_PERI:
			tmp = SC0_EXCLK_SEL_SYSCLK;
			break;
		case EXCLK_PRESCALER:
			tmp = SC0_EXCLK_SEL_TIMER;
			break;
		default:
			return -EINVAL;
		}
		__raw_writel(tmp, port->exclksel_membase);

		__raw_writel(speed.prescaler_cycle, uart->membase + SIOTOP_CYCLE0);

		break;
	case TIOPGET:
		tmp = __raw_readl(uart->membase + SC0CTR);
		switch (tmp & SC01CTR_MD) {
		case SC01CTR_MD_STST_SYNC:
			arg = PROTOCOL_ASYNC;
			break;
		case SC01CTR_MD_CLOCK_SYNC:
			arg = PROTOCOL_SYNC;
			break;
		default:
			return -EIO;
		}
		if (put_user(arg, (unsigned int __user *)a)) {
			return -EFAULT;
		}
		break;
	case TIOPSET:
		arg = a;
		switch (arg) {
		case PROTOCOL_ASYNC:
			tmp = __raw_readl(uart->membase + SC0CTR);
			tmp = (tmp & ~SC01CTR_MD) | SC01CTR_MD_STST_SYNC;
			__raw_writel(tmp, uart->membase + SC0CTR);
			break;
		case PROTOCOL_SYNC:
			tmp = __raw_readl(uart->membase + SC0CTR);
			tmp = (tmp & ~SC01CTR_MD) | SC01CTR_MD_CLOCK_SYNC;
			__raw_writel(tmp, uart->membase + SC0CTR);
			break;
		default:
			return -EINVAL;
		}
		break;
	case TIODLGET:
		tmp = __raw_readl(port->delay_membase);
		arg = (tmp >> (uart->line * 4)) & 0xf;
		if (put_user(arg, (unsigned int __user *)a)) {
			return -EFAULT;
		}
		break;
	case TIODLSET:
		arg = a;
		if (arg < 0 || arg > 0xf) {
			return -EINVAL;
		}
		tmp = __raw_readl(port->delay_membase);
		tmp &= ~(0xf << (uart->line * 4));
		tmp |= (arg << (uart->line * 4));
		__raw_writel(tmp, port->delay_membase);
		break;
	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}

/********************************************************************
 * DESCRIPTION	:
 *		It waits until transmission becomes possible.
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int		0        : normal
 *				EREMOTEIO: time out error
 *
 *******************************************************************/
static int serial_mn2ps_direct_wait_for_enable_transfer(struct serial_mn2ps_port *port)
{
	int ret = 0;
	unsigned int tmout = 10000;
	u32 status;

	status = __raw_readl(port->uart.membase + SC0STR);
	while ( !tx_can_transfer_and_empty(status) ){
		if (--tmout == 0) {
			ret = -EREMOTEIO;
			break;
		}
		udelay(1);
		status = __raw_readl(port->uart.membase + SC0STR);
	}
	return ret;
}

/********************************************************************
 * DESCRIPTION	:
 *		It waits until transmission complete.
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int		0        : normal
 *				EREMOTEIO: time out error
 *
 *******************************************************************/
static int serial_mn2ps_direct_wait_for_tx_complete(struct serial_mn2ps_port *port)
{
	return serial_mn2ps_direct_wait_for_enable_transfer(port);
}

/********************************************************************
 * DESCRIPTION	:
 *		One character is outputted in console processing.
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		int ch			output character
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_direct_console_putchar(struct uart_port *uart, int ch)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	if (!serial_mn2ps_direct_wait_for_enable_transfer(port)) {
		__raw_writel(ch, uart->membase + SC0TXB);
	}
}

/********************************************************************
 * DESCRIPTION	:
 * 		print a string to the serial port without disturbing the real user of the
 * 		port too much
 * 		- the console must be locked by the caller
 *
 * INPUT	:
 *		struct console *co	console
 *		const char *s		pointer of output buffer
 *		unsigned count		output count
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
void serial_mn2ps_direct_console_write(struct serial_mn2ps_port *port,
                                       const char *s, unsigned count)
{
	unsigned long flags;
	int locked = 1;
	u32 scxctr, tmp;

	local_irq_save(flags);
	if (port->uart.sysrq) {
		locked = 0;
	} else if (oops_in_progress) {
		locked = spin_trylock(&port->uart.lock);
	} else
		spin_lock(&port->uart.lock);

	scxctr = __raw_readl(port->uart.membase + SC0CTR);
	tmp = scxctr | SC01CTR_TXE;
	__raw_writel(tmp, port->uart.membase + SC0CTR);

	uart_console_write(&port->uart, s, count, serial_mn2ps_direct_console_putchar);

	serial_mn2ps_direct_wait_for_tx_complete(port);

	__raw_writel(scxctr, port->uart.membase + SC0CTR);

	if (locked)
		spin_unlock(&port->uart.lock);
	local_irq_restore(flags);
}

