/* MN2PS On-chip serial port UART driver
 *
 * Copyright (C) 2012 Panasonic Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version
 * 2 of the Licence, or (at your option) any later version.
 */

#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/serial_core.h>
#include <linux/serial_mn2ps.h>

#include <unit/serial_mn2ps.h>

#define no_printk(FMT, ...) do {} while(0)

#define kenter(FMT, ...) \
	printk(KERN_DEBUG "-->%s(" FMT ")\n", __func__, ##__VA_ARGS__)
#define _enter(FMT, ...) \
	no_printk(KERN_DEBUG "-->%s(" FMT ")\n", __func__, ##__VA_ARGS__)
#define kdebug(FMT, ...) \
	printk(KERN_DEBUG "--- " FMT "\n", ##__VA_ARGS__)
#define _debug(FMT, ...) \
	no_printk(KERN_DEBUG "--- " FMT "\n", ##__VA_ARGS__)
#define kproto(FMT, ...) \
	printk(KERN_DEBUG "### MNSERIAL " FMT " ###\n", ##__VA_ARGS__)
#define _proto(FMT, ...) \
	no_printk(KERN_DEBUG "### MNSERIAL " FMT " ###\n", ##__VA_ARGS__)


#define txfifo_can_transfer(scastr)	((((scastr) & SIOTOP_SCASTRN_TBSY) == SIOTOP_SCASTRN_TBSY_OK))
#define txfifo_usable_size(txtva)	(txtva)
#define txfifo_is_empty(txtva)		((txtva) == SERIAL_FIFO_SIZE)

static unsigned int serial_mn2ps_fifo_tx_empty(struct uart_port *);
static void serial_mn2ps_fifo_set_mctrl(struct uart_port *, unsigned int);
static unsigned int serial_mn2ps_fifo_get_mctrl(struct uart_port *);
static void serial_mn2ps_fifo_stop_tx(struct uart_port *);
static void serial_mn2ps_fifo_start_tx(struct uart_port *);
static void serial_mn2ps_fifo_stop_rx(struct uart_port *);
static void serial_mn2ps_fifo_enable_ms(struct uart_port *);
static void serial_mn2ps_fifo_break_ctl(struct uart_port *, int);
static int serial_mn2ps_fifo_startup(struct uart_port *);
static void serial_mn2ps_fifo_shutdown(struct uart_port *);
static void serial_mn2ps_fifo_set_termios(struct uart_port *,
					 struct ktermios *,
					 struct ktermios *);
static const char *serial_mn2ps_fifo_type(struct uart_port *);
static void serial_mn2ps_fifo_release_port(struct uart_port *);
static int serial_mn2ps_fifo_request_port(struct uart_port *);
static void serial_mn2ps_fifo_config_port(struct uart_port *, int);
static int serial_mn2ps_fifo_verify_port(struct uart_port *,
					struct serial_struct *);

const struct uart_ops serial_mn2ps_fifo_ops = {
	.tx_empty	= serial_mn2ps_fifo_tx_empty,
	.set_mctrl	= serial_mn2ps_fifo_set_mctrl,
	.get_mctrl	= serial_mn2ps_fifo_get_mctrl,
	.stop_tx	= serial_mn2ps_fifo_stop_tx,
	.start_tx	= serial_mn2ps_fifo_start_tx,
	.stop_rx	= serial_mn2ps_fifo_stop_rx,
	.enable_ms	= serial_mn2ps_fifo_enable_ms,
	.break_ctl	= serial_mn2ps_fifo_break_ctl,
	.startup	= serial_mn2ps_fifo_startup,
	.shutdown	= serial_mn2ps_fifo_shutdown,
	.set_termios	= serial_mn2ps_fifo_set_termios,
	.type		= serial_mn2ps_fifo_type,
	.release_port	= serial_mn2ps_fifo_release_port,
	.request_port	= serial_mn2ps_fifo_request_port,
	.config_port	= serial_mn2ps_fifo_config_port,
	.verify_port	= serial_mn2ps_fifo_verify_port,
};


static void serial_mn2ps_fifo_receive_interrupt(struct serial_mn2ps_port *);
static void serial_mn2ps_fifo_transmit_interrupt(struct serial_mn2ps_port *);
static irqreturn_t serial_mn2ps_fifo_interrupt(int, void *);
static int serial_mn2ps_fifo_wait_for_enable_transfer(struct serial_mn2ps_port *);
static int serial_mn2ps_fifo_wait_for_tx_complete(struct serial_mn2ps_port *);
static void serial_mn2ps_fifo_dis_tx_intr(struct serial_mn2ps_port *);
static void serial_mn2ps_fifo_en_tx_intr(struct serial_mn2ps_port *);
static void serial_mn2ps_fifo_en_rx_intr(struct serial_mn2ps_port *);
static void serial_mn2ps_fifo_dis_rx_intr(struct serial_mn2ps_port *);
static void serial_mn2ps_fifo_console_putchar(struct uart_port *, int);

/********************************************************************
 * DESCRIPTION	:
 *		disable the Transmit Interrupt
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void __attribute__((__unused__)) serial_mn2ps_fifo_dis_tx_intr(struct serial_mn2ps_port *port)
{

	u32 contrl;

	contrl = __raw_readl(port->uart.membase + SIOTOP_TXCNTN);
	contrl &= ~SIOTOP_TXCNTN_EI;
	__raw_writel(contrl, port->uart.membase + SIOTOP_TXCNTN);
}

/********************************************************************
 * DESCRIPTION	:
 *		enable the Transmit Interrupt
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_en_tx_intr(struct serial_mn2ps_port *port)
{

	u32 contrl;

	contrl = __raw_readl(port->uart.membase + SIOTOP_TXCNTN);
	contrl |= SIOTOP_TXCNTN_EI;
	__raw_writel(contrl, port->uart.membase + SIOTOP_TXCNTN);
}

/********************************************************************
 * DESCRIPTION	:
 *		disable the Receive Interrupt
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void __attribute__((__unused__)) serial_mn2ps_fifo_dis_rx_intr(struct serial_mn2ps_port *port)
{

	u32 contrl;

	contrl = __raw_readl(port->uart.membase + SIOTOP_RXCNTN);
	contrl &= ~SIOTOP_RXCNTN_EI;
	__raw_writel(contrl, port->uart.membase + SIOTOP_RXCNTN);
}

/********************************************************************
 * DESCRIPTION	:
 *		enable the Receive Interrupt
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void __attribute__((__unused__)) serial_mn2ps_fifo_en_rx_intr(struct serial_mn2ps_port *port)
{

	u32 contrl;

	contrl = __raw_readl(port->uart.membase + SIOTOP_RXCNTN);
	contrl |= SIOTOP_RXCNTN_EI;
	__raw_writel(contrl, port->uart.membase + SIOTOP_RXCNTN);
}

/********************************************************************
 * DESCRIPTION	:
 *		receive chars from the ring buffer for this serial port
 *		- must do break detection here (not done in the UART)
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_receive_interrupt(struct serial_mn2ps_port *port)
{

	struct uart_icount *icount = &port->uart.icount;
	struct tty_struct *tty = port->uart.state->port.tty;
	u32 status, rxtva;
	int max_count = 256;
	unsigned char ch;
	char flag;

	status = __raw_readl(port->uart.membase + SIOTOP_STATUSN);
	rxtva = __raw_readl(port->uart.membase + SIOTOP_RXTVAN);
	while ((rxtva > 0) && (max_count-- > 0)) {
		ch = (unsigned char)__raw_readl(port->uart.membase + SIOTOP_RXDATAN);
		flag = TTY_NORMAL;
		icount->rx++;

		if (unlikely(status & SIOTOP_STATUSN_ERROR_BITS)) {
			if (status & SIOTOP_STATUSN_SCAFECLR) {
				icount->frame++;
			}
			if (status & SIOTOP_STATUSN_SCAPECLR) {
				icount->parity++;
			}
			if (status & SIOTOP_STATUSN_RXDFOR) {
				icount->overrun++;
			}

			if (status & SIOTOP_STATUSN_SCAPECLR) {
				flag = TTY_PARITY;
			} else if (status & SIOTOP_STATUSN_SCAFECLR) {
				flag = TTY_FRAME;
			}
		}

		if (uart_handle_sysrq_char(&port->uart, ch)) {
			goto ignore_char;
		}

		uart_insert_char(&port->uart, status, SIOTOP_STATUSN_RXDFOR, ch, flag);

ignore_char:
		status = __raw_readl(port->uart.membase + SIOTOP_STATUSN);
		rxtva =__raw_readl(port->uart.membase + SIOTOP_RXTVAN);
	}
	spin_unlock(&port->uart.lock);
	tty_flip_buffer_push(tty);
	spin_lock(&port->uart.lock);
}

/********************************************************************
 * DESCRIPTION	:
 *		Transmission interruption is processed.
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port.
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_transmit_interrupt(struct serial_mn2ps_port *port)
{

	struct circ_buf *xmit = &port->uart.state->xmit;
	int count;

	if (port->tx_xchar) {
		__raw_writel(port->tx_xchar, port->uart.membase + SIOTOP_TXDATAN);
		port->uart.icount.tx++;
		port->tx_xchar = 0;
		return;
	}
	if (uart_tx_stopped(&port->uart)) {
		serial_mn2ps_fifo_stop_tx(&port->uart);
		return;
	}
	if (uart_circ_empty(xmit)) {
		serial_mn2ps_fifo_stop_tx(&port->uart);
		return;
	}

	count = __raw_readl(port->uart.membase + SIOTOP_TXTVAN);
	do {
		__raw_writel(xmit->buf[xmit->tail], port->uart.membase + SIOTOP_TXDATAN);
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		port->uart.icount.tx++;
		udelay(1);
		if (uart_circ_empty(xmit)) {
			break;
		}
	} while (--count > 0);

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS) {
		uart_write_wakeup(&port->uart);
	}

	if (uart_circ_empty(xmit)) {
		serial_mn2ps_fifo_wait_for_tx_complete(port);
		serial_mn2ps_fifo_stop_tx(&port->uart);
	}
}

/********************************************************************
 * DESCRIPTION	:
 *		The handler of interruption
 *
 * INPUT	:
 *		int  irq		interrupt number
 *		void *dev_id		device id
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		irqreturn_t
 *
 *******************************************************************/
static irqreturn_t serial_mn2ps_fifo_interrupt(int irq, void *dev_id)
{

	struct serial_mn2ps_port *port = dev_id;
	unsigned long flags;
	u32 scastr, txtva, rxtva;

	spin_lock_irqsave(&port->uart.lock, flags);
	rxtva = __raw_readl(port->uart.membase + SIOTOP_RXTVAN);
	if (rxtva != 0) {
		serial_mn2ps_fifo_receive_interrupt(port);
	}

	scastr = __raw_readl(port->uart.membase + SIOTOP_SCASTRN);
	txtva  = __raw_readl(port->uart.membase + SIOTOP_TXTVAN);
	if (txfifo_usable_size(txtva)) {
		serial_mn2ps_fifo_transmit_interrupt(port);
	}
	spin_unlock_irqrestore(&port->uart.lock, flags);

	return IRQ_HANDLED;
}

/********************************************************************
 * DESCRIPTION	:
 *		return indication of whether the hardware transmit buffer is empty
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		unsigned int
 *				0		not empty
 *				TIOCSER_TEMT	transmit buffer is empty
 *
 *******************************************************************/
static unsigned int serial_mn2ps_fifo_tx_empty(struct uart_port *uart)
{
	unsigned int ret = 0;
	struct serial_mn2ps_port *port =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 scastr, txtva;

	_enter("%s", port->name);

	scastr = __raw_readl(port->uart.membase + SIOTOP_SCASTRN);
	txtva  = __raw_readl(port->uart.membase + SIOTOP_TXTVAN);
	if (txfifo_can_transfer(scastr) &&
		txfifo_usable_size(txtva)) {
		ret = TIOCSER_TEMT;
	}
	return ret;
}

/********************************************************************
 * DESCRIPTION	:
 *		set the modem control lines (we don't have any)
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		unsigned int mctrl	modem control flag
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_set_mctrl(struct uart_port *uart,
				     unsigned int mctrl)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s,%x", port->name, mctrl);
}

/********************************************************************
 * DESCRIPTION	:
 *		get the modem control line statuses
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		unsigned int	modem control line status
 *
 *******************************************************************/
static unsigned int serial_mn2ps_fifo_get_mctrl(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	return TIOCM_CAR | TIOCM_CTS | TIOCM_DSR;
}

/********************************************************************
 * DESCRIPTION	:
 *		stop transmitting characters
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_stop_tx(struct uart_port *uart)
{

	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s", port->name);

	control = __raw_readl(uart->membase + SIOTOP_TXCNTN);
	control &= ~SIOTOP_TXCNTN_EI;
	__raw_writel(control, uart->membase + SIOTOP_TXCNTN);
}

/********************************************************************
 * DESCRIPTION	:
 *		start transmitting characters
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_start_tx(struct uart_port *uart)
{

	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	serial_mn2ps_fifo_en_tx_intr(port);
}

/********************************************************************
 * DESCRIPTION	:
 *		stop receiving characters
 *		- called whilst the port is being closed
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_stop_rx(struct uart_port *uart)
{

	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s", port->name);

	control = __raw_readl(uart->membase + SIOTOP_RXCNTN);
	control &= ~(SIOTOP_RXCNTN_RXD_EN | SIOTOP_RXCNTN_EI);
	__raw_writel(control, uart->membase + SIOTOP_RXCNTN);
}

/********************************************************************
 * DESCRIPTION	:
 *		enable modem status interrupts
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_enable_ms(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

}

/********************************************************************
 * DESCRIPTION	:
 *		transmit or cease transmitting a break signal
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		int ctl			break control flag
 *					  1: break signal on
 *					  0: break signal off
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_break_ctl(struct uart_port *uart, int ctl)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control;

	_enter("%s,%d", port->name, ctl);

	control = __raw_readl(port->uart.membase + SIOTOP_SCACTRN);
	if (ctl) {
		port->tx_break = 1;
		control |= SIOTOP_SCACTRN_SCABRE;
	} else {
		port->tx_break = 0;
		control &= ~SIOTOP_SCACTRN_SCABRE;
	}
	serial_mn2ps_fifo_en_tx_intr(port);
}

/********************************************************************
 * DESCRIPTION	:
 *		grab the interrupts and enable the port for reception
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int
 *			0
 *			-EBUSY	failure to register
 *
 *******************************************************************/
static int serial_mn2ps_fifo_startup(struct uart_port *uart)
{

	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control, threshold;

	_enter("%s,%d", port->name, ctl);

	if (request_irq(port->rx_irq, serial_mn2ps_fifo_interrupt,
			IRQF_DISABLED, port->rx_name, port) < 0) {
		goto error1;
	}
	if (request_irq(port->tx_irq, serial_mn2ps_fifo_interrupt,
			IRQF_DISABLED, port->tx_name, port) < 0) {
		goto error2;
	}

	/* set TX threshold */
	threshold = SERIAL_TX_THRESHOLD_VALUE;
	__raw_writel(threshold, uart->membase + SIOTOP_TXTRSN);

	/* send FIFO clear, intr enable */
	control = __raw_readl(uart->membase + SIOTOP_TXCNTN);
	control |= (SIOTOP_TXCNTN_TXD_EN | SIOTOP_TXCNTN_FIFO_CLR);
	__raw_writel(control, uart->membase + SIOTOP_TXCNTN);

	/* set RX threshold */
	threshold = SERIAL_RX_THRESHOLD_VALUE;
	__raw_writel(threshold, uart->membase + SIOTOP_RXTRSN);

	/* receive FIFO clear, intr enable */
	control = __raw_readl(uart->membase + SIOTOP_RXCNTN);
	control |= (SIOTOP_RXCNTN_RXD_EN | SIOTOP_RXCNTN_EI | SIOTOP_RXCNTN_FIFO_CLR);
	__raw_writel(control, uart->membase + SIOTOP_RXCNTN);

	return 0;

error2:
	free_irq(port->rx_irq, port);
error1:
	return -EBUSY;
}

/********************************************************************
 * DESCRIPTION	:
 *		shutdown the port and release interrupts
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_shutdown(struct uart_port *uart)
{

	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	u32 control, config;

	/* IP reset */
	config = __raw_readl(uart->membase + SIOTOP_CNFIGN);
	config |= SIOTOP_CNFIGN_RESET;
	__raw_writel(config, uart->membase + SIOTOP_CNFIGN);

	/* receve: intr disable / FIFO clear / function disable */
	control = __raw_readl(uart->membase + SIOTOP_RXCNTN);
	control &= ~(SIOTOP_RXCNTN_EI | SIOTOP_RXCNTN_RXD_EN);
	control |= SIOTOP_RXCNTN_FIFO_CLR;
	__raw_writel(control, uart->membase + SIOTOP_RXCNTN);

	/* send: intr disable / FIFO clear / function disable */
	control = __raw_readl(uart->membase + SIOTOP_TXCNTN);
	control &= ~(SIOTOP_TXCNTN_EI | SIOTOP_TXCNTN_TXD_EN);
	control |= SIOTOP_TXCNTN_FIFO_CLR;
	__raw_writel(control, uart->membase + SIOTOP_TXCNTN);

	control = __raw_readl(uart->membase + SIOTOP_SCACTRN);
	control &= (~SIOTOP_SCACTRN_SCATEN & ~SIOTOP_SCACTRN_SCAREN);
	__raw_writel(control, uart->membase + SIOTOP_SCACTRN);

	/* Intr Handler unregister */
	free_irq(port->rx_irq, port);
	free_irq(port->tx_irq, port);
}

/********************************************************************
 * DESCRIPTION	:
 *		set the terminal I/O parameters
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		struct ktermios *new	termios to be set up from now on
 *		struct ktermios *old	termios before a setup
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_set_termios(struct uart_port *uart,
					 struct ktermios *new,
					 struct ktermios *old)
{

	struct serial_mn2ps_port *port __attribute__ ((unused)) =
	container_of(uart, struct serial_mn2ps_port, uart);
	unsigned cflag;
	int baud, bits, xdiv;
	u32 cycle, scactr, tmp, cnfig, cnt=0;
	unsigned long flags;

	/* IP reset */
	cnfig = __raw_readl(port->uart.membase + SIOTOP_CNFIGN);
	cnfig |= SIOTOP_CNFIGN_RESET;
	__raw_writel(cnfig, port->uart.membase + SIOTOP_CNFIGN);

	/* Set byte size */
	cflag = new->c_cflag;
	switch (cflag & CSIZE) {
	case CS7: scactr = SIOTOP_SCACTRN_CLN_7BIT; bits = 9;  break;
	case CS8: scactr = SIOTOP_SCACTRN_CLN_8BIT; bits = 10; break;
	default:  scactr = SIOTOP_SCACTRN_CLN_8BIT; bits = 10; break;
	}

	/* Set stop bit mode */
	if (cflag & CSTOPB) {
		scactr |= SIOTOP_SCACTRN_STB_2BIT;
		bits++;
	}

	/* Set parity bit mode */
	if (cflag & PARENB) {
		bits++;
		if (cflag & PARODD) {
			scactr |= SIOTOP_SCACTRN_ODD;
		} else if (cflag & CMSPAR) {
			scactr |= SIOTOP_SCACTRN_FIXED0;
		} else {
			scactr |= SIOTOP_SCACTRN_EVEN;
		}
	}

	/* change Transfer bit-order (LSB/MSB) */
	if (cflag & CODMSB) {
		scactr |= SIOTOP_SCACTRN_SCAOB_MSB; /* MSB MODE */
	} else {
		scactr &= ~SIOTOP_SCACTRN_SCAOB_MSB; /* LSB MODE */
	}

	/* change protocol start-stop transmission */
	scactr |= (SIOTOP_SCACTRN_SCAPTL_SYNC | SIOTOP_SCACTRN_CK_IOCLK_16);
	port->uart.uartclk = port->ioclk / 16;

	/* Determine divisor based on baud rate */
	baud = uart_get_baud_rate(&port->uart, new, old, 0,
				  port->uart.uartclk);
	_debug("ALT [baud %d]", baud);

	if (!baud) {
		baud = 9600;	/* B0 transition handled in rs_set_termios */
	}
	xdiv = 1;
	if (baud == 134) {
		baud = 269;	/* 134 is really 134.5 */
		xdiv = 2;
	}

	cycle = port->uart.uartclk / (baud / xdiv) - 2;
	_debug("UARTCLK: %u / %hu", port->uart.uartclk, cycle);

	/* make the changes */
	spin_lock_irqsave(&port->uart.lock, flags);

	uart_update_timeout(&port->uart, new->c_cflag, baud);

	/* set prescaler */
	__raw_writel(cycle, port->uart.membase + SIOTOP_CYCLEN);

	while(cnt<0xFF){
		cnt++;
	}

	tmp = 0 | SIOTOP_SCACTRN_CK_IOCLK_16;
	__raw_writel(tmp, port->uart.membase + SIOTOP_SCACTRN);

	/* set up parity check flag */
	port->uart.read_status_mask = (1 << TTY_NORMAL) | (1 << TTY_OVERRUN);
	if (new->c_iflag & INPCK) {
		port->uart.read_status_mask |=
			(1 << TTY_PARITY) | (1 << TTY_FRAME);
	}
	if (new->c_iflag & (BRKINT | PARMRK)) {
		port->uart.read_status_mask |= (1 << TTY_BREAK);
	}

	/* characters to ignore */
	port->uart.ignore_status_mask = 0;
	if (new->c_iflag & IGNPAR) {
		port->uart.ignore_status_mask |=
			(1 << TTY_PARITY) | (1 << TTY_FRAME);
	}
	if (new->c_iflag & IGNBRK) {
		port->uart.ignore_status_mask |= (1 << TTY_BREAK);
		/*
		 * If we're ignoring parity and break indicators,
		 * ignore overruns to (for real raw support).
		 */
		if (new->c_iflag & IGNPAR) {
			port->uart.ignore_status_mask |= (1 << TTY_OVERRUN);
		}
	}

	/* Ignore all characters if CREAD is not set */
	if ((new->c_cflag & CREAD) == 0) {
		port->uart.ignore_status_mask |= (1 << TTY_NORMAL);
	}

	tmp = __raw_readl(port->uart.membase + SIOTOP_SCACTRN);
	scactr |= (tmp | (SIOTOP_SCACTRN_SCATEN | SIOTOP_SCACTRN_SCAREN));
	__raw_writel(scactr, port->uart.membase + SIOTOP_SCACTRN);

	cnfig = __raw_readl(port->uart.membase + SIOTOP_CNFIGN);
	cnfig |= (SIOTOP_CNFIGN_RXC_TXC | SIOTOP_CNFIGN_DIR_DI | SIOTOP_CNFIGN_DOE);
	__raw_writel(cnfig, port->uart.membase + SIOTOP_CNFIGN);

	spin_unlock_irqrestore(&port->uart.lock, flags);
}

/********************************************************************
 * DESCRIPTION	:
 *		return description of port type
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		const char	port description
 *
 *******************************************************************/
static const char *serial_mn2ps_fifo_type(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	return "MN2PS FIFO SIF";
}

/********************************************************************
 * DESCRIPTION	:
 *		request I/O and memory regions for port
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int	error code
 *			0	: normal
 *			-EBUSY	: error
 *
 *******************************************************************/
static int serial_mn2ps_fifo_request_port(struct uart_port *uart)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);
	int ret;

	_enter("%s", port->name);

	if (!port->uart.membase) {
		ret = serial_mn2ps_ioremap(port);
		if (!ret) {
			goto request_out;
		}
	}

	request_mem_region((unsigned long) port->uart.membase, SC_MAP_SIZE, port->name);
	return 0;
request_out:
	return ret;
}

/********************************************************************
 * DESCRIPTION	:
 *		configure the type and reserve the ports
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		int type
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_config_port(struct uart_port *uart, int type)
{
	struct serial_mn2ps_port *port __attribute__ ((unused)) =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	port->uart.type = PORT_MN2PS_FIFO;

	serial_mn2ps_fifo_request_port(uart);
}

/********************************************************************
 * DESCRIPTION	:
 *		release I/O and memory regions in use by port
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_release_port(struct uart_port *uart)
{
	struct serial_mn2ps_port *port =
		container_of(uart, struct serial_mn2ps_port, uart);

	_enter("%s", port->name);

	iounmap(port->uart.membase);

	release_mem_region((unsigned long) port->uart.membase, SC_MAP_SIZE);
}

/********************************************************************
 * DESCRIPTION	:
 *		verify serial parameters are suitable for this port type
 *
 * INPUT	:
 *		struct uart_port *uart		uart port
 *		struct serial_struct *ss	serial information
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int	0      : normal
 *			EINVAL : error
 *
 *******************************************************************/
static int serial_mn2ps_fifo_verify_port(struct uart_port *uart,
					struct serial_struct *ss)
{
	struct serial_mn2ps_port *port =
		container_of(uart, struct serial_mn2ps_port, uart);
	void *mapbase = (void *) (unsigned long) port->uart.mapbase;

	_enter("%s", port->name);

	/* these things may not be changed */
	if (ss->irq		!= port->uart.irq	||
	    ss->port		!= port->uart.iobase	||
	    ss->io_type		!= port->uart.iotype	||
	    ss->iomem_base	!= mapbase ||
	    ss->iomem_reg_shift	!= port->uart.regshift	||
	    ss->hub6		!= port->uart.hub6	||
	    ss->xmit_fifo_size	!= port->uart.fifosize) {
		return -EINVAL;
	}

	if (ss->type != port->uart.type) {
          return -EINVAL;
        }

	if (ss->type != PORT_MN2PS &&
            ss->type != PORT_MN2PS_FIFO) {
			return -EINVAL;
	}

	return 0;
}

/********************************************************************
 * DESCRIPTION	:
 *		It waits until transmission becomes possible.
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int		0        : normal
 *				EREMOTEIO: time out error
 *
 *******************************************************************/
static int serial_mn2ps_fifo_wait_for_enable_transfer(struct serial_mn2ps_port *port)
{
	int ret = 0;
	unsigned int tmout = 10000;
	u32 scastr, txtva;

	scastr = __raw_readl(port->uart.membase + SIOTOP_SCASTRN);
	txtva  = __raw_readl(port->uart.membase + SIOTOP_TXTVAN);
	while (!(txfifo_can_transfer(scastr) &&
		 txfifo_usable_size(txtva))) {
		if (--tmout == 0) {
			ret = -EREMOTEIO;
			break;
		}
		udelay(1);
		scastr = __raw_readl(port->uart.membase + SIOTOP_SCASTRN);
		txtva  = __raw_readl(port->uart.membase + SIOTOP_TXTVAN);
	}
	return ret;
}

/********************************************************************
 * DESCRIPTION	:
 *		It waits until transmission complete.
 *
 * INPUT	:
 *		struct serial_mn2ps_port *port	serial port
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		int		0        : normal
 *				EREMOTEIO: time out error
 *
 *******************************************************************/
static int serial_mn2ps_fifo_wait_for_tx_complete(struct serial_mn2ps_port *port)
{
	int ret = 0;
	unsigned int tmout = 10000;
	u32 txtva;

	txtva  = __raw_readl(port->uart.membase + SIOTOP_TXTVAN);
 	while (!txfifo_is_empty(txtva)) {
		if (--tmout == 0) {
			ret = -EREMOTEIO;
			break;
		}
		udelay(1);
		txtva  = __raw_readl(port->uart.membase + SIOTOP_TXTVAN);
	}
	return ret;
}

/********************************************************************
 * DESCRIPTION	:
 *		One character is outputted in console processing
 *
 * INPUT	:
 *		struct uart_port *uart	uart port
 *		int ch			output character
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
static void serial_mn2ps_fifo_console_putchar(struct uart_port *uart, int ch)
{
	struct serial_mn2ps_port *port =
		container_of(uart, struct serial_mn2ps_port, uart);

	/* ch -> trans buffer */
	if (!serial_mn2ps_fifo_wait_for_enable_transfer(port)){
		__raw_writel(ch, uart->membase + SIOTOP_TXDATAN);
	}
}

/********************************************************************
 * DESCRIPTION	:
 *		print a string to the serial port without disturbing the real user of the
 *		port too much
 *		- the console must be locked by the caller
 *
 * INPUT	:
 *		struct console *co	console
 *		const char *s		pointer of output buffer
 *		unsigned count		output count
 *
 * OUTPUT	:
 *		Nothing
 *
 * RETURN	:
 *		void
 *
 *******************************************************************/
void serial_mn2ps_fifo_console_write(struct serial_mn2ps_port *port,
                                       const char *s, unsigned count)
{
	unsigned long flags;
	int locked = 1;
	u32 control, tmp;

	local_irq_save(flags);
	if (port->uart.sysrq) {
		locked = 0;
	} else if (oops_in_progress) {
		locked = spin_trylock(&port->uart.lock);
	} else {
		spin_lock(&port->uart.lock);
	}

	control = __raw_readl(port->uart.membase + SIOTOP_TXCNTN);
	tmp = control;
	if (!(tmp & SIOTOP_TXCNTN_TXD_EN)) {
		tmp &= ~SIOTOP_TXCNTN_EI;
		tmp |= (SIOTOP_TXCNTN_FIFO_CLR | SIOTOP_TXCNTN_TXD_EN);
		__raw_writel(tmp, port->uart.membase + SIOTOP_TXCNTN);
	}

	uart_console_write(&port->uart, s, count, serial_mn2ps_fifo_console_putchar);

	serial_mn2ps_fifo_wait_for_tx_complete(port);

	__raw_writel(control, port->uart.membase + SIOTOP_TXCNTN);

	if (locked) {
		spin_unlock(&port->uart.lock);
	}
	local_irq_restore(flags);
}
